/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_ACCOUNTING_ACCOUNT_CODE_H_
#define DRN_ACCOUNTING_ACCOUNT_CODE_H_

#include <QString>
#include <QtGlobal>

#include <compare>
#include <iosfwd>
#include <optional>
#include <set>

#include <accounting_export.h>

#include "AccountNumber.h"


class QDebug;

namespace drn::accounting
{

/**
 * @brief The representation of a unique identifier and its associated name for an account. The only
 * identifying item is the number, the name is a convenience for the user to associate the number
 * with a name.
 */
class DRN_ACCOUNTING_EXPORT AccountCode
{
	AccountNumber number_;
	::QString name_;
	std::optional<AccountNumber> parent_;

public:
	AccountCode() = default;
	/**
	 * @brief Partially initialises the object with supplied account code name, and defaults the
	 * rest.
	 *
	 * @exception AccountError When the supplied name is effectively empty.
	 * @exception AccountError When the supplied name contains the account separator.
	 * @exception AccountError When the parent is invalid.
	 *
	 * @param name The unique name for the account, e.g. Cash.
	 * @param parent (Optional) The parent account number for the account.
	 */
	explicit AccountCode(::QString name, std::optional<AccountNumber> parent = {});
	/**
	 * @brief Partially initialises the object with supplied values, and defaults the rest.
	 *
	 * @exception AccountError When the number is not set.
	 * @exception AccountError When the parent and the value are the same.
	 * @exception AccountError When the parent is invalid.
	 *
	 * @param number The unique number for the account, e.g. 1000.
	 * @param parent (Optional) The parent account number for the account.
	 */
	explicit AccountCode(AccountNumber number, std::optional<AccountNumber> parent = {});
	/**
	 * @brief Fully initialises the object.
	 *
	 * @exception AccountError When the supplied number is zero.
	 * @exception AccountError When the supplied name is effectively empty.
	 * @exception AccountError When the supplied name contains the account separator.
	 *
	 * @param number The unique number for the account, e.g. 1000.
	 * @param name The unique name for the account, e.g. Cash.
	 * @param parent (Optional) The parent account number for the account.
	 */
	AccountCode(
		AccountNumber number,
		::QString name,
		std::optional<AccountNumber> parent = {}
	);
	/**
	 * @brief Compares the calling object's identifier with another identifier, establishing their
	 * relative order.
	 *
	 * @param other The identifier to compare against the calling object's identifier.
	 *
	 * @return Returns a comparison result of less, greater, or equal, representing the relative
	 * order between the two identifiers' account number.
	 */
	[[nodiscard]]
	std::strong_ordering operator<=>(const AccountCode& other) const noexcept;
	/**
	 * @brief Determines whether the current account code is equal to another account code.
	 *
	 * @param other The account code to compare with the current account code.
	 *
	 * @return True if the current account code’s unique identifier, i.e. account number, is equal
	 * to the provided account code; otherwise, false.
	 */
	[[nodiscard]]
	bool operator==(const AccountCode& other) const noexcept;

	[[nodiscard]]
	bool operator!=(const AccountCode&) const noexcept = default;
	/**
	 * @brief The numeric identifier for the account.
	 */
	const AccountNumber& number() const;
	/**
	 * @brief Changes the number for the account.
	 *
	 * @exception Error When the parent and the number are the same.
	 *
	 * @param number The new number for the account.
	 */
	void setNumber(AccountNumber number);
	/**
	 * @brief The human friendly name for the account.
	 */
	const ::QString& name() const;
	/**
	 * @brief Changes the human friendly name for the account.
	 *
	 * @exception AccountError When the supplied identifier is effectively empty.
	 *
	 * @param name The new name for the account.
	 */
	void setName(::QString name);
	/**
	 * @brief The number of the parent account for this account.
	 */
	const std::optional<AccountNumber>& parent() const;
	/**
	 * @brief Changes the parent for the account.
	 *
	 * @exception Error When the parent is invalid.
	 * @exception Error When the parent and the current number are the same.
	 *
	 * @param parent The new parent for the account.
	 */
	void setParent(std::optional<AccountNumber> parent);

	/// The character separator to place between the number and name when outputting the account
	/// code, currently an em-dash.
	static const ::QString separator_;
};

DRN_ACCOUNTING_EXPORT bool operator==(const AccountCode& lhs, const AccountNumber& rhs);
DRN_ACCOUNTING_EXPORT bool operator!=(const AccountCode& lhs, const AccountNumber& rhs);
DRN_ACCOUNTING_EXPORT bool operator==(const AccountNumber& lhs, const AccountCode& rhs);
DRN_ACCOUNTING_EXPORT bool operator!=(const AccountNumber& lhs, const AccountCode& rhs);
DRN_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& o, const AccountCode& ac);
DRN_ACCOUNTING_EXPORT ::QDebug operator<<(::QDebug o, const AccountCode& ac);
/**
 * @brief Converts a string representation of the account code into the strong type.
 *
 * @throw Error When the supplied string form of the account code is not convertible.
 * @throw AccountError When the supplied string is an invalid account code.
 *
 * @param accountCode The account code to convert.
 *
 * @return The strong form of the account code.
 */
DRN_ACCOUNTING_EXPORT AccountCode toAccountCode(const ::QString& accountCode);
/**
 * @brief Converts a string representation of the account code and parent into the strong type.
 *
 * @throw Error When the supplied string form of the account code is not convertible.
 * @throw AccountError When the supplied string is an invalid account code.
 *
 * @param accountCode The account code to convert.
 * @param parentNumber (Optional) The number for the parent account.
 *
 * @return The strong form of the account code.
 */
DRN_ACCOUNTING_EXPORT AccountCode toFullAccountCode(
	const ::QString& accountCode,
	const std::optional<::QString>& parentNumber
);
/**
 * @brief Finds an AccountCode in a container of account codes based on a given account number.
 *
 * @param an The account number to search for.
 * @param codes The container of account codes to search in.
 *
 * @return If an account code with the given account number is found, it gives the account code. If
 *  not found, an empty optional will be returned.
 */
DRN_ACCOUNTING_EXPORT std::optional<AccountCode> find(
	const AccountNumber& an,
	const std::set<AccountCode>& codes
) noexcept;
/**
 * @brief Converts an account code into a human-readable form.
 *
 * @param ac The account code to convert.
 *
 * @return The account code in the form: NumberSeparatorName.
 */
DRN_ACCOUNTING_EXPORT ::QString presentationText(const AccountCode& ac);

}

Q_DECLARE_METATYPE(drn::accounting::AccountCode)

#endif
