/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_ACCOUNTING_ACCOUNT_NUMBER_H_
#define DRN_ACCOUNTING_ACCOUNT_NUMBER_H_

#include <QMetaType>
#include <QtGlobal>

#include <iosfwd>

#include <accounting_export.h>


class QDebug;
class QString;

namespace drn::accounting
{

/**
 * @brief Represents a ledger book’s account through-out the entire accounting system, and is its
 * unique identifier. It encapsulates the functionality for storing and managing this identifier.
 */
class DRN_ACCOUNTING_EXPORT AccountNumber
{
public:
	/// @brief The type used to store an account’s number. The size is large enough to accommodate
	/// different accounting number styles, e.g. sequential, type-sequential, etc.
	using IntegerType = quint32;

private:
	IntegerType value_;

public:
	/// The number that is an invalid for an account code.
	static const IntegerType invalid_;
	/// The number of characters that the account code number should always be displayed as,
	/// currently four.
	static const quint8 fieldWidth_;

	/**
	 * @brief Default initialises the account number to an invalid state.
	 */
	AccountNumber();
	/**
	 * @brief Fully initialises the account number.
	 *
	 * @exception Error When the value is not set.
	 *
	 * @param value The value for the account.
	 */
	explicit AccountNumber(const IntegerType value);
	/**
	 * @brief The underlying value of the account number.
	 */
	[[nodiscard]]
	const IntegerType& integer() const noexcept;

	[[nodiscard]]
	std::strong_ordering operator<=>(const AccountNumber&) const noexcept = default;

	[[nodiscard]]
	bool operator==(const AccountNumber&) const noexcept = default;
	/**
	 * @brief Determines if the account number is equal to another integer value.
	 *
	 * @param other The integer value to compare against the current account number.
	 *
	 * @return A boolean value. Returns true if the integer value is equal to the current account
	 * number; otherwise, returns false.
	 */
	[[nodiscard]]
	bool operator==(const IntegerType& other) const noexcept;

	[[nodiscard]]
	bool operator!=(const AccountNumber&) const noexcept = default;
};

/**
 * @brief Full object standard output stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& out, const AccountNumber& data);
/**
 * @brief Full object Qt stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_ACCOUNTING_EXPORT ::QDebug operator<<(::QDebug out, const AccountNumber& data);
/**
 * @brief Generates the presentation text for a given account number.
 *
 * @param an The account number whose presentation text is to be generated.
 *
 * @return A formatted account number as a human-readable string by applying specific formatting
 * rules.
 */
DRN_ACCOUNTING_EXPORT ::QString presentationText(const AccountNumber& an);

}
namespace drn::accounting::literals
{

/**
 * @brief Provides the means to create an account number from a literal.
 *
 * @param value The literal value to be converted into an account number.
 *
 * @return The account number that was created from the literal.
 */
DRN_ACCOUNTING_EXPORT AccountNumber operator""_an(const unsigned long long value);

}
Q_DECLARE_METATYPE(drn::accounting::AccountNumber)

#endif
