/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_ACCOUNTING_ACCOUNT_TYPES_H_
#define DRN_ACCOUNTING_ACCOUNT_TYPES_H_

#include <QtGlobal>

#include <array>
#include <iosfwd>

#include <accounting_export.h>


class QString;

namespace drn::accounting
{

/**
 * @brief This represents the various forms an account may have within the accounting system.
 */
enum class DRN_ACCOUNTING_EXPORT AccountTypes : quint8
{
	/// This represents an account that is not initialised.
	Unknown,
	/// This represent the different types of economic resources owned or controlled by an entity.
	Asset,
	/// This represents the residual capital of an entity.
	Capital,
	/// This represents an entity's expenditures.
	Expense,
	/// This represents the different types of economic obligations of an entity.
	Liability,
	/// This represents an entity's earnings and/or income.
	Revenue,
};

/**
 * @brief A debug stream insertion operator.
 *
 * @param out The stream to place the information to.
 * @param at The information to place into the stream.
 *
 * @return The stream after the modification.
 */
DRN_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& out, const AccountTypes& at);
/**
 * @brief Converts a string representation of an account type into an enumeration value of
 * AccountTypes.
 *
 * This function maps a given string that describes an account type into its corresponding
 * enumeration value. It is useful in scenarios where account type information is read from
 * external inputs like user input or files, ensuring that strings are validated and consistently
 * converted into system-recognizable account types.
 *
 * It guarantees that only predefined valid account types are produced, aligning with the
 * `AccountTypes` enum. This function can be used in parsing input data, validating configurations,
 * and managing account categorization programmatically.
 *
 * @exception std::logic_error Thrown if the input string does not represent a valid account type.
 * @exception std::logic_error Thrown if the conversion is attempted on a type that is not an
 * enumeration.
 *
 * @param weak A string value describing the account type to be converted.
 *
 * @return The corresponding `AccountTypes` enumeration value for the provided input string.
 */
DRN_ACCOUNTING_EXPORT AccountTypes toAccountTypes(const ::QString& weak);
/**
 * @brief Converts a given account type into its localised presentation text for display in a GUI
 * element.
 *
 * This function maps enumeration values of account types to their corresponding string
 * representations. It is typically used when displaying account types in a user interface, such as
 * dropdown menus or descriptions. This ensures consistency and user-friendly display names for
 * account types across the system.
 *
 * @exception std::runtime_error Thrown if the provided account type cannot be mapped to a valid string.
 *
 * @param at The account type to be converted into a presentation string.
 *
 * @return The localised string representation of the provided account type. The return value will
 * align with the predefined mappings of the system.
 */
DRN_ACCOUNTING_EXPORT ::QString presentationText(const AccountTypes& at);

/**
 * @brief Contains a comprehensive collection of all account types available in the system.
 *
 * This variable acts as a reference for all possible account categories within the system. It is
 * used to validate, retrieve, or display the set of account types in different contexts, such as
 * user interfaces or back-end processing.
 */
inline constexpr std::array allAccountTypes{
	AccountTypes::Unknown,
	AccountTypes::Asset,
	AccountTypes::Capital,
	AccountTypes::Expense,
	AccountTypes::Liability,
	AccountTypes::Revenue
};

}

#endif
