/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_ACCOUNTING_BALANCEADJUSTMENTS_H_
#define DRN_ACCOUNTING_BALANCEADJUSTMENTS_H_

#include <foundation/StrongTypeDef.hpp>

#include <accounting_export.h>


namespace pecunia
{

class Money;

}

namespace drn::foundation
{

class Percentage;

}
namespace drn::accounting
{

class AccountCode;
class AccountNumber;
class Transaction;

/**
 * @brief A flag for indicating operations should only happen in regards to the major currency unit.
 */
struct ByMajorUnit final : foundation::TaggedBoolean
{
	using foundation::TaggedBoolean::TaggedBoolean;
};

/**
 * @brief Performs a theoretical adjustment of a supplied balance according to accounting rules.
 *
 * @exception AccountError When neither the crediting nor debiting accounts is the requested asset
 * account.
 *
 * @param asset The code for the asset account to adjust.
 * @param balance The balance to start with in the adjustment.
 * @param txn The transaction to apply to the balance.
 *
 * @return The calculated balance after applying a transaction.
 */
DRN_ACCOUNTING_EXPORT pecunia::Money adjustAssetBalance(
	const AccountCode& asset,
	const pecunia::Money& balance,
	const Transaction& txn
);
/**
 * @brief Performs a theoretical adjustment of a supplied balance according to accounting rules.
 *
 * @exception AccountError When neither the crediting nor debiting accounts is the requested liability
 * account.
 *
 * @param liability The code for the liability account to adjust.
 * @param balance The balance to start with in the adjustment.
 * @param txn The transaction to apply to the balance.
 *
 * @return The calculated balance after applying a transaction.
 */
DRN_ACCOUNTING_EXPORT pecunia::Money adjustLiabilityBalance(
	const AccountCode& liability,
	const pecunia::Money& balance,
	const Transaction& txn
);
/**
 * @brief Performs a check to see if a supplied asset account would increase its balance in a given
 * transaction.
 *
 * @param txn The transaction to check if it increases the asset account.
 * @param account The asset account to check for an increase in the transaction.
 *
 * @return Gives true when the transaction would increase the asset's balance, false otherwise.
 */
DRN_ACCOUNTING_EXPORT bool shouldIncreaseAsset(
	const Transaction& txn,
	const AccountNumber& account
);
/**
 * @brief Performs a check to see if a supplied asset account would decrease its balance in a given
 * transaction.
 *
 * @param txn The transaction to check if it decreases the asset account.
 * @param account The asset account to check for an decrease in the transaction.
 *
 * @return Gives true when the transaction would decrease the asset's balance, false otherwise.
 */
DRN_ACCOUNTING_EXPORT bool shouldDecreaseAsset(
	const Transaction& txn,
	const AccountNumber& account
);
/**
 * @brief Performs a check to see if a supplied liability account would increase its balance in a
 * given transaction.
 *
 * @param txn The transaction to check if it increases the liability account.
 * @param account The liability account to check for an increase in the transaction.
 *
 * @return Gives true when the transaction would increase the liability's balance, false otherwise.
 */
DRN_ACCOUNTING_EXPORT bool shouldIncreaseLiability(
	const Transaction& txn,
	const AccountNumber& account
);
/**
 * @brief Performs a check to see if a supplied liability account would decrease its balance in a
 * given transaction.
 *
 * @param txn The transaction to check if it decreases the liability account.
 * @param account The liability account to check for an decrease in the transaction.
 *
 * @return Gives true when the transaction would decrease the liability's balance, false otherwise.
 */
DRN_ACCOUNTING_EXPORT bool shouldDecreaseLiability(
	const Transaction& txn,
	const AccountNumber& account
);
/**
 * @brief Determines if a given amount of money is within +/- percentage of a goal amount.
 *
 * @param value The amount to determine if it is within range.
 * @param window The percentage window +/- that the value must lie within.
 * @param goal The amount that the supplied value must be within a percentage of.
 * @param isByMajor A flag, when true, changes the check to only consider the major unit, otherwise
 * the major and minor units are used.
 *
 * @return Gives true when the supplied value is +/- a percentage (inclusive) of the goal amount,
 * false otherwise.
 */
DRN_ACCOUNTING_EXPORT bool isWithinWindow(
	const pecunia::Money& value,
	const foundation::Percentage& window,
	const pecunia::Money& goal,
	const ByMajorUnit& isByMajor = ByMajorUnit{false}
);
/**
 * @brief Determines if a given difference is within +/- percentage of a goal amount.
 *
 * @param diff The money difference to determine if it is within range.
 * @param window The percentage window +/- that the value must lie within.
 * @param goal The amount that the supplied value must be within a percentage of.
 * @param isByMajor A flag, when true, changes the check to only consider the major unit, otherwise
 * the major and minor units are used.
 *
 * @return Gives true when the supplied value is +/- a percentage (inclusive) of the goal amount,
 * false otherwise.
 */
DRN_ACCOUNTING_EXPORT bool isDifferenceWithinWindow(
	pecunia::Money diff,
	const foundation::Percentage& window,
	const pecunia::Money& goal,
	const ByMajorUnit& isByMajor = ByMajorUnit{false}
);

}

#endif
