/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_ACCOUNTING_LEDGER_BOOKS_H_
#define DRN_ACCOUNTING_LEDGER_BOOKS_H_

#include <iosfwd>
#include <map>
#include <optional>
#include <set>
#include <vector>

#include <accounting_export.h>

#include "Account.h"
#include "Transaction.h"

namespace pecunia
{

enum class Currency : quint16;

}

namespace drn::accounting
{

class AccountNumberGenerator;
enum class OpenStates : quint8;

/**
 * @brief This is responsible for recording the transactions relating to an entity's assets,
 * liabilities, capital, revenue, and expenses of a given account.
 */
struct DRN_ACCOUNTING_EXPORT LedgerAccount
{
	/**
	 * @brief Equality operator compares all members.
	 *
	 * @param other The other object whose members should all be equal.
	 *
	 * @return When all member variables are the same it'll give true, else false.
	 */
	bool operator==(const LedgerAccount& other) const;

	/// The account information for the ledger.
	Account account_;
	/// All the transactions in an account indexed by its transaction number.
	std::map<TransactionNumber, Transaction> transactions_;
};
/**
 * @brief Represents a general ledger that manages and facilitates financial transactions,
 * account ledgers, and general accounting operations.
 *
 * It is used to handle bookkeeping operations by tracking accounts and their transactions. It
 * upholds financial integrity by validating data during these operations. Furthermore, it contains
 * all the account information relating to the accounting system’s assets, liabilities, capital,
 * revenue, and expenses.
 */
class DRN_ACCOUNTING_EXPORT GeneralLedger
{
	friend DRN_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream&, const GeneralLedger&);

	/**
	 * @brief An entity to store the ledger accounts that are debited and credited in a transaction.
	 */
	struct DebitingCrediting
	{
		/// The number of the account that is debited in a transaction.
		AccountNumber debit_{};
		/// The number of the account that is credited in a transaction.
		AccountNumber credit_{};
	};

	void validateTransaction(const Transaction& txn) const;
	void performTransaction(const Transaction& transaction, const TransactionNumber& number);
	void validateAddAccount(const Account& account) const;
	void createOpeningAccount(
		AccountNumberGenerator& generator,
		const pecunia::Currency& code
	);

	std::map<AccountNumber, LedgerAccount> ledgers_;
	/// Maps a transaction number to the account code number for the debit and credit accounts.
	std::map<TransactionNumber, DebitingCrediting> accountTransactions_;
	TransactionNumber nextNumber_;
	std::optional<AccountNumber> openingAccountNumber_;

public:
	/**
	 * @brief Default construction of an empty valid general ledger.
	 */
	GeneralLedger();
	/**
	 * @brief Fully initalises the object to the supplied values.
	 *
	 * @exception AccountError When the opening account code is not present in the ledgers.
	 * @exception AccountError When the opening account code is not an AccountTypes::Capital type.
	 * @exception AccountError When the mapped account code does not match the ledger's account
	 * code.
	 * @exception AccountError When the account for a parent does not exist.
	 * @exception AccountError When a parent account refers to a child account.
	 * @exception TransactionError When a ledger account contains an invalid transaction number.
	 * @exception TransactionError When a ledger account contains a transaction for a debiting
	 * account that does not exist.
	 * @exception TransactionError When a ledger account contains a transaction for a crediting
	 * account that does not exist.
	 * @exception TransactionError When a ledger account contains a transaction that does not
	 * contain a matching transaction in the other ledger account.
	 * @exception TransactionError When a ledger account contains a transaction number that is
	 * not mapped to the same number.
	 *
	 * @param ledgers The ledgers to store in the general ledger.
	 * @param openingAccountNumber The account number to use for the opening balance when opening an
	 * account.
	 */
	explicit GeneralLedger(
		std::map<AccountNumber, LedgerAccount> ledgers,
		std::optional<AccountNumber> openingAccountNumber = {}
	);
	/**
	 * @brief Equality operator compares all members.
	 *
	 * @param other The other object whose members should all be equal.
	 *
	 * @return When all member variables are the same it'll give true, else false.
	 */
	[[nodiscard]]
	bool operator==(const GeneralLedger& other) const;
	/**
	 * @brief Inequality operator compares all members.
	 *
	 * @param other The other object whose members should be not equal.
	 *
	 * @return When any member variable is not the same it'll give true, else false.
	 */
	[[nodiscard]]
	bool operator!=(const GeneralLedger& other) const;
	/**
	 * @brief An accessor to the current ledgers in the general ledger.
	 *
	 * @return The stored ledger accounts each indexed by its account number.
	 */
	[[nodiscard]]
	const std::map<AccountNumber, LedgerAccount>& ledgers() const noexcept;
	/**
	 * @brief Determines if an account ledger book exits in the general ledger.
	 *
	 * @param code The account code whose ledger status is desired.
	 *
	 * @return Gives true when the account ledger exists, false else-wise.
	 */
	[[nodiscard]]
	bool hasLedger(const AccountCode& code) const noexcept;
	/**
	 * @brief Determines if an account ledger book exits in the general ledger.
	 *
	 * @param number The account number whose ledger status is desired.
	 *
	 * @return Gives true when the account ledger exists, false else-wise.
	 */
	[[nodiscard]]
	bool hasLedger(const AccountNumber& number) const noexcept;
	/**
	 * @brief Provides access to the ledger of a specific account.
	 *
	 * @exception AccountError When the ledger book for the supplied account code does not exist.
	 *
	 * @param code The account code whose ledger is desired.
	 *
	 * @return The immutable account ledger book.
	 */
	[[nodiscard]]
	const LedgerAccount& ledger(const AccountCode& code) const;
	/**
	 * @brief Provides access to the ledger of a specific account.
	 *
	 * @exception AccountError When the ledger book for the supplied account code does not exist.
	 *
	 * @param number The account number whose ledger is desired.
	 *
	 * @return The immutable account ledger book.
	 */
	[[nodiscard]]
	const LedgerAccount& ledger(const AccountNumber& number) const;
	/**
	 * @brief Posts a supplied transaction to the debiting and crediting accounts.
	 *
	 * @exception TransactionError When the transaction has a debiting account that is not in the
	 * ledger books.
	 * @exception TransactionError When the ledger book for the debiting account is closed.
	 * @exception TransactionError When the transaction has a crediting account that is not in the
	 * ledger books.
	 * @exception TransactionError When the ledger book for the crediting account is closed.
	 * @exception TransactionError When the transaction is not unreconciled.
	 * @exception AccountError When the transaction is for a currency that is unknown and the
	 * opening account does not exist.
	 * @exception TransactionError When the transaction is for a currency that is unknown.
	 *
	 * @param transaction The transaction that is to be posted.
	 *
	 * @return The transaction number for the posted transaction.
	 */
	[[nodiscard]]
	TransactionNumber post(const Transaction& transaction);
	/**
	 * @brief Replaces a non-reconciled transaction with a new transaction.
	 *
	 * @exception TransactionError When the new transaction has a debiting account that is not in
	 * the ledger books.
	 * @exception TransactionError When the ledger book for the debiting account is closed.
	 * @exception TransactionError When the new transaction has a crediting account that is not in
	 * the ledger books.
	 * @exception TransactionError When the ledger book for the crediting account is closed.
	 * @exception TransactionError When the supplied original transaction number does not exist.
	 * @exception TransactionError When the original transaction is already reconciled.
	 * @exception TransactionError When the new transaction is not unreconciled.
	 * @exception TransactionError When the new transaction is for a currency that is unknown.
	 *
	 * @param number The transaction number of the original transaction to replace.
	 * @param transaction The new transaction that will replace the original one.
	 */
	void repost(const TransactionNumber& number, const Transaction& transaction);
	/**
	 * @brief Adds a new account to general ledger books.
	 *
	 * @exception AccountError When the account already exists.
	 * @exception AccountError When the account is closed.
	 * @exception AccountError When the account type is not known.
	 * @exception AccountError When the account code is not valid.
	 * @exception AccountError When the parent account type is different than the adding account.
	 *
	 * @param toAdd The new account to add to the ledger books.
	 * @param generator The generator for new account numbers already to be used.
	 *
	 * @post The general ledger contains a new ledger book for the account supplied. If there's an
	 * opening balance, it is entered as the first of the previous month.
	 *
	 * @return The number of the transaction of the posted opening balance when there is one,
	 * not set when there isn't a balance to post.
	 */
	[[nodiscard]]
	std::optional<TransactionNumber> add(
		const Account& toAdd,
		AccountNumberGenerator& generator
	);
	/**
	 * @brief Attempts to remove an account from the general ledger. This operation modifies the
	 * ledger books by erasing the specified account.
	 *
	 * @exception AccountError Thrown when the account is not present in the ledger books.
	 * @exception AccountError Thrown when the account is already closed.
	 * @exception AccountError Thrown when attempting to remove an opening account with
	 * transactions.
	 * @exception AccountError Thrown when attempting to remove a parent account with dependent
	 * child accounts.
	 *
	 * @param an Refers to the unique identifier of the account to be removed.
	 *
	 * @post The specified account, if successfully removed, is no longer present in the ledger
	 * books.
	 * @post If the specified account is the opening account, it will no longer be designated as
	 * such after the removal.
	 */
	void remove(const AccountNumber& an);
	/**
	 * @brief Closes a ledger account.
	 *
	 * @exception AccountError Thrown if the account does not exist in the ledger.
	 * @exception AccountError Thrown if the account has non-reconciled transactions.
	 * @exception AccountError Thrown if the account is a parent account of other non-closed
	 * accounts.
	 *
	 * @param an The unique identifier of the account to be closed.
	 *
	 * @post The specified account is marked as closed.
	 * @post If the account is the opening account, the opening account reference is reset.
	 */
	void close(const AccountNumber& an);
	/**
	 * @brief Opens a ledger.
	 *
	 * @exception AccountError When the account does not exist.
	 * @exception AccountError When the account to open is a child account whose parent is closed.
	 *
	 * @param code The account code of the account ledger book to open.
	 */
	void open(const AccountCode& code);
	/**
	 * @brief Changes the code of a ledger book and recodes all transactions.
	 *
	 * @exception AccountError When the current account does not exist.
	 * @exception AccountError When the account is closed.
	 * @exception AccountError When the changed account code is already in use.
	 * @exception AccountError When the account for a parent does not exist.
	 * @exception AccountError When a parent account refers to a child account.
	 *
	 * @param toRecode The account code of the account ledger book to change.
	 * @param recodeAs The account code of the account ledger book to change to.
	 */
	void recode(const AccountCode& toRecode, const AccountCode& recodeAs);
	/**
	 * @brief Marks the transaction for both accounts involved in the transaction as cleared.
	 *
	 * @exception Error When the transaction number does not exist in the ledger books.
	 *
	 * @param txnNum The number of the transaction to be cleared.
	 */
	void clear(const TransactionNumber& txnNum);
	/**
	 * @brief Marks the transaction of a single ledger account as cleared.
	 *
	 * @exception TransactionError When the transaction number does not exist in the ledger books.
	 * @exception AccountError When the account number does not exist in the ledger books.
	 *
	 * @param txnNum The number of the transaction to be cleared.
	 * @param actNum The number of the ledger account whose transaction should be cleared.
	 */
	void clear(
		const TransactionNumber& txnNum,
		const AccountNumber& actNum
	);
	/**
	 * @brief Marks all the transactions supplied as cleared.
	 *
	 * @param txnNums The numbers of the transactions to be cleared.
	 */
	void clear(const std::set<TransactionNumber>& txnNums);
	/**
	 * @brief Marks the transaction supplied as unreconciled.
	 *
	 * @exception Error When the transaction number does not exist in the ledger books.
	 *
	 * @param txnNum The number of the transaction to be unreconciled.
	 */
	void unreconcile(const TransactionNumber& txnNum);
	/**
	 * @brief Marks the transaction supplied as unreconciled.
	 *
	 * @exception TransactionError When the transaction number does not exist in the ledger books.
	 * @exception AccountError When the account number does not exist in the ledger books.
	 *
	 * @param txnNum The number of the transaction to be unreconciled.
	 * @param actNum The number of the ledger account whose transaction should be cleared.
	 */
	void unreconcile(
		const TransactionNumber& txnNum,
		const AccountNumber& actNum
	);
	/**
	 * @brief Marks all the transactions supplied as unreconciled.
	 *
	 * @exception Error When a transaction number does not exist in the ledger books.
	 *
	 * @param txnNums The numbers of the transactions to be unreconciled.
	 */
	void unreconcile(const std::set<TransactionNumber>& txnNums);
	/**
	 * @brief Reconciles the transaction supplied.
	 *
	 * @exception TransactionError When the transaction number does not exist in the ledger books.
	 *
	 * @param txnNum The number of the transaction to be reconciled.
	 */
	void reconcile(const TransactionNumber& txnNum);
	/**
	 * @brief Reconciles all transactions supplied.
	 *
	 * @param txnNums The numbers of the transactions to be reconciled.
	 */
	void reconcile(const std::set<TransactionNumber>& txnNums);
	/**
	 * @brief Reconciles only those transactions supplied for a given account number.
	 *
	 * @exception TransactionError When a transaction number does not exist in the ledger books.
	 * @exception AccountError When the account to only reconcile is not a part of the transaction.
	 * @exception AccountError When the account to reconcile does not exist.
	 *
	 * @param actNum The account whose transactions are to be reconciled.
	 * @param txnNum The numbers of the transactions to be reconciled.
	 */
	void reconcile(
		const TransactionNumber& txnNum,
		const AccountNumber& actNum
	);
	/**
	 * @brief Removes the transaction supplied.
	 *
	 * @exception Error When a transaction number does not exist in the ledger books.
	 * @exception TransactionError When a transaction in the ledger books is already reconciled.
	 *
	 * @param number The number of the transaction to be removed.
	 */
	void remove(const TransactionNumber& number);
	/**
	 * @brief Removes all the transactions supplied.
	 *
	 * @exception Error When a transaction number does not exist in the ledger books.
	 * @exception TransactionError When a transaction in the ledger books is already reconciled.
	 *
	 * @param transactionNumbers The numbers of the transactions to be removed.
	 */
	void remove(const std::set<TransactionNumber>& transactionNumbers);
	/**
	 * @brief Provides all the known accounts in the general ledger.
	 */
	[[nodiscard]]
	std::set<Account> accounts() const noexcept;
	/**
	 * @brief Provides all the known account codes in the general ledger.
	 */
	[[nodiscard]]
	std::set<AccountCode> accountCodes() const noexcept;
	/**
	 * @brief Provides all the known account codes in the general ledger of a specific type.
	 *
	 * @param type The type of account to provide the codes for.
	 */
	[[nodiscard]]
	std::set<AccountCode> accountCodes(const AccountTypes& type) const noexcept;
	/**
	 * @brief Provides the account code number for the ledger book tracking the opening account
	 * balances.
	 *
	 * @return When set provides the account code number in the ledger book, otherwise null.
	 */
	[[nodiscard]]
	const std::optional<AccountNumber>& openingAccountNumber() const noexcept;
	/**
	 * @brief Determines if the general ledger is empty.
	 *
	 * @return Gives true when there are not any accounts, false otherwise.
	 */
	[[nodiscard]]
	bool isEmpty() const noexcept;
	/**
	 * @brief Converts a ledger account's running balance from one currency into another. If the
	 * opening balance account is not created, it'll be created.
	 *
	 * @exception TransactionError When the supplied currency code to convert to is unknown.
	 * @exception TransactionError When the account to convert is closed.
	 * @exception AccountError When the account to convert does not exist.
	 * @exception AccountError When the account in already in the currency code to convert to.
	 *
	 * @param convertAccount The ledger account to convert.
	 * @param code The currency to convert the ledger account into.
	 * @param generator The generator for new account numbers already to be used.
	 */
	void convertTo(
		const AccountNumber& convertAccount,
		const pecunia::Currency& code,
		AccountNumberGenerator& generator
	);
	/**
	 * @brief Determines if an account can be removed.
	 *
	 * @param code The account code that should be checked.
	 *
	 * @return Gives true when the criteria for removing an account is true, false otherwise.
	 */
	[[nodiscard]]
	bool canRemove(const AccountCode& code) const noexcept;
	/**
	 * @brief Provides the account codes for a requested selection of account numbers.
	 *
	 * @exception AccountError When an account number does not have an account code.
	 *
	 * @param accountNumbers The account numbers to get the account codes for.
	 *
	 * @return The fully populated account codes for the requested account numbers.
	 */
	[[nodiscard]]
	std::set<AccountCode> accountCodes(const std::set<AccountNumber>& accountNumbers) const;
	/**
	 * @brief Provides the account code for a requested account number.
	 *
	 * @exception AccountError When an account number does not have an account code.
	 *
	 * @param accountNumber The account number to get the account code for.
	 *
	 * @return The fully populated account code for the requested account number.
	 */
	[[nodiscard]]
	AccountCode accountCode(const AccountNumber& accountNumber) const;
	/**
	 * @brief Searches for a specific transaction.
	 *
	 * @param txnNum The transaction to search for.
	 *
	 * @return Gives the transaction found with the given number, otherwise null.
	 */
	[[nodiscard]]
	std::optional<Transaction> lookUpTransaction(
		const TransactionNumber& txnNum
	) const noexcept;
	/**
	 * @brief Finds and retrieves the identifiers of all child accounts linked to the specified
	 * parent account. This method queries accounts hierarchically tied to the given parent and
	 * returns the results for further processing or display.
	 *
	 * @param parent The identifier of the parent account whose child accounts are to be retrieved.
	 *
	 * @return A set containing the identifiers of the child accounts if there are any linked to the
	 * specified parent, or an empty set if no children exist.
	 */
	[[nodiscard]]
	std::set<AccountNumber> lookUpChildren(const AccountNumber& parent) const noexcept;
	/**
	 * @brief Identifies and retrieves all child accounts for a specified parent account, filtering
	 * by their open or closed status.
	 *
	 * This function parses through a collection of accounts, finding those that are directly
	 * subordinate to a specified parent account. It allows filtering based on whether the accounts
	 * are open, closed, or regardless of their state.
	 *
	 * @param parent The account identifier for which child accounts are sought.
	 * @param openState The desired open or closed status of accounts to be included in the result
	 * set.
	 *
	 * @return A collection of account identifiers representing the child accounts of the specified
	 * parent account, filtered by the open or closed status.
	 */
	[[nodiscard]]
	std::set<AccountNumber> lookUpChildren(
		const AccountNumber& parent,
		const OpenStates& openState
	) const noexcept;
	/**
	 * @brief Locates the opening balance transaction of an account. The oldest transaction
	 * involving the general ledger's opening account is considered to be the opening balance.
	 *
	 * @param account The account number of the ledger to search for the opening balance.
	 *
	 * @return Gives the opening transaction when it is present, null otherwise.
	 */
	[[nodiscard]]
	std::optional<Transaction> openingTransaction(
		const AccountNumber& account
	) const noexcept;
	/**
	 * @brief Processes a refund for a previously recorded transaction.
	 *
	 * @exception TransactionError if the specified original transaction does not exist.
	 * @exception TransactionError if the refund amount is greater than the original transaction
	 * amount.
	 *
	 * @param txnNum The identifier of the transaction for which the refund is to be made.
	 * @param amount The amount to be refunded.
	 *
	 * @post A new transaction is created to record the refund and is posted to the ledger.
	 *
	 * @return The transaction number of the newly created refund transaction.
	 */
	[[nodiscard]]
	TransactionNumber refund(
		const TransactionNumber& txnNum,
		const pecunia::Money& amount
	);
};

/**
 * @brief Stream insertion operator for a human readable ledger account.
 *
 * @param o The stream to insert the information into.
 * @param la The account to generate the information for.
 *
 * @post The stream's filled with the information in the form: Member Name (Member Value).
 *
 * @return The stream populated with the information.
 */
DRN_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& o, const LedgerAccount& la);
/**
 * @brief Stream insertion operator for a human readable general ledger account.
 *
 * @param o The stream to insert the information into.
 * @param la The ledgers to generate the information for.
 *
 * @post The stream's filled with the information in the form: Member Name (Member Value).
 *
 * @return The stream populated with the information.
 */
DRN_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& o, const GeneralLedger& gl);
/**
 * @brief Collects all currencies used by the accounts in the general ledger.
 *
 * @param gl The general ledger to collect from.
 *
 * @return All the unique currency codes for the accounts are using.
 */
DRN_ACCOUNTING_EXPORT std::set<pecunia::Currency> containedCurrencies(
	const GeneralLedger& gl
) noexcept;

}

#endif
