/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_ACCOUNTING_TRANSACTION_H_
#define DRN_ACCOUNTING_TRANSACTION_H_

#include <pecunia/Money.hpp>

#include <QDate>
#include <QString>

#include <compare>
#include <iosfwd>
#include <optional>

#include <accounting_export.h>

#include "Account.h"
#include "TransactionNumber.h"


class QDebug;

namespace drn::accounting::internal
{

class AccountTransactionKey;

}
namespace drn::accounting
{

/**
 * @brief Enumerates the different statuses that a financial transaction can hold.
 *
 * These statuses track the progress of a financial transaction from being unprocessed, through
 * initial clearing, to final reconciliation. Each status reflects a specific stage in the
 * transaction lifecycle.
 *
 * This is designed to assist in managing and querying the state of a transaction within financial
 * systems, providing clarity and control over its progression.
 */
enum class DRN_ACCOUNTING_EXPORT TransactionStatuses : quint8
{
	/// The transaction has not been cleared nor reconciled with the account.
	Unreconciled,
	/// The transaction has been cleared through the account, but has not been reconciled.
	Cleared,
	/// The transaction has been reconciled with the account.
	Reconciled,
};
/**
 * @brief Represents a financial transaction, typically involving the transfer of money
 * between two accounts.
 *
 * This encapsulates all relevant details associated with a financial transaction, including the
 * date of occurrence, the accounts involved (debited and credited), the amount transferred, a
 * description of the transaction, its identifier if available, and its reconciliation status. This
 * is immutable for most of its core attributes and is designed to safely manage essential
 * transaction details. Mutation of the essential details is only allowed via a key enforced by the
 * compiler.
 */
class DRN_ACCOUNTING_EXPORT Transaction final
{
	::QDate occurredOn_;
	AccountCode debiting_;
	AccountCode crediting_;
	pecunia::Money amount_;
	std::optional<::QString> description_;
	std::optional<TransactionNumber> number_;
	TransactionStatuses status_;

public:
	/**
	 * @brief Default constructor with all values defaulted and the transaction status unreconciled.
	 */
	Transaction();
	/**
	 * @brief Basic date constructor. Useful for when constructing piece-wise. All other values are
	 * defaulted and the transaction status unreconciled.
	 *
	 * @param occurredOn The date the transaction occurred on.
	 */
	explicit Transaction(const ::QDate& occurredOn);
	/**
	 * @brief New transaction initialisation constructor.
	 *
	 * @exception TransactionError When the transaction date is invalid.
	 * @exception TransactionError When the code of the debit account is the same as the code
	 * of the credit account.
	 * @exception TransactionError When the amount is less than zero.
	 *
	 * @param occurredOn The date the Transaction occurred on.
	 * @param debiting The code of the account that is debited.
	 * @param crediting The code of the account that is credited.
	 * @param amount The amount of money applied to each account.
	 * @param description A human friendly description of the transaction.
	 * @param id The identifier for the transaction in the general ledger book iff its posted.
	 * @param status The current status of the transaction.
	 */
	Transaction(
		const ::QDate& occurredOn,
		const AccountCode& debiting,
		const AccountCode& crediting,
		const pecunia::Money& amount,
		std::optional<::QString> description = std::nullopt,
		std::optional<TransactionNumber> id = std::nullopt,
		const TransactionStatuses& status = TransactionStatuses::Unreconciled
	);
	/**
	 * @brief New transaction initialisation constructor.
	 *
	 * @exception TransactionError When the transaction date is invalid.
	 * @exception TransactionError When the number of the debit account is the same as the number
	 * of the credit account.
	 * @exception TransactionError When the amount is less than zero.
	 *
	 * @param occurredOn The date the Transaction occurred on.
	 * @param debiting The ledger number of the account that is debited.
	 * @param crediting The ledger number  of the account that is credited.
	 * @param amount The amount of money applied to each account.
	 * @param description A human friendly description of the transaction.
	 * @param id The identifier for the transaction in the general ledger book iff its posted.
	 * @param status The current status of the transaction.
	 */
	Transaction(
		const ::QDate& occurredOn,
		AccountNumber debiting,
		AccountNumber crediting,
		pecunia::Money amount,
		std::optional<::QString> description = std::nullopt,
		std::optional<TransactionNumber> id = std::nullopt,
		const TransactionStatuses& status = TransactionStatuses::Unreconciled
	);

	Transaction(const Transaction&) = default;

	Transaction(Transaction&&) noexcept = default;
	/**
	 * @brief Assigns all attributes of another transaction to the current transaction.
	 *
	 * This copies the details of another transaction into the current transaction. It ensures
	 * that the current transaction has identical properties to the other transaction.
	 *
	 * @param other The transaction whose attributes are to be assigned to the current transaction.
	 *
	 * @post The current transaction object will have identical attributes to the other transaction
	 * object.
	 *
	 * @return The updated current transaction object.
	 */
	Transaction& operator=(const Transaction& other);
	/**
	 * @brief Transfers the state of another transaction object to the current object, ensuring the
	 * current object’s data reflects that of the source. This is useful for operations involving
	 * ownership transfer of resources from one transaction to another.
	 *
	 * @param other The source transaction object whose data will be transferred.
	 *
	 * @post The current transaction object will have its state updated to match that of the source
	 * transaction object.
	 * @post The source transaction object will be left in a valid but unspecified state.
	 *
	 * @return The current transaction object with updated state.
	 */
	Transaction& operator=(Transaction&& other) noexcept;
	/**
	 * @brief Performs a three-way comparison between this transaction and another transaction.
	 *
	 * The purpose is to determine the order of transactions based on a set of prioritised
	 * attributes such as status, number, etc. This is helpful in sorting, equality checks, and
	 * other operations requiring transactional order evaluation.
	 *
	 * @param other The transaction against which this transaction will be compared.
	 *
	 * @return The result of the three-way comparison.
	 */
	[[nodiscard]]
	std::strong_ordering operator<=>(const Transaction& other) const noexcept;
	/**
	 * @brief Compares this transaction with another to determine if they are equal.
	 *
	 * This method checks if two transactions are equivalent by comparing all their properties. It
	 * is useful for identifying whether two transactions represent the same entity, aiding in
	 * tasks such as deduplication, or validation within financial systems.
	 *
	 * @param other The transaction to be compared against.
	 *
	 * @return True if the transactions are equivalent; otherwise, false.
	 */
	[[nodiscard]]
	bool operator==(const Transaction& other) const noexcept;

	[[nodiscard]]
	bool operator!=(const Transaction& other) const noexcept = default;
	/**
	 * @brief Retrieves the date when the transaction occurred.
	 *
	 * @return The date when the transaction took place.
	 */
	[[nodiscard]]
	const ::QDate& occurredOn() const noexcept;
	/**
	 * @brief Retrieves the account code associated with the debiting side of the transaction.
	 *
	 *
	 * @return The account code representing the debiting part of the transaction.
	 */
	[[nodiscard]]
	const AccountCode& debiting() const noexcept;
	/**
	 * @brief Retrieves the account code associated with the crediting side of the transaction.
	 *
	 *
	 * @return The account code representing the crediting part of the transaction.
	 */
	[[nodiscard]]
	const AccountCode& crediting() const noexcept;
	/**
	 * @brief Retrieves the monetary value associated with the transaction.
	 *
	 * @return The monetary value of the transaction.
	 */
	[[nodiscard]]
	const pecunia::Money& amount() const noexcept;
	/**
	 * @brief Retrieves the description associated with the transaction if available.
	 *
	 * This provides access to a textual description that explains the transaction. This description
	 * can be useful in providing contextual information for their purpose.
	 *
	 * @return An optional value that holds the description of the transaction if one exists. If no
	 * description is available, the optional value will hold no data.
	 */
	[[nodiscard]]
	const std::optional<::QString>& description() const noexcept;
	/**
	 * @brief Retrieves the current transaction number if it exists.
	 *
	 * @return The optional transaction number associated with the transaction. If no transaction
	 * number has been associated, it returns an uninitialised optional state.
	 */
	[[nodiscard]]
	const std::optional<TransactionNumber>& number() const noexcept;
	/**
	 * @brief Changes the number of the transaction. Once set it cannot be changed.
	 *
	 * @exception TransactionError When the transaction number is already set.
	 *
	 * @param value The new unique identifier for the transaction in the accounting system.
	 * @param allowed The object that allows access to modify the transaction.
	 */
	void setNumber(TransactionNumber value, const internal::AccountTransactionKey& allowed);
	/**
	 * @brief Retrieves the current state of the transaction regarding its reconciliation status.
	 *
	 * @return The current reconciliation status of the transaction.
	 */
	[[nodiscard]]
	const TransactionStatuses& status() const noexcept;
	/**
	 * @brief Changes the transaction to have not cleared nor reconciled with the account.
	 *
	 * @exception TransactionError When the transaction is already reconciled.
	 */
	void markUnreconciled();
	/**
	 * @brief Changes the transaction to have cleared the account.
	 *
	 * @exception TransactionError When the transaction is already reconciled.
	 */
	void markCleared();
	/**
	 * @brief Changes the transaction to have reconciled with the account.
	 */
	void markReconciled();
	/**
	 * @brief Changes the debiting account code.
	 *
	 * @exception TransactionError When the new account code is the same as the crediting account.
	 *
	 * @param code The new account code.
	 * @param allowed The object that allows access to modifying the transaction.
	 */
	void changeDebit(const AccountCode& code, const internal::AccountTransactionKey& allowed);
	/**
	 * @brief Changes the crediting account code.
	 *
	 * @exception TransactionError When the new account code is the same as the debiting account.
	 *
	 * @param code The new account code.
	 * @param allowed The object that allows access to modifying the transaction.
	 */
	void changeCredit(const AccountCode& code, const internal::AccountTransactionKey& allowed);
	/**
	 * @brief Converts the transaction amount into a requested currency.
	 *
	 * @exception TransactionError When the currency of the amount is the same as the request
	 * currency.
	 *
	 * @param changeTo The currency to change the transaction into.
	 * @param allowed The object that allows access to modifying the transaction.
	 */
	void convertAmount(
		const pecunia::Currency& changeTo,
		const internal::AccountTransactionKey& allowed
	);
};

/**
 * @brief Compares two transactions to determine if they are “equal-ish” by predefined criteria. The
 * function checks multiple attributes of the transactions for equality. It is useful for
 * identifying whether two transactions likely represent the same logical operation, but not
 * necessarily the same entity, e.g. comparing two transactions created at different times.
 *
 * @param lhs The first transaction to compare with.
 * @param rhs The second transaction to compare against.
 *
 * @return True if the transactions have identical key attributes (e.g., dates, numbers, debiting
 * accounts, crediting accounts, and statuses). False otherwise.
 *
 * @note This method does not use a strict equality check but rather compares specific fields
 * to allow for a reasonable match between the two transactions.
 */
[[nodiscard]]
DRN_ACCOUNTING_EXPORT bool areEqualish(const Transaction& lhs, const Transaction& rhs);
/**
 * @brief Converts a weak string representation of a transaction status into its strong from.
 *
 * @exception logic_error When the weak representation is not valid for a transaction status.
 *
 * @param weak The weak representation of a transaction status.
 *
 * @return The strong form of the transaction status.
 */
[[nodiscard]]
DRN_ACCOUNTING_EXPORT TransactionStatuses toTransactionStatuses(const ::QString& weak);
/**
 * @brief Full object standard out stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& out, const TransactionStatuses& data);
/**
 * @brief Full object standard out stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& out, const Transaction& data);
/**
 * @brief Full object Qt stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_ACCOUNTING_EXPORT ::QDebug operator<<(::QDebug out, const TransactionStatuses& data);
/**
 * @brief Full object Qt stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_ACCOUNTING_EXPORT ::QDebug operator<<(::QDebug out, const Transaction& data);
/**
 * @brief Retrieves the common abbreviation used for a credit entry in accounting.
 *
 * This function provides a standardised shorthand for referring to credit transactions. It is useful
 * for presenting financial information in a concise and consistent manner, such as in GUI elements
 * or reports.
 *
 * @return The abbreviation representing a credit entry.
 */
[[nodiscard]]
DRN_ACCOUNTING_EXPORT const ::QString& creditAbbreviation();
/**
 * @brief Retrieves the common abbreviation used for a debit entry in accounting.
 *
 * This function provides a standardised shorthand for referring to debit transactions. It is useful
 * for presenting financial information in a concise and consistent manner, such as in GUI elements
 * or reports.
 *
 * @return The abbreviation representing a debit entry.
 */
[[nodiscard]]
DRN_ACCOUNTING_EXPORT const ::QString& debitAbbreviation();
/**
 * @brief Provides a textual representation of the given transaction status to be displayed
 * in the interface.
 *
 * @param ts The transaction status whose textual representation is to be obtained.
 *
 * @return A descriptive string corresponding to the provided transaction status.
 */
[[nodiscard]]
DRN_ACCOUNTING_EXPORT ::QString presentationText(const TransactionStatuses& ts);
/**
 * @brief Generates a formatted descriptive text for a transaction.
 *
 * This function creates a user-readable presentation of a transaction, extracting and formatting
 * details such as the transaction date, number, description, debit, credit, amount, and status.
 * It is commonly used in displaying transaction details in a GUI item or exporting transaction
 * summaries for user-facing purposes.
 *
 * @param t The transaction object whose details are to be formatted for presentation.
 *
 * @return A formatted text representation of the transaction.
 */
[[nodiscard]]
DRN_ACCOUNTING_EXPORT ::QString presentationText(const Transaction& t);

}

#endif
