/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_ACCOUNTING_TRANSACTION_NUMBER_H_
#define DRN_ACCOUNTING_TRANSACTION_NUMBER_H_

#include <QtCore>

#include <compare>
#include <iosfwd>

#include <accounting_export.h>

class QDebug;
class QString;

namespace drn::accounting
{

/**
 * @brief Represents a unique identifier for a financial transaction.
 *
 * This is designed to encapsulate the concept of a transaction number, ensuring that it is managed
 * consistently, is always valid, and adapted to specific transactional contexts where unique
 * identification is critical.
 */
class DRN_ACCOUNTING_EXPORT TransactionNumber
{
public:
	/// @brief The type used to store a transaction’s number.
	using IntegerType = quint16;

private:
	IntegerType value_;

public:
	/// The number that indicates that it is invalid and empty.
	static const IntegerType invalid_;

	/**
	 * @brief Default initialises the object to an invalid numerical value.
	 */
	TransactionNumber();
	/**
	 * @brief Fully initialises the object with the supplied values.
	 *
	 * @exception TransactionError When the supplied new value is zero.
	 *
	 * @param value The number for the transaction.
	 */
	explicit TransactionNumber(const IntegerType& value);

	[[nodiscard]]
	std::strong_ordering operator<=>(const TransactionNumber&) const noexcept = default;

	[[nodiscard]]
	bool operator==(const TransactionNumber&) const noexcept = default;

	[[nodiscard]]
	bool operator!=(const TransactionNumber&) const noexcept = default;
	/**
	 * @brief Retrieves the integer value representing the transaction number.
	 *
	 * @return The integer value that represents the transaction number.
	 */
	[[nodiscard]]
	const IntegerType& integer() const noexcept;
	/**
	 * @brief Increment the object by one.
	 *
	 * @exception Error When the result of the incrementation would exceed the limit of the integer
	 * storage type.
	 *
	 * @return The instance after the increment operation is performed.
	 */
	TransactionNumber& operator++();
	/**
	 * @brief Increment the object by one.
	 *
	 * @exception Error When the result of the incrementation would exceed the limit of the integer
	 * storage type.
	 *
	 * @return The instance before the increment operation is performed.
	 */
	[[nodiscard]]
	TransactionNumber operator++(int);
};

/**
 * @brief Full object standard output stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& out, const TransactionNumber& data);
/**
 * @brief Full object Qt stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_ACCOUNTING_EXPORT ::QDebug operator<<(::QDebug out, const TransactionNumber& data);
/**
 * @brief Generates the text representation of the given transaction number for presentation
 * purposes.
 *
 * @param tn The transaction number to be converted into a presentation text format.
 *
 * @return A textual representation of the provided transaction number.
 */
DRN_ACCOUNTING_EXPORT ::QString presentationText(const TransactionNumber& tn);

}
namespace drn::accounting::literals
{

/**
 * @brief Provides the means to create a transaction number from a literal.
 *
 * @param value The literal value to be converted into a transaction number.
 *
 * @return The transaction number that was created from the literal.
 */
DRN_ACCOUNTING_EXPORT TransactionNumber operator""_tn(const unsigned long long value);

}

#endif
