/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BANKING_BANK_ACCOUNT_H_
#define DRN_BANKING_BANK_ACCOUNT_H_

#include <compare>
#include <iosfwd>

#include <banking_export.h>
#include <foundation/PresentationTags.hpp>

#include "BankAccountType.h"
#include "BankName.h"


class QDebug;
class QString;

namespace drn::banking
{

/**
 * @brief Represents a simple bank account with basic operations.
 *
 * This class allows users to manage a bank account by performing actions such as
 * depositing funds, withdrawing funds, and checking the account balance.
 * It is designed for scenarios where basic financial account management is required.
 *
 * Use this class to maintain an account's financial state and interact with its balance
 * through a set of straightforward operations. You can create an account, deposit or withdraw
 * funds, and obtain the current account balance.
 */
struct DRN_BANKING_EXPORT BankAccount
{
	std::strong_ordering operator<=>(const BankAccount& other) const = default;

	/**
	 * @brief Represents the name of the associated financial institution for a bank account.
	 *
	 * This member variable serves as an identifier for the financial institution associated
	 * with the bank account. It ensures that each bank account is uniquely tied to its source
	 * institution, enabling clear and consistent usage in a variety of banking operations or
	 * financial reconciliations.
	 *
	 * Use this member to store or retrieve the name of the financial institution when performing
	 * actions that require institution-specific identification, such as reconciliation, reporting,
	 * and interbank communication.
	 */
	BankName name_{};
	/**
	 * @brief Indicates the type and ledger number of bank account.
	 *
	 * This variable specifies whether the account is, for example, a savings account, checking
	 * account, or other types as implemented in the system.
	 */
	BankAccountType accountType_{};
};

/**
 * @brief Full object stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param ba The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_BANKING_EXPORT std::ostream& operator<<(std::ostream& out, const BankAccount& ba);
/**
 * @brief Full object Qt stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param ba The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_BANKING_EXPORT ::QDebug operator<<(::QDebug out, const BankAccount& ba);
/**
 * @brief Generates a localized presentation text for a bank account.
 *
 * This function creates a formatted string to display information about a bank account,
 * including its name and type. It supports localization to adapt the presentation
 * depending on the specified locale preferences. The output is designed for display
 * in GUI elements or logs where an account summary is needed.
 *
 * Use this method to generate a readable and context-appropriate representation of a bank account.
 *
 * @param ba The bank account object for which the presentation text is generated.
 * @param isLocaleAware Indicates whether the output should adhere to locale-aware formatting rules.
 *
 * @return A formatted string representing the bank account, localized if specified.
 */
DRN_BANKING_EXPORT ::QString presentationText(
	const BankAccount& ba,
	const foundation::LocaleAware isLocaleAware = foundation::LocaleAware{true}
);

}

#endif
