/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BANKING_BANK_ACCOUNT_TYPE_H_
#define DRN_BANKING_BANK_ACCOUNT_TYPE_H_

#include <QtGlobal>

#include <array>
#include <compare>
#include <iosfwd>

#include <accounting/AccountNumber.h>
#include <accounting/AccountTypes.h>
#include <banking_export.h>
#include <foundation/PresentationTags.hpp>


class QString;

namespace drn::banking
{

/**
 * @brief The types of financial accounts supported.
 */
enum class DRN_BANKING_EXPORT SupportedAccountTypes
{
	/// The bank account type is not known.
	Unknown,
	/// The bank account type is a for chequeing/current account.
	Chequeing,
	/// The bank account type is a credit card.
	CreditCard,
	/// The bank account type is for a loan.
	Loan,
	/// The bank account type is an interest bearing savings account.
	Savings,
	/// The bank account type is a goal budget item account that is stored as a sub-total inside
	/// another bank account.
	Goal,
};
/**
 * @brief The representation of the bank account ledger and the type of account for a financial
 * institution.
 */
struct DRN_BANKING_EXPORT BankAccountType
{
	std::strong_ordering operator<=>(const BankAccountType& other) const = default;

	/// The account number of the ledger for the bank account.
	accounting::AccountNumber number_{};
	/// The type of bank account for a financial institution.
	SupportedAccountTypes type_{SupportedAccountTypes::Unknown};
};

/**
 * @brief Maps a bank account type string to its corresponding supported account type enumeration.
 * This method processes a provided bank account type string and converts it into a supported
 * account type enumeration value, which can be used for further classification or logic in
 * financial account management.
 *
 * @exception std::invalid_argument Thrown if the provided bank account type string cannot be mapped
 * to a valid supported account type enumeration.
 *
 * @param accountType A textual representation of a bank account type. This can be a localized
 * string matching predefined values for different account types such as "Chequeing", "Savings", or
 * "Loan".
 *
 * @return The enumeration value corresponding to the provided bank account type string.
 */
DRN_BANKING_EXPORT SupportedAccountTypes toSupportedAccountTypes(const ::QString& accountType);
/**
 * @brief Converts the given account type into its non-localized textual representation.
 *
 * This function maps the supported account types to their corresponding non-localized
 * string representations. It is useful when a consistent textual representation of account
 * types is required, irrespective of localization or display formatting preferences.
 *
 * @exception Throws a runtime exception if the provided account type is invalid or unhandled.
 *
 * @param sat The supported account type that is to be converted to a non-localized text.
 *
 * @return A non-localized string representing the given account type. If the account typ is
 * `Unknown`, the string representation will also state "Unknown".
 */
DRN_BANKING_EXPORT ::QString toNonLocalisedText(const SupportedAccountTypes& sat);
/**
 * @brief Converts a supported bank account type to a corresponding ledger account type.
 *
 * This method maps a given supported bank account type to its equivalent ledger account type
 * based on pre-defined mappings. It is useful when translating account types between two different
 * systems or domains, ensuring compatibility and proper categorization of account data.
 *
 * @exception std::invalid_argument Throws this exception if the provided supported account type is
 * invalid or unrecognized.
 *
 * @param sat The supported bank account type to be converted.
 *
 * @return The converted bank account type corresponding to the provided supported ledger account
 * type.
 */
DRN_BANKING_EXPORT accounting::AccountTypes toAccountTypes(const SupportedAccountTypes& sat);
/**
 * @brief Full object stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param sat The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_BANKING_EXPORT std::ostream& operator<<(std::ostream& out, const SupportedAccountTypes& sat);
/**
 * @brief Full object stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param bat The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_BANKING_EXPORT std::ostream& operator<<(std::ostream& out, const BankAccountType& bat);
/**
 * @brief Determines if a supported account type is an asset account type.
 *
 * @param type The supported type to check.
 *
 * @return Gives true when the supported type is of a known asset type, false otherwise.
 */
[[nodiscard]]
DRN_BANKING_EXPORT bool isAssetAccount(const SupportedAccountTypes type);
/**
 * @brief Determines if a supported account type is a liability account type.
 *
 * @param type The supported type to check.
 *
 * @return Gives true when the supported type is of a known liability type, false otherwise.
 */
[[nodiscard]]
DRN_BANKING_EXPORT bool isLiabilityAccount(const SupportedAccountTypes type);
/**
 * @brief Generates a user-friendly localised presentation text for a given supported account type.
 *
 * This method emits a human-readable string corresponding to the financial account type,
 * such as "Chequeing", "Savings", or "Credit Card". It is useful in GUI applications or other
 * user-facing components that need a descriptive label for account types.
 *
 * @exception std::logic_error If the provided account type is not recognized.
 *
 * @param sat The account type for which a descriptive string is required.
 *
 * @return A descriptive label for the specified account type.
 */
DRN_BANKING_EXPORT ::QString presentationText(const SupportedAccountTypes& sat);
/**
 * @brief Generates a user-friendly presentation text for a bank account type.
 *
 * This method creates a textual representation of the bank account type, including account number
 * and type. It is useful in GUI applications or any user-facing component to display a descriptive
 * representation of an account. Localization is applied to the presentation when specified.
 *
 * @exception std::logic_error If the provided bank account type is invalid or unrecognized.
 *
 * @param bat The bank account type for which the presentation text is generated.
 * @param isLocaleAware A flag indicating whether to localize the presentation text or not.
 *
 * @return A localized or non-localized descriptive string representing the bank account type and
 * number.
 */
DRN_BANKING_EXPORT ::QString presentationText(
	const BankAccountType& bat,
	const foundation::LocaleAware isLocaleAware = foundation::LocaleAware{true}
);

/**
 * @brief A collection of all supported financial account types.
 *
 * This array contains all the predefined account types supported by the system. It provides
 * a centralized list that can be used for validations, comparisons, or any functionality that
 * requires iterating over the supported account types. It is especially useful to avoid redundancy
 * and ensure consistency when handling account types in business logic or graphical interfaces.
 *
 * Use this array when you need to check or iterate over all types of accounts supported in the
 * system.
 *
 * Example usage:
 * @code
 * for (const auto& accountType : allSupportedAccountTypes) {
 *     // Perform operations based on the accountType
 * }
 * @endcode
 */
inline constexpr std::array allSupportedAccountTypes{
	SupportedAccountTypes::Unknown,
	SupportedAccountTypes::Chequeing,
	SupportedAccountTypes::Goal,
	SupportedAccountTypes::CreditCard,
	SupportedAccountTypes::Loan,
	SupportedAccountTypes::Savings
};

}

Q_DECLARE_METATYPE(drn::banking::SupportedAccountTypes)

#endif
