/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BANKING_BANK_NAME_H_
#define DRN_BANKING_BANK_NAME_H_

#include <compare>
#include <iosfwd>

#include <QDebug>
#include <QString>

#include <banking_export.h>


namespace drn::banking
{

/**
 * @brief A representation of the name of a financial institution.
 */
class DRN_BANKING_EXPORT BankName
{
	::QString value_;

public:
	/**
	 * @brief Constructs a default-initialized object representing the name of a financial
	 * institution.
	 *
	 * Use this constructor when you need to initialize a financial institution's name
	 * without immediately assigning a specific value. It sets all internal state to its default.
	 */
	BankName() = default;
	/**
	 * @brief Constructs an object representing the name of a financial institution.
	 *
	 * This constructor should be used when you need to instantiate an object representing
	 * the name of a financial institution, with a specific value provided during construction.
	 *
	 * @exception BankError Thrown when the provided name is empty or contains only whitespace.
	 *
	 * @param name The name of the financial institution to be initialized.
	 */
	explicit BankName(::QString name);
	/**
	 * @brief An accessor to the value for the name of the financial institution.
	 */
	[[nodiscard]]
	const ::QString& institution() const noexcept;

	std::strong_ordering operator<=>(const BankName&) const noexcept = default;
	/**
	 * @brief Determines if the financial institution's name value is set.
	 *
	 * @return True when the name value is not set, false otherwise.
	 */
	[[nodiscard]]
	bool isEmpty() const noexcept;
};

/**
 * @brief Full object stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param bn The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_BANKING_EXPORT std::ostream& operator<<(std::ostream& out, const BankName& bn);
/**
 * @brief Full object Qt stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param bn The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_BANKING_EXPORT ::QDebug operator<<(::QDebug out, const BankName& bn);
/**
 * @brief Generates a user-friendly presentation text for a financial institution's name.
 *
 * This method retrieves the name of the financial institution
 * and formats it for presentation through a graphical or textual user interface.
 *
 * @param bn The object representing the financial institution whose name is to be presented.
 *
 * @return A formatted string representation of the financial institution's name, suitable for
 * presentation.
 */
DRN_BANKING_EXPORT ::QString presentationText(const BankName& bn);

}

#endif
