/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BANKING_RECONCILED_BANK_ACCOUNT_H_
#define DRN_BANKING_RECONCILED_BANK_ACCOUNT_H_

#include <pecunia/Money.hpp>

#include <QDate>

#include <compare>
#include <iosfwd>
#include <optional>

#include <banking_export.h>
#include "BankAccount.h"


namespace pecunia
{

enum class Currency : quint16;

}

namespace drn::accounting
{

class AccountNumber;

}
namespace drn::banking
{

/**
 * @brief Represents a bank account that has undergone a reconciliation process to ensure that its
 * transactions are consistent with bank statements.
 *
 * This class is designed to manage and reflect accurate financial records by verifying and
 * reconciling transactions against external bank statements. It is particularly useful for
 * financial reporting, auditing, and ensuring the integrity of account balances.
 *
 * To use this class, provide the necessary details of the bank account and initiate reconciliation
 * through the provided methods. Once reconciled, you can access the updated account state, balance,
 * and any identified discrepancies.
 */
class DRN_BANKING_EXPORT ReconciledBankAccount
{
	BankAccount bankAccount_;
	pecunia::Money balance_;
	std::optional<::QDate> reconciledOn_;

public:
	/**
	 * @brief Constructs an empty instance of a reconciled bank account, representing a default
	 * state for an unreconciled account.
	 *
	 * This constructor initializes the reconciled bank account with default values, ensuring that
	 * no unnecessary data is loaded or manipulated. The default-constructed account allows you to
	 * safely proceed with further processes without pre-existing account data.
	 */
	ReconciledBankAccount();
	/**
	 * @brief Minimum value constructor.
	 *
	 * Initializes a bank account that is prepared for reconciliation by associating it with a
	 * specific currency. This constructor streamlines the setup process necessary to perform
	 * reconciliations by creating a baseline using the provided bank account details and a
	 * currency framework.
	 *
	 * Use this constructor when you need to create a reconciled bank account efficiently with only
	 * the essential details like account and currency. Post-instantiation, you can proceed to
	 * manage reconciliations and examine the associated financial data.
	 *
	 * @param bankAccount The initial bank account to be reconciled.
	 * @param currency The currency to associate with the bank account, ensuring all balances are
	 * denominated consistently.
	 */
	ReconciledBankAccount(BankAccount bankAccount, const pecunia::Currency& currency);
	/**
	 * @brief Full initialisation constructor.
	 *
	 * This constructor initializes a bank account with its associated balance and an optional
	 * reconciliation date. It is particularly useful when you have precise information about
	 * an account's reconciled state and need to immediately create a usable object that reflects
	 * these details.
	 *
	 * @param bankAccount The bank account being reconciled.
	 * @param balance The reconciled balance associated with the bank account.
	 * @param reconciledOn An optional date representing when the reconciliation occurred.
	 */
	ReconciledBankAccount(
		BankAccount bankAccount,
		pecunia::Money balance,
		std::optional<::QDate> reconciledOn
	);
	/**
	 * @brief Copy constructor.
	 *
	 * This constructor creates a new instance as an exact copy of an existing one. It ensures that
	 * all internal data of the original object is preserved in the copy.
	 *
	 * @param other The object to copy from.
	 */
	ReconciledBankAccount(const ReconciledBankAccount& other) noexcept;
	/**
	 * @brief Move constructor, transferring ownership of all state from another instance without
	 * duplicating data.
	 *
	 * This constructor allows efficient transfer of data from one  instance to another, avoiding
	 * the overhead of copying. Use this when you need to transfer an already initialized reconciled
	 * bank account object into a new context without reinitializing or duplicating its contents.
	 *
	 * @param other The instance whose state will be transferred. It will leave the source instance
	 * in a valid, but unspecified, state after the operation.
	 */
	ReconciledBankAccount(ReconciledBankAccount&& other) noexcept;
	/**
	 * @brief Cleans up resources associated with a reconciled bank account instance.
	 */
	~ReconciledBankAccount() noexcept;
	/**
	 * @brief Copy assignment operator.
	 *
	 * @param other The object instance to copy its data from.
	 *
	 * @post The current instnace's data is all replaced with the data from the supplied object.
	 *
	 * @return The current instance after modification.
	 */
	ReconciledBankAccount& operator=(const ReconciledBankAccount& other) noexcept;
	/**
	 * @brief Move assignment operator.
	 *
	 * @param other  The instance whose state will be transferred. It will leave the source instance
	 * in a valid, but unspecified, state after the operation.
	 *
	 * @post The current instnace's data is all replaced with the data from the supplied object.
	 *
	 * @return The current instance after modification.
	 */
	ReconciledBankAccount& operator=(ReconciledBankAccount&& other) noexcept;
	/**
	 * @brief An accessor to the bank account that is reconciled.
	 */
	[[nodiscard]]
	const BankAccount& bankAccount() const noexcept;
	/**
	 * @brief An accessor to the amount of money in the bank account on the date reconciled.
	 *
	 * @return When the bank account was reconciled, gives the reconciled balance, zero when the
	 * bank account has yet to be reconciled.
	 */
	[[nodiscard]]
	const pecunia::Money& balance() const noexcept;
	/**
	 * @brief An accessor to the date the bank account was reconciled on.
	 *
	 * @return When the bank account was reconciled, gives the date, null when the bank account has
	 * yet to be reconciled.
	 */
	[[nodiscard]]
	const std::optional<::QDate>& reconciledOn() const noexcept;
	/**
	 * @brief Updates the reconciled information.
	 *
	 * @exception BankError When the date for reconciliation is in the past of the current
	 * reconciliation.
	 *
	 * @param reconciledOn The date for the new balance after a successful reconciliation.
	 * @param balance The new balance after reconciliation.
	 */
	void reconcile(const ::QDate& reconciledOn, const pecunia::Money& balance);

	std::strong_ordering operator<=>(const ReconciledBankAccount& other) const = default;
};

/**
 * @brief Full object stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param rba The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_BANKING_EXPORT std::ostream& operator<<(std::ostream& out, const ReconciledBankAccount& rba);
/**
 * @brief Full object Qt stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param rba The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_BANKING_EXPORT ::QDebug operator<<(::QDebug out, const ReconciledBankAccount& rba);

}

#endif
