/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BUDGETING_BUDGET_H_
#define DRN_BUDGETING_BUDGET_H_

#include <pecunia/Money.hpp>

#include <compare>
#include <iosfwd>
#include <iterator>
#include <map>
#include <set>
#include <typeindex>
#include <utility>

#include <budgeting_export.h>
#include <foundation/ObserverPtr.hpp>

#include "BillMap.hpp"
#include "BudgetItemIdentifier.h"
#include "DebtMap.h"
#include "GoalMap.h"
#include "NontrackMap.hpp"
#include "WageMap.hpp"


namespace drn::budgeting
{

enum class BudgetItemTypes : quint8;
class BudgetedMoney;
class BudgetSource;

/**
 * @brief Represents a budget management system, allowing creation, modification, and analysis of
 * budgets.
 *
 * All methods respect the structure of the budget, maintaining the integrity of categories,
 * expenses, and incomes.
 */
class DRN_BUDGETING_EXPORT Budget
{
public:
	/**
	 * @brief Provides functionality to traverse through a collection of elements sequentially,
	 * allowing controlled access to each element in the collection.
	 *
	 * The iterator ensures encapsulation by abstracting the traversal mechanism, making it
	 * independent of the collection's internal structure. It can be utilised to iterate through
	 * various types of collections, such as lists, arrays, or other custom data structures.
	 *
	 * The class provides essential operations for iteration, such as advancing to the next element,
	 * checking for the end of the collection, and accessing the current element.
	 */
	class DRN_BUDGETING_EXPORT Iterator
	{
		foundation::ObserverPtr<const Budget> budget_;
		BillMap::const_iterator billIter_;
		DebtMap::const_iterator debtIter_;
		GoalMap::const_iterator goalIter_;
		NontrackMap::const_iterator nontrackIter_;
		WageMap::const_iterator wageIter_;

	public:
		/// Standard iterator category tag.
		using iterator_category = std::forward_iterator_tag;
		/// Type used to represent the distance between two iterators
		using difference_type = std::ptrdiff_t;
		/// The type of value that the iterator points to, representing a the budget item ID and
		/// its monetary budget item.
		using value_type = std::pair<
			BudgetItemIdentifier,
			foundation::ObserverPtr<const BudgetedMoney>
		>;
		/// The type used when the value is to be a pointer to the value.
		using const_pointer = const value_type*;
		/// The type used when the value is to be a const-ref to the value.
		using const_reference = const value_type&;

		/**
		 * @brief Fully initialises the object to point to the budget. Upon successful construction
		 * will be pointing just past the end of the budget.
		 *
		 * @param budget The budget to iterate over.
		 */
		explicit Iterator(const Budget& budget);
		/**
		 * @brief Equality operator with the same semantics: both being at the end or both observe
		 * the same budgt item is considered to be equivalent.
		 *
		 * @pre The other iterator must come from the same budget object.
		 *
		 * @param other The other object to compare against.
		 *
		 * @return True when the two iterators point to the same object, false otherwise.
		 */
		[[nodiscard]]
		bool operator==(const Iterator& other) const noexcept;
		/**
		 * @brief Star-operator to the current budget item.
		 *
		 * @return The current budget item.
		 */
		[[nodiscard]]
		const_reference operator*() const noexcept;
		/**
		 * @brief Arrow-operator to the current budget item.
		 *
		 * @return The current budget item.
		 */
		[[nodiscard]]
		const_pointer operator->() const noexcept;
		/**
		 * @brief Prefix increment operator.
		 *
		 * @return The object after the increment operation.
		 */
		Iterator& operator++();
		/**
		 * @brief Postfix increment operator.
		 *
		 * @return The object before the increment operation.
		 */
		Iterator operator++(int);
		/**
		 * @brief Returns an iterator pointing to the first item in the collection within the
		 * budget.
		 *
		 * This method initializes an iterator to start iterating through the available budget
		 * items, starting with the first category containing items. If no items are present in any
		 * category, the iterator will be initialized to its final state, indicating the end of the
		 * collection.
		 *
		 * @return An iterator initialized to the first budget item or indicating the end of the
		 * collection if no items are available.
		 */
		[[nodiscard]]
		Iterator begin() const noexcept;
		/**
		 * @brief Returns an iterator pointing to the position after the last element in the
		 * budget's managed collection.
		 *
		 * This method allows iteration over the budget's items to stop gracefully. By calling this
		 * method, the iterator returned can be used to check termination of a loop or other forms
		 * of traversal.
		 *
		 * @return An iterator positioned at the end of the budget's collection, signifying
		 * traversal completion.
		 */
		[[nodiscard]]
		Iterator end() const noexcept;

	private:
		static const value_type endValue_;

		value_type current_;
	};

	Budget() = default;
	/**
	 * @brief Full initialisation constructor.
	 *
	 * It is designed to create a consolidated representation of a budget by accepting various
	 * components that contribute to financial planning.
	 *
	 * @param wages A collection representing the income sources for the budget.
	 * @param bills A collection representing recurring expenses within the budget.
	 * @param debts A collection representing obligations to be repaid over time.
	 * @param goals A collection representing savings or investment targets within the budget.
	 * @param nontracks A collection representing items that are not tracked, typically incidental
	 * or minor expenses.
	 */
	explicit Budget(
		WageMap wages,
		BillMap bills = {},
		DebtMap debts = {},
		GoalMap goals = {},
		NontrackMap nontracks = {}
	);
	/**
	 * @brief Three-way comparison operator that performs the check in the order: wages, bills,
	 * debts, goals, and then nontracks.
	 *
	 * @param other The other object to compare against.
	 *
	 * @return The ordering relationship between them.
	 */
	[[nodiscard]]
	std::strong_ordering operator<=>(const Budget& other) const noexcept;

	[[nodiscard]]
	bool operator==(const Budget&) const noexcept = default;
	/**
	 * @brief Determines if the budget is empty.
	 *
	 * @return When the budget doesn't contain any wages, bills, debts, nor goals gives
	 * true, false else-wise
	 */
	[[nodiscard]]
	bool isEmpty() const noexcept;
	/**
	 * @brief Calculates the size of all the budget items in the current budget.
	 *
	 * @return The number of items in the budget.
	 */
	[[nodiscard]]
	std::size_t size() const noexcept;
	/**
	 * @brief Provides all the sources in the budget.
	 *
	 * @return The container mapping the type to a container of budget sources in the budget.
	 */
	[[nodiscard]]
	std::map<std::type_index, std::set<BudgetSource>> sources() const noexcept;
	/**
	 * @brief Returns an iterator pointing to the beginning of the budget's underlying structure.
	 *
	 * This method enables iteration through the budget data. It can be used to traverse and access
	 * elements starting from the first position in a sequential manner.
	 *
	 * @return An iterator positioned at the beginning of the budget's underlying structure.
	 */
	[[nodiscard]]
	Iterator begin() const noexcept;
	/**
	 * @brief Returns an iterator pointing to the end of the budget structure.
	 *
	 * This method provides a means to signify the end boundary of the budget structure when
	 * performing iteration. It is commonly used in conjunction with other methods or algorithms
	 * that traverse or manipulate budget data.
	 *
	 * @return An iterator pointing to the end boundary of the budget structure.
	 */
	[[nodiscard]]
	Iterator end() const noexcept;

	/// The container of all the wage coming into the budget.
	WageMap wages_;
	/// The container of all the living bills leaving the budget.
	BillMap bills_;
	/// The container of all the borrowed money leaving the budget.
	DebtMap debts_;
	/// The container of all the money being saved in the budget.
	GoalMap goals_;
	/// The container of all the money that is not tracked in the budget.
	NontrackMap nontracks_;
};
/**
 * @brief Calculates the totals for all budgetary items for a given period.
 */
struct DRN_BUDGETING_EXPORT BudgetTotals
{
	/**
	 * @brief Constructor for calculating the totals.
	 *
	 * @param budget The budget from which all totals come from.
	 * @param currency The currency the totals should be calculated in.
	 * @param period The frequency for which the total should be calculated in.
	 */
	BudgetTotals(
		const Budget& budget,
		const pecunia::Currency& currency,
		const EventFrequency period
	);

	/// The total of all the wage for the requested period.
	const pecunia::Money wage_;
	/// The total of all the goals for the requested period.
	const pecunia::Money goals_;
	/// The total of all the bills for the requested period.
	const pecunia::Money bills_;
	/// The total of all the debts for the requested period.
	const pecunia::Money debts_;
	/// The total of all the nontrack spending for the requested period.
	const pecunia::Money nontrack_;
	/// The frequency period all totals are calculated to.
	const EventFrequency period_;
	/// The total of money remaining after all the incoming and outgoing monies are calculated for
	/// the period.
	const pecunia::Money remaining_;
};

/**
 * @brief Finds all the unique currencies used throughout the budget.
 *
 * @param b The budget to examine for currencies.
 *
 * @return Only the unique currencies in the budget.
 */
[[nodiscard]]
DRN_BUDGETING_EXPORT std::set<pecunia::Currency> containedCurrencies(const Budget& b);
/**
 * @brief Full object stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param b The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_BUDGETING_EXPORT std::ostream& operator<<(std::ostream& out, const Budget& b);
/**
 * @brief Searches for a specific budget item within a budget using a given source as a reference.
 *
 * @tparam BudgetItemType The type of budget item to search for; should be derived from a base
 * class representing general budgeted components.
 *
 * @param needle A specific budget source to locate the corresponding budget item for.
 * @param haystack A collection representing the budget in which the search will be performed.
 *
 * @return The budget item corresponding to the given source if found, otherwise an empty optional.
 */
template<typename BudgetItemType>
[[nodiscard]]
std::optional<BudgetItemType> findBudgetItem(
	const BudgetSource& needle,
	const Budget& haystack
);
/**
 * @brief Searches for a specific budget item within a budget using a given source as a reference.
 *
 * @param needle A specific budget source to locate the corresponding budget item for.
 * @param haystack A collection representing the budget in which the search will be performed.
 *
 * @return The budget item corresponding to the given source if found, otherwise an empty optional.
 */
template<>
[[nodiscard]]
DRN_BUDGETING_EXPORT std::optional<Bill> findBudgetItem(
	const BudgetSource& needle,
	const Budget& haystack
);
/**
 * @brief Searches for a specific budget item within a budget using a given source as a reference.
 *
 * @param needle A specific budget source to locate the corresponding budget item for.
 * @param haystack A collection representing the budget in which the search will be performed.
 *
 * @return The budget item corresponding to the given source if found, otherwise an empty optional.
 */
template<>
[[nodiscard]]
DRN_BUDGETING_EXPORT std::optional<Debt> findBudgetItem(
	const BudgetSource& needle,
	const Budget& haystack
);
/**
 * @brief Searches for a specific budget item within a budget using a given source as a reference.
 *
 * @param needle A specific budget source to locate the corresponding budget item for.
 * @param haystack A collection representing the budget in which the search will be performed.
 *
 * @return The budget item corresponding to the given source if found, otherwise an empty optional.
 */
template<>
[[nodiscard]]
DRN_BUDGETING_EXPORT std::optional<Goal> findBudgetItem(
	const BudgetSource& needle,
	const Budget& haystack
);
/**
 * @brief Searches for a specific budget item within a budget using a given source as a reference.
 *
 * @param needle A specific budget source to locate the corresponding budget item for.
 * @param haystack A collection representing the budget in which the search will be performed.
 *
 * @return The budget item corresponding to the given source if found, otherwise an empty optional.
 */
template<>
[[nodiscard]]
DRN_BUDGETING_EXPORT std::optional<Nontrack> findBudgetItem(
	const BudgetSource& needle,
	const Budget& haystack
);
/**
 * @brief Searches for a specific budget item within a budget using a given source as a reference.
 *
 * @param needle A specific budget source to locate the corresponding budget item for.
 * @param haystack A collection representing the budget in which the search will be performed.
 *
 * @return The budget item corresponding to the given source if found, otherwise an empty optional.
 */
template<>
[[nodiscard]]
DRN_BUDGETING_EXPORT std::optional<Wage> findBudgetItem(
	const BudgetSource& needle,
	const Budget& haystack
);
/**
 * @brief Locates a budget item within a budget using its identifier.
 *
 * This function searches through various categories of a budget to find and return the budget item
 * that matches the given identifier.
 *
 * @exception BudgetItemError Thrown when the provided identifier does not match any budget item
 * within the budget, with a message indicating that the source does not exist.
 *
 * @param id The unique identifier of the budget item to be located.
 * @param b The budget from which the item should be retrieved.
 *
 * @return The budgeted item that matches the identifier.
 */
[[nodiscard]]
DRN_BUDGETING_EXPORT const BudgetedMoney& findBudgetItem(
	const BudgetItemIdentifier& id,
	const Budget& b
);

}

#endif
