/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BUDGETING_BUDGET_ITEM_IDENTIFIER_H_
#define DRN_BUDGETING_BUDGET_ITEM_IDENTIFIER_H_

#include <iosfwd>

#include <compare>
#include <budgeting_export.h>

#include "BudgetItemTypes.h"
#include "BudgetSource.h"


class QDebug;

namespace drn::budgeting
{

/**
 * @brief Identifies a budget item by its type and source. The structure is used to uniquely
 * represent a budget item by combining its type (e.g., bill, debt, wage) and its source
 * information.
 */
struct DRN_BUDGETING_EXPORT BudgetItemIdentifier final
{
	[[nodiscard]]
	std::strong_ordering operator<=>(const BudgetItemIdentifier& rhs) const noexcept = default;

	/// Represents the type of a budget item.
	BudgetItemTypes type_{BudgetItemTypes::Unknown};
	/// Represents the source information of a budget item.
	BudgetSource source_{};
};

/**
 * @brief Full object standard output stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_BUDGETING_EXPORT std::ostream& operator<<(std::ostream& out, const BudgetItemIdentifier& data);
/**
 * @brief Full object Qt stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_BUDGETING_EXPORT ::QDebug operator<<(::QDebug out, const BudgetItemIdentifier& data);
/**
 * @brief Generates a text for the identifier of a budget item. The text is formatted to concisely
 * display the source’s name and a description of the item’s type. This is often used for displaying
 * budget item details in textual format in GUI elements or other presentation layers.
 *
 * @param bii The budget item identifier to format.
 *
 * @return The formatted presentation text.
 */
DRN_BUDGETING_EXPORT ::QString presentationText(const BudgetItemIdentifier& bii);

}

#endif
