/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BUDGETING_BUDGET_ITEM_TYPES_H_
#define DRN_BUDGETING_BUDGET_ITEM_TYPES_H_

#include <QtGlobal>

#include <array>
#include <iosfwd>
#include <typeindex>

#include <budgeting_export.h>


class QString;

namespace drn::budgeting
{

class BudgetedMoney;

/**
 * @brief Represents different types of budget-related items for categorization purposes. This
 * defines a variety of budgeting categories, providing a structured way to classify items.
 */
enum class DRN_BUDGETING_EXPORT BudgetItemTypes : quint8
{
	/// Represents an undefined or unspecified budget item type.
	Unknown,
	/// Represents periodic payment obligations, such as utility or service bills.
	Bill,
	/// Represents financial liabilities, such as loans or credit balances.
	Debt,
	/// Represents personal or financial saving objectives.
	Goal,
	/// Represents items that should not be tracked as part of the budget.
	Nontrack,
	/// Represents income sources, such as salaries or earnings.
	Wage,
};

/**
 * @brief Stream insertion operator for logging purposes.
 *
 * @param out The stream to log the values into.
 * @param bit The information to be logged.
 *
 * @return The stream after it is populated with the logging information.
 */
DRN_BUDGETING_EXPORT std::ostream& operator<<(std::ostream& out, const BudgetItemTypes& bit);
/**
 * @brief Translates a value into a weakly-typed English string representation.
 *
 * @param bit The value to be converted into its English string equivalent.
 *
 * @return An English string representation of the given value.
 */
[[nodiscard]]
DRN_BUDGETING_EXPORT ::QString toWeak(const BudgetItemTypes& bit);
/**
 * @brief Provides a user-friendly, localized text representation of a budget item type. This
 * function takes a budget item type and converts it into a meaningful, human-readable string. The
 * resulting text is typically localized to ensure it is intuitive to the end user in their
 * preferred language. It is particularly useful for displaying budget item types in a user
 * interface or report.
 *
 * @param bit The budget item type to be converted into a localized string representation.
 *
 * @return A localized string representing the given budget item type.
 */
[[nodiscard]]
DRN_BUDGETING_EXPORT ::QString presentationText(const BudgetItemTypes& bit);
/**
 * @brief Converts a raw string into its corresponding enumeration value.
 *
 * @exception std::logic_error When the supplied string does not match any known category.
 *
 * @param raw The raw string representing a budget category, which may originate from user input,
 * external sources, or localized strings.
 *
 * @return The corresponding value based on the input string.
 */
[[nodiscard]]
DRN_BUDGETING_EXPORT BudgetItemTypes toBudgetItemTypes(const ::QString& raw);
/**
 * @brief Converts a type index to its corresponding budget item type.
 *
 * @note N.B. the value BudgetItemTypes::Unknown is explicitly not handled as it cannot be a budget
 * item type.
 *
 * @exception std::logic_error When the supplied index does not match any known category.
 *
 * @param type The type index of the object or category to be mapped.
 *
 * @return The corresponding budget item type for the specified type index.
 */
[[nodiscard]]
DRN_BUDGETING_EXPORT BudgetItemTypes toBudgetItemTypes(const std::type_index& type);
/**
 * @brief Converts a monetary budget item into its corresponding budget item type.
 *
 * This function calculates and returns the budget item type based on the input monetary value.
 * It is designed to help categorize budgeted monetary items into meaningful categories for further
 * processing or display in a system. This ensures consistency and accuracy in the classification of
 * budgeted items.
 *
 * @param money The monetary value to be categorized into a budget item type. This input represents
 * a budgeted amount or entity.
 *
 * @return The budget item type corresponding to the provided monetary value.
 */
[[nodiscard]]
DRN_BUDGETING_EXPORT BudgetItemTypes toBudgetItemTypes(const BudgetedMoney& money) noexcept;

/// Provides a complete list of available budget item types.
inline constexpr std::array allBudgetItemTypes{
	BudgetItemTypes::Unknown,
	BudgetItemTypes::Bill,
	BudgetItemTypes::Debt,
	BudgetItemTypes::Goal,
	BudgetItemTypes::Nontrack,
	BudgetItemTypes::Wage
};

}

#endif
