/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BUDGETING_BUDGET_SOURCE_H_
#define DRN_BUDGETING_BUDGET_SOURCE_H_

#include <QString>

#include <compare>
#include <iosfwd>

#include <budgeting_export.h>


class QDebug;

namespace drn::budgeting
{

/**
 * @brief The manager for the information on the source of a budgetary item.
 */
class DRN_BUDGETING_EXPORT BudgetSource
{
	::QString name_;

public:
	BudgetSource() = default;
	/**
	 * @brief Fully initialising constructor. All leading and trailing white-space is removed
	 * from the name of the source.
	 *
	 * @exception Error When the name of the budgetary item is not valid.
	 *
	 * @param name The name of the source for the budgetary item.
	 */
	explicit BudgetSource(::QString name);

	[[nodiscard]]
	std::strong_ordering operator<=>(const BudgetSource& other) const noexcept = default;
	/**
	 * @brief The name of the source.
	 */
	[[nodiscard]]
	const ::QString& name() const noexcept;
	/**
	 * @brief Changes the name of the budgetary item source. All leading and trailing white-space
	 * is removed from the name.
	 *
	 * @exception Error When the name of the budgetary item is not valid.
	 *
	 * @param name The new name for the budgetary item source.
	 */
	void setName(const ::QString& name);
	/**
	 * @brief Determines if the object is empty, e.g. constructed by default.
	 *
	 * @return When everything is the default value gives true, false else-wise.
	 */
	[[nodiscard]]
	bool isEmpty() const noexcept;
};

/**
 * @brief Full object standard stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_BUDGETING_EXPORT std::ostream& operator<<(std::ostream& out, const BudgetSource& data);
/**
 * @brief Full object Qt stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_BUDGETING_EXPORT ::QDebug operator<<(::QDebug out, const BudgetSource& data);
/**
 * @brief Retrieves the textual representation to be displayed on a GUI element or used elsewhere.
 *
 * This method provides the presentation-ready text that can be displayed in a GUI item or
 * utilised programmatically when a user facing textual description is required.
 *
 * @return A string containing the presentation-ready text for display or further processing.
 */
DRN_BUDGETING_EXPORT ::QString presentationText(const BudgetSource& bs);

}

#endif
