/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BUDGETING_BUDGETING_ERRORS_H_
#define DRN_BUDGETING_BUDGETING_ERRORS_H_

#include <exception>

#include <foundation/Error.h>
#include <budgeting_export.h>



namespace drn::budgeting
{

class BudgetSource;

/**
 * @brief Represents an error while working with a budget item.
 */
struct DRN_BUDGETING_EXPORT BudgetItemError final : foundation::Error
{
	/**
	 * @brief Constructs a new error with a specific source and error message.
	 * This can be used when an operation related to a budget item fails and needs clear
	 * identification of the source and issue.
	 *
	 * Use this to provide detailed error information while performing operations on budget items.
	 * Typically triggered when invalid inputs, state violations, or logical inconsistencies occur
	 * during budget-related operations.
	 *
	 * @param bs The source of the budget item involved in the error. This is required to identify
	 * the particular budget item that caused the issue.
	 * @param message A descriptive message specifying the details of the error. This is essential
	 * to convey meaningful information about the issue.
	 */
	BudgetItemError(const BudgetSource& bs, const ::QString& message);
	BudgetItemError(const BudgetSource&, const char* message) = delete;
	/**
	 * @brief Constructs a new error object for issues related to a budget item. This constructor
	 * emits an error description, identifying the budget item causing the problem and providing a
	 * detailed explanation along with the root exception.
	 *
	 * This should be used to report errors that occur during budget item operations, such as
	 * invalid manipulations, state inconsistencies, or failures in processing. By using this
	 * constructor, you ensure clear communication of the issue and its origin.
	 *
	 * @param bs The source of the budget item causing the error. This provides context
	 * for identifying the problematic budget item.
	 * @param message A descriptive message explaining the error encountered during
	 * the operation.
	 * @param root The root exception that triggered this error. This allows chaining
	 * of exceptions to provide a clear traceback of the issue.
	 */
	BudgetItemError(const BudgetSource& bs, const ::QString& message, const std::exception& root);
	BudgetItemError(const BudgetSource&, const char* message, const std::exception) = delete;
};

}

#endif
