/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BUDGETING_DEBT_H_
#define DRN_BUDGETING_DEBT_H_

#include <compare>
#include <iosfwd>

#include <budgeting_export.h>
#include <foundation/Percentage.h>

#include "BudgetedMoney.h"


namespace drn::budgeting
{

/**
 * @brief The representation of an amount of money that is borrowed and how often it occurs.
 */
class DRN_BUDGETING_EXPORT Debt final : public BudgetedMoney
{
	pecunia::Money borrowed_;
	foundation::Percentage interest_;

public:
	/**
	 * @brief Constructs all member variables to their empty state.
	 */
	Debt();
	explicit Debt(const BudgetSource& source);
	/**
	 * @brief Fully initialises all member variables to their supplied values.
	 *
	 * @exception BudgetItemError When the source is empty.
	 * @exception BudgetItemError When the borrowed is less then zero.
	 * @exception BudgetItemError When the borrowed is not issued in a known currency.
	 * @exception BudgetItemError When the borrowed is less then the minimum payment.
	 * @exception BudgetItemError When the period is not known.
	 * @exception BudgetItemError When the minimum is less than zero.
	 * @exception BudgetItemError When the minimum is not issued in a known currency.
	 * @exception BudgetItemError When the interest rate is less than zero.
	 * @exception BudgetItemError When the next occur date is invalid.
	 *
	 * @param source The source of the debt.
	 * @param minimum The minimum payment required per payment period.
	 * @param period The frequency period of payment for the debt source.
	 * @param nextOccurOn The date the money budget item last occurred on.
	 * @param interest The interest rate applied to the borrowed amount.
	 * @param borrowed The total amount of money borrowed from the debt source.
	 */
	Debt(
		const BudgetSource& source,
		const pecunia::Money& minimum,
		const EventFrequency& period,
		const ::QDate& nextOccurOn,
		const pecunia::Money& borrowed,
		const foundation::Percentage& interest
	);
	[[nodiscard]]
	std::strong_ordering operator<=>(const Debt& other) const noexcept;
	[[nodiscard]]
	bool operator==(const Debt& other) const noexcept = default;
	/**
	 * @brief The total amount of money borrowed.
	 */
	const pecunia::Money& borrowed() const;
	/**
	 * @brief The current interest rate for the debt.
	 */
	const foundation::Percentage& interest() const;
};

DRN_BUDGETING_EXPORT std::ostream& operator<<(std::ostream& o, const Debt& d);

}

#endif
