/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BUDGETING_EVENT_FREQUENCY_H_
#define DRN_BUDGETING_EVENT_FREQUENCY_H_

#include <QtGlobal>

#include <array>
#include <iosfwd>

#include <budgeting_export.h>


class QDate;
class QDebug;
class QDateTime;
class QString;

namespace drn::foundation
{

struct PartialDays;

}
namespace drn::budgeting
{

/**
 * @brief The available frequencies at which an event may regularly occur. They are in ascending
 * order of time.
 */
enum class DRN_BUDGETING_EXPORT EventFrequency : quint8
{
	/// It is not known when the event will reoccur.
	Unknown,
	/// The event will happen every day, i.e. every 24 hours.
	Daily,
	/// The event will happen roughly twice a week, i.e. every 84 hours (3.5 days).
	SemiWeekly,
	/// The event will happen once a week, i.e. every 168 hours (7 days).
	Weekly,
	/// The event will happen every two weeks, i.e. every 336 hours (14 days).
	BiWeekly,
	/// The event will happen twice a month, i.e. every 360 hours (15 days).
	SemiMonthly,
	/// The event will happen once a month.
	Monthly,
	/// The event will happen every two months, i.e. six times a year.
	Bimonthly,
	/// The event will happen every three months, i.e. four times a year.
	Quarterly,
	/// The event will happen every four months, i.e. three times a year.
	Quadrimesterly,
	/// The event will happen every six months, i.e. two times a year.
	Semiannually,
	/// The event will happen once a year.
	Annually,
	/// The event will happen every two years.
	Biennially,
	/// The event will happen every three years.
	Triennially,
	/// The event will happen every four years.
	Quadrennially,
	/// The event will happen every five years.
	Quinquennially,
	/// The event will happen every six years.
	Sexennially,
	/// The event will happen every seven years.
	Septennially,
	/// The event will happen every eight years.
	Octennially,
	/// The event will happen every nine years.
	Novennially,
	/// The event will happen every ten years.
	Decennially,
	/// The event will never happen.
	Never,
};

/**
 * @brief Converts a strong typed enumeration to its corresponding weak type.
 *
 * @exception logic_error When the supplied value cannot be converted.
 *
 * @param strong The strong typed enumeration value to be converted.
 *
 * @return The representing weak type equivalent.
 */
DRN_BUDGETING_EXPORT quint8 toWeak(const EventFrequency strong);
/**
 * @brief Converts an integer representation of an event frequency to its corresponding enumeration
 * value.
 *
 * @exception logic_error When the supplied value cannot be converted.
 *
 * @param weak An integer that represents the event frequency.
 *
 * @return The representing enumeration type equivalent.
 */
DRN_BUDGETING_EXPORT EventFrequency toEventFrequency(const quint8 weak);
/**
 * @brief Converts a string representation of an event frequency to its corresponding enumeration
 * value.
 *
 * @exception logic_error When the supplied value cannot be converted.
 *
 * @param weak A string that represents the event frequency.
 *
 * @return The representing enumeration type equivalent.
 */
DRN_BUDGETING_EXPORT EventFrequency toEventFrequency(const ::QString& weak);
/**
 * @brief Standard stream insertion operator.
 *
 * @param out The output stream where the value will be inserted.
 * @param ef The value to be inserted into the stream.
 *
 * @post The output buffer contains the contents: Value (ef)
 *
 * @return A reference to the provided output stream.
 */
DRN_BUDGETING_EXPORT std::ostream& operator<<(std::ostream& out, const EventFrequency ef);
/**
 * @brief Qt Debug stream insertion operator.
 *
 * @param out The output stream where the value will be inserted.
 * @param ef The value to be inserted into the stream.
 *
 * @post The output buffer contains the contents: Value (ef)
 *
 * @return A reference to the provided output stream.
 */
DRN_BUDGETING_EXPORT ::QDebug operator<<(::QDebug out, const EventFrequency ef);
/**
 * @brief Provides a textual representation, human-readable format of an event frequency.
 *
 * @param ef The event frequency to convert.
 *
 * @return The textual description of the event frequency.
 */
DRN_BUDGETING_EXPORT ::QString presentationText(const EventFrequency& ef);
/**
 * @brief Calculates the next time point from a given point using a specific event frequency.
 *
 * @exception Error When the event frequency is not known.
 * @exception Error When the current time point is not valid.
 *
 * @param current The current time point to calculate from.
 * @param ef The event frequency to calculate the next time point for.
 *
 * @return The calculated time point. If the event is never to happen, the largest possible point
 * in time is given.
 */
DRN_BUDGETING_EXPORT ::QDateTime nextTimePoint(
	const ::QDateTime& current,
	const EventFrequency& ef
);
/**
 * @brief Calculates the conversion ratio to get from one event frequency period to another. The
 * "never" frequency conversion ratio is considered to be zero.
 *
 * @exception Error When either the supplied from or two values are Unknown.
 * @exception Error When either the from or to values are not known values.
 *
 * @param from The frequency period to convert from.
 * @param to The frequency period to convert to.
 *
 * @return The ratio that when multiplying the "from" value by it will be equal to the "to" value.
 */
DRN_BUDGETING_EXPORT double convert(const EventFrequency from, const EventFrequency to);
/**
 * @brief Calculates the number of days in a period of time.
 *
 * @exception Error When the period is not a known value.
 *
 * @param period The frequency period to calculate the number of days in.
 *
 * @return The calculated average number of days for the provided period accounting for leap years.
 */
DRN_BUDGETING_EXPORT foundation::PartialDays daysIn(const EventFrequency period);
/**
 * @brief Calculates the number of times an event occurs within a specific time range, based on its
 * periodicity and other scheduling constraints such as weekly active hours and days.
 * This function determines how frequent events repeat in the given period.
 *
 * @pre The `from` input date must be valid.
 * @pre The `to` input date must be valid.
 *
 * @param period Defines the frequency at which the event occurs.
 * @param from The start date of the period for which occurrences are being calculated.
 * @param to The end date of the period for which occurrences are being calculated.
 *
 * @return The computed number of occurrences of the event within the specified period. This will be
 * negative if the `to` date is earlier than the `from` date.
 */
DRN_BUDGETING_EXPORT double occurrencesIn(
	const EventFrequency period,
	const ::QDate& from,
	const ::QDate to
);

/// @brief All the possible frequencies at which an event may occur in chronological order.
inline constexpr std::array allEventFrequencies{
	EventFrequency::Unknown,
	EventFrequency::Daily,
	EventFrequency::SemiWeekly,
	EventFrequency::Weekly,
	EventFrequency::BiWeekly,
	EventFrequency::SemiMonthly,
	EventFrequency::Monthly,
	EventFrequency::Bimonthly,
	EventFrequency::Quarterly,
	EventFrequency::Quadrimesterly,
	EventFrequency::Semiannually,
	EventFrequency::Annually,
	EventFrequency::Biennially,
	EventFrequency::Triennially,
	EventFrequency::Quadrennially,
	EventFrequency::Quinquennially,
	EventFrequency::Sexennially,
	EventFrequency::Septennially,
	EventFrequency::Octennially,
	EventFrequency::Novennially,
	EventFrequency::Decennially,
	EventFrequency::Never
};

}

#endif
