/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BUDGETING_GOAL_H_
#define DRN_BUDGETING_GOAL_H_

#include <compare>
#include <iosfwd>

#include <budgeting_export.h>

#include "BudgetedMoney.h"


namespace drn::budgeting
{

/**
 * @brief The representation of an amount of money that set aside each period for a particular
 * usage.
 */
class DRN_BUDGETING_EXPORT Goal final : public BudgetedMoney
{
	pecunia::Money final_;

public:
	Goal() = default;
	/**
	 * @brief Minimum initialisation constructor.
	 *
	 * It initializes the internal data to represent a specific goal that aligns with the properties
	 * of the provided budget source. This should be used to create an instance representing a
	 * focused financial target derived from a specific source. The associated source allows for
	 * tracking and manipulation based on the user's budgeting decisions.
	 *
	 * @param source The budget source from which the goal is derived.
	 */
	explicit Goal(const BudgetSource& source);
	/**
	 * @brief Fully initialises all member variables to their supplied values.
	 *
	 * @exception BudgetItemError When the source is empty.
	 * @exception BudgetItemError When the amount is less then zero.
	 * @exception BudgetItemError When the amount is not issued in a known currency.
	 * @exception BudgetItemError When the period is not known.
	 * @exception BudgetItemError When the next occur date is invalid.
	 *
	 * @param source The source of the money budget item.
	 * @param amount The amount the money is for every period.
	 * @param period The frequency period of the event for the money.
	 * @param nextOccurOn The date the money budget item last occurred on.
	 * @param final The final amount of money the goal is completed at.
	 */
	Goal(
		const BudgetSource& source,
		const pecunia::Money& amount,
		const EventFrequency& period,
		const ::QDate& nextOccurOn,
		const pecunia::Money& final
	);
	[[nodiscard]]
	std::strong_ordering operator<=>(const Goal& other) const noexcept;
	[[nodiscard]]
	bool operator==(const Goal& other) const noexcept = default;
	/**
	 * @brief The final amount of money the goal is completed at.
	 */
	const pecunia::Money& final() const;
};

DRN_BUDGETING_EXPORT std::ostream& operator<<(std::ostream& o, const Goal& g);

}

#endif
