/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_CONVERSION_CONVERSION_MAP_H_
#define DRN_CONVERSION_CONVERSION_MAP_H_

#include <pecunia/Currency.hpp>
#include <pecunia/FloatingPoint.hpp>

#include <chrono>
#include <map>

#include <conversion_export.h>


namespace drn::conversion
{

/**
 * @brief A container for mapping from one currency code to the conversion ratio of another.
 */
class DRN_CONVERSION_EXPORT ConversionMap final :
		std::map<
			pecunia::Currency,
			std::map<pecunia::Currency, pecunia::FloatingPoint>
		>
{
	std::chrono::system_clock::time_point lastFetched_;

public:
	using std::map<
		pecunia::Currency,
		std::map<pecunia::Currency, pecunia::FloatingPoint>
	>::begin;
	using std::map<
		pecunia::Currency,
		std::map<pecunia::Currency, pecunia::FloatingPoint>
	>::cbegin;
	using std::map<
		pecunia::Currency,
		std::map<pecunia::Currency, pecunia::FloatingPoint>
	>::cend;
	using std::map<
		pecunia::Currency,
		std::map<pecunia::Currency, pecunia::FloatingPoint>
	>::clear;
	using std::map<
		pecunia::Currency,
		std::map<pecunia::Currency, pecunia::FloatingPoint>
	>::contains;
	using std::map<
		pecunia::Currency,
		std::map<pecunia::Currency, pecunia::FloatingPoint>
	>::empty;
	using std::map<
		pecunia::Currency,
		std::map<pecunia::Currency, pecunia::FloatingPoint>
	>::end;
	using std::map<
		pecunia::Currency,
		std::map<pecunia::Currency, pecunia::FloatingPoint>
	>::size;

	/**
	 * @brief Initialisation constructor populated with the unknown currency code conversion rate.
	 */
	ConversionMap();
	/**
	 * @brief Initialisation constructor populated with the supplied conversion rates.
	 *
	 * @param lastFetched The point at which the conversions were last fetched.
	 * @param ratios The mapping of rations from one currency to another.
	 */
	ConversionMap(
		std::chrono::system_clock::time_point lastFetched,
		std::map<
			pecunia::Currency,
			std::map<pecunia::Currency, pecunia::FloatingPoint>
		> ratios
	);
	/**
	 * @brief Member access for when the conversions were last set.
	 *
	 * @return The time-point the fetch occurred.
	 */
	[[nodiscard]]
	const std::chrono::system_clock::time_point& lastFetched() const noexcept;
	/**
	 * @brief Determines if the currency conversion ratio exists.
	 *
	 * @param from The currency code converting from the given currency.
	 * @param to The currency code to convert to the given currency.
	 *
	 * @return Gives true when the currency conversion ratio exits to go from a supplied code to
	 * another, false else-wise.
	 */
	[[nodiscard]]
	bool hasConversion(
		const pecunia::Currency& from,
		const pecunia::Currency& to
	) const;
	/**
	 * @brief Inserts or updates a conversion ratio entry for a given currency to another.
	 *
	 * @exception Error When the supplied from and to currency codes are the same, but the
	 * conversion ratio is not an identity.
	 *
	 * @param from The currency code the ratio is converting from.
	 * @param to The currency code the ratio is converting to.
	 * @param ratio The ratio, when multiplied, converts to the given currency.
	 */
	void setConversion(
		const pecunia::Currency& from,
		const pecunia::Currency& to,
		const pecunia::FloatingPoint ratio
	);
	/**
	 * @brief Retrieves the conversion mappings for a specified currency.
	 *
	 * @param from The currency for which the conversion mappings are to be retrieved.
	 *
	 * @return The container holding the conversion ratios for the specified currency.
	 */
	[[nodiscard]]
	const std::map<pecunia::Currency, pecunia::FloatingPoint>& conversions(
		const pecunia::Currency& from
	) const noexcept;
	/**
	 * @brief Retrieves the conversion ratio between two currencies.
	 *
	 * @exception Error Thrown if the conversion between the source and target currencies does not
	 * exist.
	 *
	 * @param from The source currency for the conversion.
	 * @param to The target currency for the conversion.
	 *
	 * @return The conversion ratio from the source currency to the target currency.
	 */
	[[nodiscard]]
	const pecunia::FloatingPoint& conversion(
		const pecunia::Currency& from,
		const pecunia::Currency& to
	) const;
};

}

#endif
