/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_CONVERSION_CURRENCY_CONVERTER_H_
#define DRN_CONVERSION_CURRENCY_CONVERTER_H_

#include <QtGlobal>

#include <set>

#include <conversion_export.h>

#include "ConversionMap.h"


namespace pecunia
{

class FloatingPoint;

}
namespace pecunia
{

enum class Currency : quint16;

}

namespace drn::conversion
{

struct CurrencySource;

/**
 * @brief Provides functionality for currency conversion operations.
 *
 * It is designed to facilitate the conversion of monetary values between different currencies
 * using predefined or dynamically fetched conversion rates. It supports both static configuration
 * and runtime retrieval of conversion mappings.
 */
class DRN_CONVERSION_EXPORT CurrencyConverter
{
	static ConversionMap conversions_;

	CurrencySource& source_;

public:
	/**
	 * @brief Retrieves the collection of available currency conversion mappings.
	 *
	 * This method provides access to the cache of conversion mappings, which associate currency
	 * pairs with their respective conversion rates. These mappings are essential for accurate
	 * currency conversion operations throughout the system.
	 *
	 * @return The collection of available conversion mappings.
	 */
	[[nodiscard]]
	static const ConversionMap& conversions() noexcept;
	/**
	 * @brief Configures a conversion rate between two currencies.
	 *
	 * This function allows setting a direct conversion ratio between two currencies. It ensures
	 * that the conversion rate is properly stored and can subsequently be used for currency
	 * conversion operations.
	 *
	 * @param from The source currency for the conversion.
	 * @param to The target currency for the conversion.
	 * @param ratio The conversion rate between the two currencies.
	 *
	 * @post Updates the cache to include the new or updated conversion rate.
	 */
	static void setConversion(
		const pecunia::Currency& from,
		const pecunia::Currency& to,
		const pecunia::FloatingPoint ratio
	);
	/**
	 * @brief Initializes the currency conversion cache used in the converter.
	 *
	 * It is used to load or update the internal mapping of currency conversions, enabling the
	 * system to perform accurate and efficient conversion operations.
	 *
	 * @param conversions A mapping of currency pairs to their respective conversion rates to be
	 * used in future conversions.
	 *
	 * @post Replaces the current currency conversion cache with the provided one. The updated
	 * mapping will serve as the basis for any subsequent currency conversion operations.
	 */
	static void initialise(ConversionMap&& conversions);
	/**
	 * @brief Full initialisation constructor.
	 *
	 * It initializes the instance with a source of currency data for use in conversion
	 * calculations. The source provides the necessary exchange rates, either preloaded or fetched
	 * at runtime. It can then be used to perform lookups and fetch conversions for specified
	 * currencies.
	 *
	 * @param source The source of currency data, used to fetch exchange rates during conversion
	 * operations.
	 */
	explicit CurrencyConverter(CurrencySource& source);
	/**
	 * @brief Retrieves the conversion rate for the specified pair of currencies.
	 *
	 * Ensures that a conversion rate exists for the provided currency pair. If a rate is not
	 * already available in the cache, it attempts to fetch it using a data source and updates the
	 * cache. This ensures that subsequent lookups for the same pair do not require additional
	 * fetching.
	 *
	 * @param from The source currency for the conversion.
	 * @param to The target currency for the conversion.
	 *
	 * @post If the conversion does not exist in the cache, it is fetched and stored for future
	 * lookups.
	 *
	 * @return The conversion rate between the two specified currencies.
	 */
	[[nodiscard]]
	pecunia::FloatingPoint lookUp(
		const pecunia::Currency& from,
		const pecunia::Currency& to
	) const;
	/**
	 * @brief Retrieves all conversion rates between the specified currencies.
	 *
	 * This method calculates and returns a map containing conversion rates for all possible
	 * pairs of the supplied currencies. It ensures that self-conversion rates are set correctly,
	 * and other rates are fetched from an underlying data source. Useful for operations requiring
	 * comprehensive conversion data in a single query.
	 *
	 * @param currencies A collection of currencies for which conversion rates need to be fetched.
	 *
	 * @post The fetched conversion is stored in the currency mapping cache.
	 *
	 * @return A mapping representing the conversion rates between all pairs of the provided
	 * currencies.
	 */
	[[nodiscard]]
	ConversionMap fetchAll(const std::set<pecunia::Currency>& currencies) const;
};

}

#endif
