/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_CONVERSION_CURRENCY_SOURCE_HPP_
#define DRN_CONVERSION_CURRENCY_SOURCE_HPP_

#include <QObject>

#include <conversion_export.h>
#include <foundation/QtMemory.hpp>


namespace pecunia
{

class FloatingPoint;

}
namespace pecunia
{

enum class Currency : quint16;

}

namespace drn::conversion
{

/**
 * @brief Provides the interface functionality for currency exchange rate retrieval.
 *
 * It is designed to facilitate the retrieval of currency conversion rates. It serves as a central
 * mechanism to manage and provide up-to-date foreign exchange rates for calculations, or any system
 * requiring dynamic currency conversion.
 */
struct DRN_CONVERSION_EXPORT CurrencySource : ::QObject
{
	Q_OBJECT

public:
	/**
	 * @brief Full initialisation constructor.
	 *
	 * This constructor initializes the instance with a parent object, ensuring that it integrates
	 * seamlessly with parent-child relationships in the application.
	 *
	 * @param parent A smart pointer to the parent object that the instance will be associated with.
	 */
	explicit CurrencySource(const foundation::QtPtr<QObject>& parent);
	/**
	 * @brief When implemented, it retrieves the conversion rate between two currencies.
	 *
	 * This method is designed to serve as an interface for retrieving conversion rates between two
	 * specific currencies.
	 *
	 * @param from The original currency from which the conversion is initiated.
	 * @param to The target currency to which the conversion is calculated.
	 *
	 * @return The conversion rate between the specified currencies.
	 */
	virtual pecunia::FloatingPoint fetch(
		const pecunia::Currency& from,
		const pecunia::Currency& to
	) const = 0;
};

}

#endif
