/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_CONVERSION_OPEN_EXCHANGE_RATES_SOURCE_H_
#define DRN_CONVERSION_OPEN_EXCHANGE_RATES_SOURCE_H_

#include <QString>

#include <conversion_export.h>
#include <foundation/QtMemory.hpp>

#include "CurrencySource.hpp"


namespace pecunia
{

class FloatingPoint;

}
namespace pecunia
{

enum class Currency : quint16;

}

namespace drn::networking
{

struct NotifyingSender;

}
namespace drn::conversion
{

/**
 * @brief Manages the details of fetching conversion rates between two currencies from the
 * Open Exchange Rates, https://openexchangerates.org/ .
 */
class DRN_CONVERSION_EXPORT OpenExchangeRatesSource final : public CurrencySource
{
	networking::NotifyingSender& sender_;
	QString appId_;

public:
	static constexpr auto host_{"openexchangerates.org"};

	/**
	 * @brief Full initialisation constructor.
	 *
	 * @param sender The object through which the network is accessible for sending requests.
	 * @param appId The application identifier for the OER servers.
	 * @param parent The parent widget of the object.
	 */
	OpenExchangeRatesSource(
		networking::NotifyingSender& sender,
		QString appId,
		foundation::QtPtr<QObject> parent
	);
	~OpenExchangeRatesSource() override;
	/**
	 * @brief Synchronously retrieves the conversion from one currency code to another.
	 *
	 * @exception ReplyError When it fails to during the reply of a conversion request.
	 * @exception ParsingError When the data returned contains an error.
	 *
	 * @param from The currency to convert from.
	 * @param to The currency to convert to.
	 *
	 * @return The ratio that when multiplied will take the value of "from" to the value of "to".
	 */
	pecunia::FloatingPoint fetch(
		const pecunia::Currency& from,
		const pecunia::Currency& to
	) const override;
	/**
	 * @brief The application identifier for the OER servers.
	 */
	[[nodiscard]]
	const QString& applicationId() const;
};

}

#endif
