/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_DESKTOP_UI_CURRENCY_COMBOBOX_H_
#define DRN_DESKTOP_UI_CURRENCY_COMBOBOX_H_

#include <pecunia/Currency.hpp>

#include <QComboBox>

#include <set>

#include <navigation/DisplayCurrency.h>

#include <foundation/QtMemory.hpp>
#include <desktop-ui_export.h>


namespace drn::desktop_ui
{

/**
 * @brief A GUI item providing a selectable list of currencies.
 *
 * This class represents a specialized GUI element that allows users to select
 * from a predefined set of currencies. It supports dynamic updates to its list
 * of options and emits signals when the selection changes, enabling seamless
 * integration with other components in the application.
 */
class DRN_DESKTOP_UI_EXPORT CurrencyComboBox final : public ::QComboBox
{
	Q_OBJECT

	using ::QComboBox::addItem;
	using ::QComboBox::clear;
	using ::QComboBox::itemText;

	std::set<pecunia::Currency> usableCodes_;
	navigation::CurrencyDisplayOptions displayAs_;

public slots:
	/**
	 * @brief Updates the currently selected currency in the GUI item.
	 *
	 * This function sets the current selection of a GUI element representing a selectable list
	 * of currencies. It ensures logical display behavior for a specific currency, adjusting
	 * the display text or index based on predefined conditions. It integrates seamlessly with
	 * other parts of the application for real-time updates.
	 *
	 * @param code The currency to be set as the currently selected value.
	 *
	 * @post The selected currency in the GUI item is updated.
	 */
	void setCurrentCode(const pecunia::Currency& code);

public:
	/**
	 * @brief Constructs a specialized GUI item for selecting currencies.
	 *
	 * This constructor initializes a GUI element that provides users with the ability to select
	 * currencies from a list. The list is displayed and managed in an alphabetically sorted order
	 * to ensure usability and clarity. The widget is created within the given parent GUI container,
	 * if specified.
	 *
	 * @param parent The GUI element acting as the container for this widget.
	 */
	explicit CurrencyComboBox(const foundation::QtPtr<::QWidget>& parent);
	/**
	 * @brief Retrieves the set of currencies currently displayed in the GUI item.
	 *
	 * This method should be used to get a snapshot of the selectable currency options at any given
	 * point in time.
	 *
	 * @return A collection of all valid currencies displayed by the GUI item. The returned
	 * collection reflects the current state of the GUI item and excludes any invalid or empty
	 * entries.
	 */
	[[nodiscard]]
	std::set<pecunia::Currency> currencies() const;
	/**
	 * @brief Retrieves the currently selected currency code from the GUI item.
	 *
	 * This function returns the currency code corresponding to the user's current selection
	 * in the GUI element. It ensures that the value aligns with predefined and usable
	 * currency codes.
	 *
	 * @exception DataEntryError Throws this exception if the selected country code is not supported
	 * or recognized. The error will indicate the problematic input and provide additional context.
	 *
	 * @return The currency code corresponding to the user's current selection in the GUI item.
	 */
	[[nodiscard]]
	pecunia::Currency currentCode() const;
	/**
	 * @brief Populates the GUI item with a list of usable currencies.
	 *
	 * This function updates the selectable options within the GUI element to reflect a given
	 * set of usable currencies. The GUI item's contents are kept consistent with the provided data,
	 * ensuring accurate and usable output for end-users.
	 *
	 * @param usableCodes A collection of currency codes that are considered valid and should
	 * be added to the list.
	 * @param displayAs The specific manner in which each currency should be displayed within
	 * the GUI element.
	 *
	 * @post All preexisting selectable options in the GUI item are cleared.
	 * @post The GUI item is populated with new options corresponding to the `usableCodes`
	 * collection.
	 * @post The default selection in the GUI item is updated based on the provided data. If only
	 * one currency exists, it is selected.
	 */
	void populateUsableCurrencies(
		const std::set<pecunia::Currency>& usableCodes,
		const navigation::CurrencyDisplayOptions& displayAs
	);
};

}

#endif
