/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_DESKTOP_UI_EVENT_FREQUENCY_ENTRY_WIDGET_H_
#define DRN_DESKTOP_UI_EVENT_FREQUENCY_ENTRY_WIDGET_H_

#include <QWidget>

#include <vector>

#include <desktop-ui_export.h>
#include <foundation/QtMemory.hpp>
#include <ui_EventFrequencyEntryWidget.h>


namespace drn::budgeting
{

enum class EventFrequency : quint8;

}
namespace drn::desktop_ui
{

/**
 * @brief A widget that facilitates the entry of frequency-based event data via a GUI element. This
 * widget emits signals when new frequency information is entered by the user.
 *
 * This class is designed to be used in applications where users need to input frequency-related
 * data, such as intervals, repetitions, or periodic events in a visually intuitive manner. The
 * widget ensures valid data entry and provides feedback for errors. Integration of this widget can
 * help streamline frequency event configuration tasks.
 */
class DRN_DESKTOP_UI_EXPORT EventFrequencyEntryWidget final :
		public QWidget,
		Ui::EventFrequencyEntryWdgt
{
	Q_OBJECT

private slots:
	void onCurrentIndexChanged(const int index);

signals:
	/**
	 * @brief Triggered when the user changes the selected event frequency using a GUI element.
	 *
	 * It is emitted each time the user modifies the current selection in an event frequency
	 * chooser. It enables listeners to respond dynamically, such as updating related fields or
	 * recalculating dependent values. It is particularly useful for budget or scheduling operations
	 * where event frequencies directly influence computations or visualisations.
	 *
	 * @param changed The newly selected event frequency, indicating how often events are expected
	 * to occur. This parameter reflects the user's selection.
	 */
	void selectionChanged(const budgeting::EventFrequency& changed);

public:
	/**
	 * @brief Minimum initialisation constructor.
	 *
	 * This constructor initialises the widget by setting up associated GUI components, populating a
	 * selection-based GUI element with all the predefined event frequencies, and connecting signals
	 * to internal slots.
	 *
	 * @param parent The parent GUI object for hierarchical ownership or scope management. Can also
	 * be null for no parent association.
	 */
	explicit EventFrequencyEntryWidget(const foundation::QtPtr<::QWidget>& parent);
	/**
	 * @brief Full initialisation constructor, initialising its GUI elements and populating them
	 * with the provided frequency options.
	 *
	 * It sets up the GUI elements, ensuring that an intuitive and interactive interface is
	 * available. If the frequency list contains the unknown entry, it is represented with an empty
	 * item in the selection list.
	 *
	 * @pre The `frequencies` list must not be empty to ensure meaningful initialisation.
	 *
	 * @param frequencies A collection of possible event frequencies to populate the GUI element.
	 * These frequencies are displayed for selection and interaction.
	 * @param parent An optional GUI container that manages memory and hierarchies; this is usually
	 * the parent widget owning the entry widget.
	 */
	EventFrequencyEntryWidget(
		const std::vector<budgeting::EventFrequency>& frequencies,
		const foundation::QtPtr<::QWidget>& parent
	);
	/**
	 * @brief Sets the period description on the relevant GUI element for the widget.
	 *
	 * @exception DataEntryError Thrown if the provided description is empty or composed only of
	 * white spaces. This ensures that valid descriptions are always used.
	 *
	 * @param description The new description to be displayed on the associated GUI element. It
	 * should clearly convey the intended period's representation.
	 *
	 * @post The GUI element reflecting the period description is updated with the specified text.
	 * This change is visible to the user.
	 */
	void setDescription(const ::QString& description);
	/**
	 * @brief Updates the GUI element to reflect the specified frequency.
	 *
	 * This method is used to programmatically select a frequency in the relevant GUI element.
	 *
	 * @exception DataEntryError Thrown if the specified frequency is not a valid option in the GUI
	 * element.
	 *
	 * @param period The period frequency to be selected in the GUI element.
	 *
	 * @post The GUI element reflecting frequency selection is updated to match the input frequency.
	 */
	void select(const budgeting::EventFrequency& period);
	/**
	 * @brief Retrieves the currently selected event frequency from the selection-based GUI element.
	 *
	 * @return The frequency currently selected in the GUI element.
	 */
	[[nodiscard]]
	budgeting::EventFrequency period() const noexcept;
};

}

#endif
