/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_DESKTOP_UI_MAIN_WINDOW_H_
#define DRN_DESKTOP_UI_MAIN_WINDOW_H_

#include <pecunia/Currency.hpp>
#include <pecunia/Money.hpp>

#include <QMainWindow>
#include <QString>
#include <QToolBar>

#include <chrono>
#include <map>
#include <memory>
#include <optional>
#include <set>
#include <utility>
#include <vector>

#include <accounting/AccountCode.h>
#include <accounting/Account.h>
#include <accounting/LedgerBooks.h>
#include <accounting/TransactionNumber.h>
#include <adviser/BudgetHealth.h>
#include <banking/BankAccount.h>
#include <banking/BankAccountType.h>
#include <banking/BankMap.h>
#include <banking/BankName.h>
#include <banking/ReconciledBalance.h>
#include <budgeting/BillMap.hpp>
#include <budgeting/BudgetItemDue.h>
#include <budgeting/BudgetItemIdentifier.h>
#include <budgeting/BudgetItemTypes.h>
#include <budgeting/DebtMap.h>
#include <budgeting/GoalMap.h>
#include <budgeting/NontrackMap.hpp>
#include <budgeting/WageMap.hpp>
#include <conversion/ConversionMap.h>
#include <desktop-ui_export.h>
#include <foundation/Percentage.h>
#include <foundation/QtMemory.hpp>
#include <navigation/AllowSubAccounts.hpp>
#include <navigation/BankAccountCode.hpp>
#include <navigation/BudgetBankAccount.h>
#include <navigation/BudgetWizardTemplates.h>
#include <surveying/IndexedIncreaseDecreases.hpp>
#include <surveying/MappingSurvey.h>
#include <updating/ReleaseInformation.hpp>


class QCloseEvent;

namespace Ui
{

class BwmMnwndw;

}
namespace drn::navigation
{

struct ConfigurationOptions;
enum class CurrencyDisplayOptions;

}
namespace drn::desktop_ui
{

class DRN_DESKTOP_UI_EXPORT MainWindow final : public QMainWindow
{
	Q_OBJECT

	static const ::QString budgetFileExtensionFilter_;

	std::unique_ptr<Ui::BwmMnwndw> ui_;
	const navigation::ConfigurationOptions& options_;
	bool confirmedExit_;
	std::optional<bool> requestedNewBudgetViaWizard_;
	foundation::QtPtr<QToolBar> accountingTbr_;

private slots:
	//{ Help Menu

	void onShowAbout();
	void onShowAboutQt();

	//}

	void onDisplayBudgetInterface(const drn::budgeting::BudgetItemTypes& type);

protected:
	void closeEvent(QCloseEvent* event) override;

public slots:
	void onShowBills(
		const drn::budgeting::BillMap& bills,
		const pecunia::Currency& preferred,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void onShowDebts(
		const drn::budgeting::DebtMap& debts,
		const std::set<drn::navigation::BudgetBankAccount>& accounts,
		const std::set<drn::accounting::AccountCode>& codes,
		const std::set<drn::banking::BankName>& banks,
		const pecunia::Currency& preferred,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void onShowGoals(
		const drn::budgeting::GoalMap& goals,
		const std::set<drn::navigation::BudgetBankAccount>& accounts,
		const std::set<drn::accounting::AccountCode>& codes,
		const std::set<drn::banking::BankName>& banks,
		const pecunia::Currency& preferred,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const pecunia::Money& wagesTotal
	);
	void onShowNontracks(
		const drn::budgeting::NontrackMap& nontracks,
		const pecunia::Currency& preferred,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void onShowWages(
		const drn::budgeting::WageMap& wages,
		const pecunia::Currency& preferred,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void onShowAccounts(
		const std::set<drn::accounting::Account>& accounts,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void onShowLedgers(
		const std::set<drn::accounting::Account>& accounts,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void onShowBanks(
		const drn::banking::BankMap& banks,
		const std::set<drn::accounting::Account>& allAccounts,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const std::set<drn::banking::BankAccountType>& accountTypes
	);
	void onShowProcessBudgetItems(
		const std::chrono::days& dueWithin,
		const std::vector<drn::budgeting::BudgetItemAmountDue>& items,
		const std::set<drn::banking::BankName>& banks,
		const pecunia::Currency& preferredCurrency,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const std::set<drn::accounting::AccountCode>& codes
	);
	void onShowSurvey(
		const std::set<drn::banking::BankName>& banks,
		const std::set<pecunia::Currency>& usable,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const drn::foundation::Percentage& balanceWindow,
		const std::set<drn::accounting::AccountCode>& codes
	);
	/**
	 * @brief Displays a confirmation dialog to verify the user's intent to create a new budget and
	 * emits appropriate signals based on the user's response.
	 *
	 * This should be used when a user initiates the creation of a new budget, ensuring they are
	 * aware of potential loss of unsaved changes and providing an opportunity to confirm or cancel
	 * the action. Signals emitted notify subscribed components or systems about the user's decision
	 * and the context of the requested budget creation (using a wizard or starting empty).
	 *
	 * @pre The application must have a valid budget creation request, stored as a value in the
	 * relevant member variable.
	 *
	 * @post Displays a confirmation dialog using a GUI element to prompt the user.
	 * @post If the user confirms, emits a signal indicating the type of new budget to create, such
	 * as using a wizard, or starting with an empty budget.
	 * @post Resets the tracked request for new budget creation to an uninitialized state.
	 */
	void onShowConfirmNewBudget();
	/**
	 * @brief Displays a status message about the creation of a new budget and emits a signal to
	 * prepare the display of the primary dashboard GUI element.
	 *
	 * This should be used when initializing or transitioning the user to a new budgeting process,
	 * ensuring they are informed about the new budget creation and directing them to the
	 * appropriate dashboard view.
	 *
	 * @post A status message is shown in the status bar for a specified duration.
	 * @post A signal is emitted to notify subscribers to prepare and show the dashboard interface.
	 */
	void onShowNewBudget();
	void onShowOpenBudgetFile();
	void onShowConfirmLoadBudgetFile(const QString& filePathName);
	void onLoadedBudgetFile(const QString& filePathName, const bool isReloaded);
	void onSavedBudgetFile(const QString& filePathName);
	void onShowSaveAsBudgetFile();
	void onShowNoBudgetFileOpened();
	void onCheckCleanExit(const bool isDirty);
	void show();
	void onShowDashboard(
		const drn::budgeting::BudgetTotals& bt,
		const std::vector<drn::budgeting::BudgetItemAmountDue>& due,
		const drn::adviser::BudgetHealth& bh,
		const std::chrono::days& dueWithin,
		const pecunia::Currency& preferred,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void onShowExchangeRates(
		const drn::conversion::ConversionMap& currentConversions,
		const std::set<pecunia::Currency>& currencies
	);
	void onShowConfigure(
		const std::chrono::hours& hrInWeek,
		const std::chrono::days& dayInWeek,
		const std::chrono::days& notifyDueWithin,
		const pecunia::Currency& preferred,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const std::set<pecunia::Currency>& usable,
		const std::set<pecunia::Currency>& used,
		const drn::foundation::Percentage& balanceWindow
	);
	void onFetchedReleaseNotes(
		const std::set<drn::updating::ReleaseInformation>& versions
	);
	void onUpdateReleased(const drn::updating::SemanticVersion& latest);
	void onShowMoneyTransfer(
		const std::set<drn::navigation::BankAccountCode>& fromBanks,
		const std::set<drn::navigation::BankAccountCode>& toBanks,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	/**
	 * @brief Sets up and displays a GUI element for budget item refunds, enabling interactions such
	 * as looking up sources, transactions, or posting refunds. Additionally, connects signals and
	 * slots for transaction save results and the widget's lifecycle events.
	 *
	 * This should be used when the user needs to view and process refunds related to specific
	 * budget items. The associated signals provide functionality for retrieving associated data and
	 * managing refund operations in an interactive interface.
	 *
	 * @pre The application framework must be initialized, and this method must only be called from
	 * the main thread of the application.
	 *
	 * @param codes A collection of currency codes that determine the scope of refundable budget
	 * items.
	 * @param displayAs Options that define how refund information should be displayed within the
	 * GUI.
	 *
	 * @post A GUI element is created and set as the primary view for the application window.
	 * Internal  signal and slot connections are established to handle interactions such as looking
	 * up budget item sources, looking up refundable transactions, posting refunds, and handling
	 * save operations.
	 */
	void onShowBudgetItemRefund(
		const std::set<pecunia::Currency>& codes,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	/**
	 * @brief Sets up and displays a GUI element for a first-time setup wizard, enabling users to
	 * configure settings such as preferred currency and display options.
	 *
	 * This should be used when the application is launched for the first time, allowing users to
	 * define initial preferences and streamline their onboarding experience. The wizard provides an
	 * interactive GUI element for setting up the application.
	 *
	 * @param preferred The default display currency option for the application.
	 * @param usable The set of currencies marked as usable, which will be reflected in the GUI
	 * list.
	 * @param displayAs The display preferences defining how currency values should be formatted or
	 * shown within the application during configuration.
	 *
	 * @post A GUI element for the first-time setup wizard is created and displayed. A connection is
	 * established between the wizard's closing event and the method that switches to the dashboard
	 * view.
	 */
	void onShowNewBudgetFirstTimeWizard(
		const pecunia::Currency& preferred,
		const std::set<pecunia::Currency>& usable,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);

signals:
	void prepareShowBills();
	void prepareShowDebts();
	void prepareShowGoals();
	void prepareShowNontracks();
	void prepareShowWages();
	void added(
		const drn::budgeting::BudgetItemTypes& type,
		const drn::budgeting::BudgetedMoney& item,
		const std::optional<drn::banking::BankAccount>& ba
	);
	void renamed(
		const drn::budgeting::BudgetItemTypes& type,
		const drn::budgeting::BudgetSource& original,
		const drn::budgeting::BudgetSource& changed
	);
	void updated(
		const drn::budgeting::BudgetItemTypes& type,
		const drn::budgeting::BudgetedMoney& item,
		const std::optional<drn::banking::BankAccount>& ba
	);
	void removed(
		const drn::budgeting::BudgetItemTypes& type,
		const drn::budgeting::BudgetSource& source
	);
	void saveSucceeded(const drn::budgeting::BudgetSource& source);
	void saveFailed(
		const std::optional<drn::budgeting::BudgetSource>& original,
		const QString& errorMessage,
		const std::optional<drn::budgeting::BudgetSource>& displayed
	);
	void prepareShowAccounts();
	void addedAct(const drn::accounting::Account& a);
	void recoded(
		const drn::accounting::AccountCode& current,
		const drn::accounting::AccountCode& changed
	);
	void closed(const drn::accounting::AccountCode& code);
	void opened(const drn::accounting::AccountCode& code);
	void removedAct(const drn::accounting::AccountCode& code);
	void accountSaveSucceeded(const drn::accounting::AccountCode& code);
	void accountSaveFailed(
		const std::optional<drn::accounting::AccountCode>& original,
		const QString& errorMessage,
		const std::optional<drn::accounting::AccountCode>& displayed
	);
	void prepareShowLedgers();
	void selectedLedger(const drn::accounting::AccountCode& code);
	void showLedger(const drn::accounting::LedgerAccount& ledger);
	void addedTransaction(const drn::accounting::Transaction& txn);
	void updatedTransaction(
		const drn::accounting::TransactionNumber& tn,
		const drn::accounting::Transaction& txn
	);
	void markCleared(const std::set<drn::accounting::TransactionNumber>& numbers);
	void markUnreconciled(const std::set<drn::accounting::TransactionNumber>& numbers);
	void markReconciled(const drn::accounting::TransactionNumber& number);
	void removedTransaction(const drn::accounting::TransactionNumber& number);
	void transactionSaveSucceeded(
		const std::optional<drn::accounting::TransactionNumber>& displayed,
		const std::optional<drn::accounting::TransactionNumber>& posted
	);
	void transactionSaveFailed(
		const std::optional<drn::accounting::TransactionNumber>& tn,
		const QString& errorMessage
	);
	void prepareShowBanks();
	void bankAdded(
		const drn::banking::Bank& b,
		const std::map<
			drn::accounting::AccountNumber,
			drn::banking::SupportedAccountTypes
		>& accountNumbersBankTypes
	);
	void bankUpdated(
		const drn::banking::Bank& b,
		const std::map<
			drn::accounting::AccountNumber,
			drn::banking::SupportedAccountTypes
		>& accountNumbersBankTypes
	);
	void bankRenamed(
		const drn::banking::BankName& original,
		const drn::banking::BankName& change
	);
	void bankClosed(const drn::banking::BankName& bn);
	void bankOpened(const drn::banking::BankName& bn);
	void bankRemoved(const drn::banking::BankName& bn);
	void bankSaveSucceeded(const drn::banking::BankName& bn);
	void bankSaveFailed(
		const std::optional<drn::banking::BankName>& original,
		const QString& errorMessage,
		const std::optional<drn::banking::BankName>& displayed
	);
	void addedBankAccount(
		const QString& name,
		const drn::banking::SupportedAccountTypes& type,
		const pecunia::Money& openingBalance
	);
	void renamedBankAccount(
		const drn::accounting::AccountCode& original,
		const drn::accounting::AccountCode& changed
	);
	void removedBankAccount(const drn::accounting::AccountCode& ac);
	void bankAccountSaveSucceeded(const drn::accounting::AccountCode& ac);
	void bankAccountSaveFailed(
		const std::optional<drn::accounting::AccountCode>& original,
		const QString& errorMessage,
		const std::optional<drn::accounting::AccountCode>& displayed
	);
	void prepareShowProcessBudgetItems();
	void prepareShowSurvey();
	void lookUpBankAccounts(
		const drn::banking::BankName& bn,
		const drn::navigation::AllowSubAccounts& shouldAllow
	);
	void lookedUpBankAccounts(
		const drn::banking::BankName& bn,
		const std::set<drn::banking::BankAccountType>& accountTypes
	);
	/**
	 * @brief Emitted when the user requests to look up all sources related to a specific budget
	 * item type.
	 *
	 * This should be used when there is a need to identify all associated sources for a given
	 * budget item, categorized by its type.
	 *
	 * @param requestedBy A unique identifier specifying the budget item for which sources need to
	 * be retrieved.
	 */
	void lookUpBudgetItemSources(const drn::budgeting::BudgetItemDueIdentifier& requestedBy);
	/**
	 * @brief Emitted when the available sources for a specific budget item type are retrieved.
	 *
	 * This signal should be used to notify the completion of a source lookup for a specific budget
	 * item. It facilitates downstream processes or UI updates by providing the set of associated
	 * sources for the identified budget item type.
	 *
	 * @param requestedBy A unique identifier specifying the exact budget item for which sources
	 * were retrieved.
	 * @param sources The set of sources determined to be associated with the budget item type.
	 */
	void lookedUpBudgetItemSources(
		const drn::budgeting::BudgetItemDueIdentifier& requestedBy,
		const std::set<drn::budgeting::BudgetSource>& sources
	);
	void processBudgetItemPosted(
		const drn::budgeting::BudgetItemAmountDue& bid,
		const drn::accounting::AccountCode& ac,
		const QDate& postOn
	);
	void processBudgetItemPostedUnexpected(
		const drn::budgeting::BudgetItemDueIdentifier& bidi,
		const pecunia::Money& amount,
		const drn::accounting::AccountCode& ac,
		const QDate& postOn
	);
	void processBudgetItemSkipped(const drn::budgeting::BudgetItemDueIdentifier& bidi);
	void processBudgetItemSaveSucceeded(
		const drn::budgeting::BudgetItemDueIdentifier& bidi,
		const std::optional<drn::accounting::TransactionNumber>& txnNum
	);
	void processBudgetItemSaveFailed(
		const drn::budgeting::BudgetItemDueIdentifier& bidi,
		const QString& errorMessage
	);
	/**
	 * @brief Emitted when the user wishes to prepare a new budget item using an empty one.
	 */
	void prepareNewBudgetEmpty();
	/**
	 * @brief Triggered when the user confirms that they wish to have a new budget via an empty one.
	 */
	void confirmedNewBudgetEmpty();
	void loadBudgetFile(const QString& filePathName);
	void confirmedLoadBudgetFile(const QString& filePathName, const bool isReloaded);
	void saveBudgetFile();
	void saveAsBudgetFile(const QString& filePathName);
	void reloadBudgetFile();
	void checkIsBudgetFileDirty();
	void prepareShowDashboard();
	void prepareShowExchangeRates();
	void fetchExchangeRates(const std::set<pecunia::Currency>& currencies);
	void fetchedExchangeRates(
		const std::set<pecunia::Currency>& currencies,
		const drn::conversion::ConversionMap& conversions
	);
	void storeExchangeRates(const drn::conversion::ConversionMap& conversions);
	void prepareShowConfigure();
	void configureUpdated(
		const std::chrono::hours& hrInWeek,
		const std::chrono::days& dayInWeek,
		const std::chrono::days& notifyDueWithin,
		const pecunia::Currency& preferred,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const std::set<pecunia::Currency>& usable,
		const drn::foundation::Percentage& balanceWindow
	);
	void configureSaveSucceeded();
	void configureSaveFailed(const QString& errorMessage);
	/**
	 * @brief Triggered when the user wishes to look up and processes the most recent
	 * reconciliation for a specified account.
	 *
	 * This method should be used when there is a need to analyze or display the latest
	 * reconciliation details for a given account.
	 *
	 * @param an The identifier representing the account for which to look up the latest
	 * reconciliation.
	 */
	void lookUpLastReconciliation(const drn::accounting::AccountNumber& an);
	/**
	 * @brief This is to be triggered when the details of the last reconciliation for the current
	 * financial account displayed is available.
	 *
	 * @param code The currency associated with the selected account reconciliation process.
	 * @param reconciliation The last reconciled balance, if available.
	 *
	 * @post Retrieves the last reconciliation details and displays them using a GUI item.
	 * @post Updates the internal state of the application to reflect the data of the displayed
	 * reconciliation.
	 */
	void lookedUpLastReconciliation(
		const pecunia::Currency& code,
		const std::optional<drn::banking::ReconciledBalance>& reconciliation
	);
	void generateMappingSurvey(
		const drn::banking::BankName& bn,
		const drn::accounting::AccountNumber& an,
		const QDate& surveyingOn
	);
	void generatedMappingSurvey(const drn::surveying::MappingSurvey& survey);
	void completedSurvey(
		const QDate& completedOn,
		const drn::banking::BankAccount& ba,
		const pecunia::Money& balance,
		const std::set<drn::accounting::TransactionNumber>& allReconciled,
		const drn::surveying::IndexedIncreaseDecreases<
			drn::budgeting::BudgetItemIdentifier
		>& allAdded,
		const drn::surveying::IndexedIncreaseDecreases<drn::accounting::AccountNumber>& allExtra,
		const std::map<
			drn::budgeting::BudgetItemIdentifier,
			drn::surveying::DistributedAmount
		>& distributed,
		const pecunia::Money& nondistributed
	);
	void completedSurveySucceeded();
	void completedSurveyFailed(const QString& errorMessage);
	void fetchReleaseNotes();
	void prepareShowMoneyTransfer();
	void transferRequested(
		const drn::accounting::AccountNumber& from,
		const drn::accounting::AccountNumber& to,
		const pecunia::Money& amount,
		const QString& reason
	);
	void prepareShowBudgetItemRefund();
	/**
	 * @brief Emitted when the users wishes to look up all refundable transactions for a specific
	 * budget item.
	 *
	 * @param bii A unique identifier that specifies the budget item for which refundable
	 * transactions need to be located.
	 */
	void lookUpRefundableTransactions(const drn::budgeting::BudgetItemIdentifier& bii);
	/**
	 * @brief Triggered when the result of a refundable transactions lookup for a specific budget
	 * item is completed.
	 *
	 * @param bii The unique identifier specifying the budget item for which refundable transactions
	 * were retrieved. This ensures that the processing is scoped to the relevant budget item.
	 * @param ac The identifier in the ledger books for the account of the budget item.
	 * @param txns A set of transactions identified as refundable for the specified budget item.
	 */
	void lookedUpRefundableTransactions(
		const drn::budgeting::BudgetItemIdentifier& bii,
		const drn::accounting::AccountCode& ac,
		const std::set<drn::accounting::Transaction>& txns
	);
	/**
	 * @brief Emitted when the users clicks to save a budget item refund.
	 *
	 * @param txnNum The unique identifier representing the transaction to which the budget item
	 * is to be refunded.
	 * @param refund The refund amount to be posted against the specified transaction.
	 */
	void postRefund(
		const drn::accounting::TransactionNumber& txnNum,
		const pecunia::Money& refund
	);
	/**
	 * @brief Emitted when the user wishes to prepare a new budget item using a first-time wizard.
	 */
	void prepareNewBudgetFirstTimeWizard();
	/**
	 * @brief Triggered when the user confirms that they wish to display the first-time budget
	 * wizard.
	 */
	void confirmedNewBudgetFirstTimeWizard();
	/**
	 * @brief Triggered when the user completes the first-time setup process.
	 *
	 * This function provides the first-time budget setup process by capturing and processing the
	 * user-provided details such as currency preferences, bank account details, and budget template
	 * selections.
	 *
	 * Ensure the required data is valid and complete before invoking this method.
	 *
	 * @pre The wizard process must have all necessary preferences inputted and validated.
	 *
	 * @param preferred The user's preferred currency for displaying financial information.
	 * @param usable A collection of currencies that are usable in the application.
	 * @param displayAs Specifies options for how the currencies will be displayed to the user.
	 * @param primary The primary banking institution selected by the user.
	 * @param kind Supported account type associated with the primary banking institution.
	 * @param accountName The user-defined name for the account being set up.
	 * @param initialBalance The initial monetary balance for the specified account.
	 * @param selectedTemplate The primary wizard template chosen by the user for setup guidance.
	 */
	void completedNewBudgetFirstTimeWizard(
		const pecunia::Currency& preferred,
		const std::set<pecunia::Currency>& usable,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const drn::banking::BankName& primary,
		const drn::banking::SupportedAccountTypes& kind,
		const QString& accountName,
		const pecunia::Money& initialBalance,
		const drn::navigation::BudgetWizardTemplates& selectedTemplate
	);
	void updateAvailableAcknowledged();

public:
	/**
	 * @brief Constructs the main application window.
	 *
	 * @param options Configuration settings specifying application behaviour and UI elements.
	 *
	 * @post Creates and displays the main application window as the central interface for users.
	 * The first widget to be displayed is the dashboard view.
	 * @post Initialises key components such as toolbars based on the specified options.
	 */
	explicit MainWindow(const navigation::ConfigurationOptions& options);
	/**
	 * @brief Destroys the instance and cleans up any allocated resources.
	 *
	 * @post Releases all allocated resources associated with the instance.
	 * @post Ensures that any ongoing or pending operations relevant to the are
	 * finalised.
	 */
	~MainWindow() override;
};

}

#endif
