/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_DESKTOP_UI_MONEYENTRYWIDGET_H_
#define DRN_DESKTOP_UI_MONEYENTRYWIDGET_H_

#include <pecunia/Currency.hpp>

#include <QString>
#include <QWidget>

#include <set>

#include <foundation/QtMemory.hpp>
#include <ui_MoneyEntryWidget.h>

#include <desktop-ui_export.h>


namespace pecunia
{

class Money;

}

namespace drn::desktop_ui::tests
{

struct MoneyEntryWidgetUT;

}

namespace drn::desktop_ui::details
{
class DRN_DESKTOP_UI_EXPORT MoneyAmountValidator final : public QValidator
{
	Q_OBJECT

	pecunia::Currency code_;

public:
	MoneyAmountValidator(
		const pecunia::Currency& code,
		foundation::QtPtr<QObject> parent
	);
	State validate(QString& input, int& pos) const override;
};
}
namespace drn::desktop_ui
{
/**
 * @brief A Qt Widget to display a money entry.
 */
class DRN_DESKTOP_UI_EXPORT MoneyEntryWidget final : public QWidget, Ui::MoneyEntryWidget
{
	Q_OBJECT

	friend drn::desktop_ui::tests::MoneyEntryWidgetUT;

	bool eventFilter(QObject* obj, QEvent* event) override;

	quint8 displayDigits_;
	bool isReadOnly_;

private slots:
	void setAmountValidator();

public slots:
	/**
	 * @brief Sets the current index of the currency code.
	 *
	 * @exception Error if the index exceeds the number of available codes in the GUI element.
	 *
	 * @param index The index to set as current.
	 *
	 * @post The currency code GUI element displays the currency code at the supplied index.
	 */
	void setCurrentIndex(const int index);

signals:
	//{ QLineEdit Signals

	/**
	 * @brief Triggered when the user finishes editing the displayed money amount.
	 */
	void editingFinished();

	//}

	//{ QComboBox Signals

	/**
	 * @brief Triggered when the current index of the currency selection changes.
	 *
	 * @param index The new selected index in the currency GUI element.
	 */
	void currentIndexChanged(int index);

	//}

	/**
	 * @brief Triggered when the value of the money changes.
	 *
	 * @param m The new value for the money when valid, default constructed value when it's invalid.
	 */
	void moneyChanged(const pecunia::Money& m);

public:
	/**
	 * @brief Fully initialises the widget.
	 *
	 * @param parent The parent responsible for the memory of the widget.
	 */
	explicit MoneyEntryWidget(const foundation::QtPtr<QWidget>& parent);
	/**
	 * @brief An accessor to the possible currencies that are selectable.
	 */
	[[nodiscard]]
	std::set<pecunia::Currency> currencies() const noexcept;
	/**
	 * @brief Changes the possible currencies that a user can select.
	 *
	 * @param code The code the user can select from.
	 * @param displayAs Options for how the currencies should be displayed.
	 *
	 * @post The GUI element for currency selection is populated with the supplied currencies. The
	 * first entry is empty (Currency::XXX), to allow selecting no currency.
	 */
	void setUsableCurrency(
		const pecunia::Currency& code,
		const navigation::CurrencyDisplayOptions& displayAs
	);
	/**
	 * @brief Changes the set of usable currencies for the widget.
	 *
	 * @param codes A set of Currency codes that are allowed to be used.
	 * @param displayAs Options for how the currencies should be displayed.
	 *
	 * @post The GUI element for currency selection is populated with the supplied currencies. The
	 * first entry is empty (Currency::XXX), to allow selecting no currency.
	 */
	void setUsableCurrencies(
		const std::set<pecunia::Currency>& codes,
		const navigation::CurrencyDisplayOptions& displayAs
	);
	/**
	 * @brief Changes the displayed amount of money.
	 *
	 * @param money The new money value to be displayed.
	 *
	 * @post The GUI elements displaying the monetary amount are updated to the supplied values.
	 */
	void setAmount(const pecunia::Money& money);
	/// An accessor to the displayed amount of money.
	[[nodiscard]]
	pecunia::Money amount() const;
	/// An accessor to the displayed amount of money.
	[[nodiscard]]
	std::optional<pecunia::Money> tryAmount() const;
	/**
	 * @brief Changes the label text for the money entry widget.
	 *
	 * @param text The new text to be set as the label.
	 *
	 * @post The GUI element displaying the label information about the money widget to the supplied
	 * values.
	 */
	void setLabel(const QString& text);
	/**
	 * @brief Hides the currency code selection GUI element.
	 *
	 * @post This the visibility of the currency code GUI element is hidden from the user interface.
	 */
	void hideCodes();
	/**
	 * @brief Shows the currency code selection GUI element.
	 *
	 * @post This the visibility of the currency code GUI element is shown on the user interface.
	 */
	void showCodes();
	/**
	 * @brief Retrieves the current code from the code GUI element.
	 *
	 * @return The currently selected currency code.
	 */
	[[nodiscard]]
	pecunia::Currency code() const;
	/**
	 * @brief Sets the widget to a read-only or editable state.
	 *
	 * @param isReadOnly A boolean flag indicating whether the widget should be read-only (true) or
	 * editable (false).
	 *
	 * @post If set to read-only, the amount input is disabled, and the currency code selection is
	 * disabled. Otherwise, both are enabled.
	 */
	void setReadOnly(const bool isReadOnly);
	/**
	 * @brief Controls the visibility of the amount label.
	 *
	 * @param shouldShow A boolean flag indicating whether the label should be shown or hidden.
	 *
	 * @post When the flag is true, the labeling is shown, else it's hidden.
	 */
	void showLabel(const bool shouldShow);
	/**
	 * @brief Retrieves the textual contents of the monetary amount GUI element.
	 *
	 * @return Textual representation of the amount, with no validation.
	 */
	[[nodiscard]]
	QString textAmount() const;
	/**
	 * @brief Clears the money entry fields within the widget.
	 *
	 * @post The monetary amount and currency GUI elements are changed to an empty state.
	 */
	void clear();
	/// An accessor that, when true, shows if the widget is read-only.
	[[nodiscard]]
	bool isReadOnly() const;
	/**
	 * @brief Enables the currency selection GUI element in the money entry widget.
	 */
	void enableCurrencySelection();
	/**
	 * @brief Disables the currency selection GUI element.
	 */
	void disableCurrencySelection();
	/**
	 * @brief Determines if the money as entered in the widget is valid, i.e. no errors.
	 *
	 * @return A null optional when the money is valid, else the error text of why the money is not
	 * valid.
	 */
	[[nodiscard]]
	std::optional<::QString> amountErrorMessage() const noexcept;
};

}

#endif
