/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_FILE_STORAGE_BUDGET_FILE_H_
#define DRN_FILE_STORAGE_BUDGET_FILE_H_

#include <QString>

#include <map>

#include <file-storage_export.h>
#include <storage/BudgetStorage.hpp>


namespace drn::accounting
{

class AccountCode;
class AccountNumber;
class GeneralLedger;

}
namespace drn::banking
{

enum class SupportedAccountTypes;
struct BankMap;
class BankName;

}
namespace drn::budgeting
{

class Budget;
class BudgetSource;

}
namespace drn::foundation
{

template<typename, typename>
class TypeIndexMap;

}
namespace drn::storage
{

struct BudgetLoadValues;

}
namespace drn::surveying
{

class CompletedSurveys;

}
namespace drn::file_storage
{

/**
 * @brief The manager for interacting with a budget file.
 */
class DRN_FILE_STORAGE_EXPORT BudgetFile final : public storage::BudgetStorage
{
	QString directoryPath_;
	QString name_;
	QString backUpExtension_;

public:
	/// The default extension used when backing-up a budget file, on systems like KDE this will show
	/// a nice back-up icon.
	static const QChar defaultBackUpExtension_;
	/// The extension used when backing-up a back-up to a budget file.
	static const QString oldBackUpExtension_;

	BudgetFile();
	/**
	 * @brief Initialises the entire object using the supplied values.
	 *
	 * @exception Error When the supplied file location is empty.
	 * @exception Error When the supplied file location is not absolute.
	 * @exception Error When the supplied back-up extension starts with the full stop file
	 * separator ('.').
	 *
	 * @param fileLocation The file name and absolute directory path of the budget file. The text
	 * after the last directory separator is assumed to be the name of the budget file.
	 * @param backUpExtension (Optional) The extension to use when backing-up a pre-existing budget
	 * file. When this is not the default, it must not start with the full stop.
	 */
	explicit BudgetFile(
		const QString& fileLocation,
		const QString& backUpExtension = BudgetFile::defaultBackUpExtension_
	);
	/// The absolute directory path of the budget file.
	[[nodiscard]]
	const QString& directoryPath() const;
	/// The file name of the budget file.
	[[nodiscard]]
	const QString& name() const noexcept override;
	/// The extension to use when backing-up a budget file.
	[[nodiscard]]
	const QString& backUpExtension() const;
	/**
	 * @brief Changes the location of the budget file to load.
	 *
	 * @exception StorageError When the location is invalid.
	 *
	 * @param fileLocation The directory path and file-name of the budget file to use.
	 */
	void changeLocation(const ::QString& fileLocation) override;
	/**
	 * @brief Loads the budget file from the local file-system.
	 *
	 * @exception StorageError When the budget file cannot be found.
	 *
	 * @return The values loaded for the budget file.
	 */
	[[nodiscard]]
	storage::BudgetLoadValues load() const override;
	/**
	 * @brief Creates or overwrites the budget file to the local file-system. When overwriting the
	 * budget file, a back-up copy is created in the same directory.
	 *
	 * @exception StorageError When the budget file cannot be saved.
	 * @exception StorageError When the budget file cannot be backed-up prior to saving.
	 *
	 * @param[in] gl The general ledger whose contents are to be stored in the file.
	 * @param[in] b The budget whose contents are to be stored in the file.
	 * @param[in] budgetCodes The mapping between each budget item and its ledger account number.
	 * @param[in] banks The financial institutions whose contents are to be stored in the file.
	 * @param[in] bankAccountTypes The associations between the financial institutions and the
	 * general ledger whose contents are to be stored in the file.
	 */
	void store(
		const accounting::GeneralLedger& gl,
		const budgeting::Budget& b,
		const foundation::TypeIndexMap<
			budgeting::BudgetSource,
			accounting::AccountNumber
		>& budgetCodes,
		const banking::BankMap& banks,
		const std::map<
			banking::BankName,
			std::map<accounting::AccountNumber, banking::SupportedAccountTypes>
		>& bankAccountTypes,
		const surveying::CompletedSurveys& surveys
	) const override;
};

}

#endif
