/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_FOUNDATION_CALENDAR_TIME_H_
#define DRN_FOUNDATION_CALENDAR_TIME_H_

#include <chrono>

#include <foundation_export.h>


class QDate;

namespace drn::foundation
{

/**
 * @brief A duration type representing partial (fractional) days.
 *
 * This type alias defines a std::chrono::duration that can represent fractional days. The ratio
 * represents the number of seconds in a day (24 hours * 60 minutes * 60 seconds = 86400 seconds).
 *
 * @code
 * PartialDays days{2.5}; // Represents two and a half days
 * @endcode
 *
 * @note The underlying representation uses double for storing the number of days,
 * allowing for precise fractional day values.
 */
struct DRN_FOUNDATION_EXPORT PartialDays : std::chrono::duration<double, std::ratio<86400>>
{
	using std::chrono::duration<double, std::ratio<86400>>::duration;
};

/// The total number of hours in a day.
inline constexpr std::chrono::hours hoursInDay{24};
/// The total number of days in a standard Gregorian calendar year.
inline constexpr std::chrono::days daysInWeek{7};
/// The total number of hours in a standard Gregorian calendar year.
inline constexpr std::chrono::hours hoursInWeek{hoursInDay.count() * daysInWeek.count()};
/// The total number of months in a standard Gregorian calendar year.
inline constexpr std::chrono::months monthsInYear{12};
/**
 * @brief The average number of days per year, accounting for leap years, in a standard Gregorian
 * calendar year.
 */
inline constexpr PartialDays daysInYear{365.2425};
/**
 * @brief The average number of days per month, accounting for leap years, in a standard Gregorian
 * calendar year.
 */
inline constexpr PartialDays daysInMonth{daysInYear.count() / monthsInYear.count()};

/**
 * @brief Calculates the first day of the month for a given date.
 *
 * This function is useful when operations require aligning a date to the start of its month, such
 * as calculating monthly periods, budgeting, or generating reports.
 *
 * @pre The input date must be valid.
 *
 * @param date The input date from which the first day of the month is derived.
 *
 * @return A new date set to the first day of the same month and year as the provided input date.
 */
[[nodiscard]]
DRN_FOUNDATION_EXPORT ::QDate firstOfMonth(const ::QDate& date) noexcept;
/**
 * @brief Calculates the last day of the month for a given date.
 *
 * This function is useful in scenarios where date alignment to the end of the month is required,
 * such as calculating monthly report deadlines, financial summaries, or creating recurring
 * schedules.
 *
 * @pre The input date must be valid.
 *
 * @param date The input date from which the last day of the month is derived.
 *
 * @return A new date set to the last day of the same month and year as the provided input date.
 */
[[nodiscard]]
DRN_FOUNDATION_EXPORT ::QDate lastOfMonth(const ::QDate& date) noexcept;
/**
 * @brief Calculates the midpoint of the month for a given date.
 *
 * This function is useful when operations require determining a date approximately halfway
 * through the month, such as splitting periods into early/late segments, creating mid-month
 * reports, or sending out reminders.
 *
 * @pre The input date must be valid.
 *
 * @param date The input date from which the midpoint of the month is derived.
 *
 * @return A new date set to a day approximately halfway through the same month and year as the
 * provided input date.
 */
[[nodiscard]]
DRN_FOUNDATION_EXPORT ::QDate halfOfMonth(const ::QDate& date) noexcept;
/**
 * @brief Calculates the number of complete months between two dates.
 *
 * This function is ideal for determining the duration between two dates in terms of full months.
 *
 * @pre The starting date must be valid.
 * @pre The ending date must be valid.
 *
 * @param from The starting date for the calculation.
 * @param to The ending date for the calculation.
 *
 * @return The number of full months between the two dates. This will be negative if the `to` date
 * is earlier than the `from` date.
 */
[[nodiscard]]
DRN_FOUNDATION_EXPORT std::chrono::months monthsBetween(
	const ::QDate& from,
	const ::QDate& to
) noexcept;
/**
 * @brief Calculates the difference in days between two dates.
 *
 * This function is useful for determining the number of days between two points in time, which can
 * be used for scheduling, event planning, or calculating durations for reports or metrics.
 *
 * @pre The `from` input date must be valid.
 * @pre The `to` input date must be valid.
 *
 * @param from The starting date for the calculation.
 * @param to The ending date for the calculation.
 *
 * @return The number of days between the two dates. This will be negative if the `to` date is
 * earlier than the `from` date.
 */
[[nodiscard]]
DRN_FOUNDATION_EXPORT std::chrono::days daysBetween(const ::QDate& from, const ::QDate& to) noexcept;

}

#endif
