/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_FOUNDATION_ENUM_FLAGS_HPP_
#define DRN_FOUNDATION_ENUM_FLAGS_HPP_

/**
 * @file EnumFlags.hpp
 * @brief Provides utilities for handling enumeration flags and bitwise operations.
 * 
 * This file defines utilities that enable type-safe bitwise operations on enumeration types that
 * are registered as flags. It provides a mechanism to perform logical operations (OR, AND) on
 * enumeration values while maintaining type safety.
 * 
 * Key features:
 * - Type-safe bitwise operations for enumeration flags
 * - Registration mechanism for enum types to be used as flags
 * - Compile-time checking of registered flag types
 * 
 * @see REGISTER_ENUM_AS_FLAG
 */

#include <type_traits>

#include <foundation_export.h>


namespace drn::foundation::detail
{

template<typename EnumType>
concept EnumerationFlag = std::is_enum_v<EnumType> && ! std::is_convertible_v<EnumType, int>;

template<EnumerationFlag>
struct IsRegisteredEnumerationFlag : std::false_type
{};

}
namespace drn::foundation
{

/**
 * @brief Combines two sets of flags using a logical OR operation. This operation can be used to
 * create a new flag set that contains all flags present in at least one of the inputs.
 *
 * @pre The `FlagType` must be registered with the enumeration flag system.
 *
 * @tparam FlagType The type of enumeration that describes the possible flags.
 *
 * @param lhs The first set of flags to be included in the resulting combination.
 * @param rhs The second set of flags to be included in the resulting combination.
 *
 * @return The combined result flags containing all flags from both the input sets. The result
 * reflects the logical OR operation applied to the inputs.
 */
template<typename FlagType>
[[nodiscard]]
std::enable_if_t<detail::IsRegisteredEnumerationFlag<FlagType>::value, FlagType> operator|(
	const FlagType lhs,
	const FlagType rhs
) noexcept;
/**
 * @brief Computes the logical AND operation of two flags. This is used to evaluate the intersection
 * of two flags, where a flag is enabled only if it is set in both operands. It simplifies the
 * handling of bitwise operations when working with flags.
 *
 * @pre The `FlagType` must be registered with the enumeration flag system.
 *
 * @tparam FlagType The type of enumeration that describes the possible flags.
 *
 * @param lhs The first flag to be evaluated.
 * @param rhs The second flag to be evaluated.
 *
 * @return The resulting flag representing the intersection of the two input flags.
 */
template<typename FlagType>
[[nodiscard]]
std::enable_if_t<detail::IsRegisteredEnumerationFlag<FlagType>::value, FlagType> operator&(
	const FlagType lhs,
	const FlagType rhs
) noexcept;
/**
 * @brief Compares two flag values and determines if they are equal.
 *
 * @pre The `FlagType` must be registered with the enumeration flag system.
 *
 * @tparam FlagType The type of enumeration that describes the possible flags.
 *
 * @param lhs The first flag to be compared.
 * @param rhs The second flag to be compared.
 *
 * @return `true` if the flags are equal; `false` otherwise.
 */
template<typename FlagType>
[[nodiscard]]
std::enable_if_t<detail::IsRegisteredEnumerationFlag<FlagType>::value, bool> operator==(
	const FlagType lhs,
	const FlagType rhs
) noexcept;
/**
 * @brief Checks whether a specific flag is set within a given mask. This determines if a specified
 * flag is active based on the provided mask. It is typically used to evaluate configuration or
 * state flags where binary flags are combined using bitwise operations.
 *
 * @pre The `FlagType` must be registered with the enumeration flag system.
 *
 * @tparam FlagType The type of enumeration that describes the possible flags.
 *
 * @param flag The specific flag to check.
 * @param mask The mask against which the flag is evaluated.
 *
 * @return `true` if the flag is set within the provided mask, otherwise `false`.
 */
template<typename FlagType>
[[nodiscard]]
std::enable_if_t<detail::IsRegisteredEnumerationFlag<FlagType>::value, bool> isSet(
	const FlagType flag,
	const FlagType mask
) noexcept;


}

/**
 * @def REGISTER_ENUM_AS_FLAG(enumType)
 *
 * @brief Registers an enumeration type to be used with bitwise flag operations. This macro enables
 * the use of bitwise operators and flag checking functionality for the specified enumeration type.
 *
 * @param enumType The enumeration type to be registered as a flag type.
 *
 * @note The macro must be used outside of any namespace
 *
 * @pre The enumeration type must be a scoped enumeration (enum class).
 * @pre The enumeration values should be defined using power-of-two values to ensure proper bitwise
 * operations.
 *
 * @post The registered enumeration type will work with all the flag-related functions, e.g.
 * - Bitwise OR operator (@c |) for combining flags
 * - Bitwise AND operator (@c &) for intersecting flags
 * - @c isSet() function for checking if specific flags are set
 *
 * Example usage:
 * @code{.cpp}
 * namespace example
 * {
 * enum class Permissions
 * {
 *     None    = 0b000,
 *     Read    = 0b001,
 *     Write   = 0b010
 * };
 * }
 *
 * REGISTER_ENUM_AS_FLAG(example::Permissions);
 *
 * // After registration, you can use:
 * Permissions flags{Permissions::Read | Permissions::Write};
 * bool canRead{isSet(Permissions::Read, flags)};  // true
 * @endcode
 */
#define REGISTER_ENUM_AS_FLAG(enumType)  \
	template<>  \
	struct DRN_FOUNDATION_EXPORT drn::foundation::detail::IsRegisteredEnumerationFlag<enumType>  \
			: std::true_type  \
	{}

#include "EnumFlags.tpp"

#endif
