/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_FOUNDATION_ERROR_H_
#define DRN_FOUNDATION_ERROR_H_

#include <QString>

#include <stdexcept>

#include <foundation_export.h>


namespace drn::foundation
{

/**
 * @brief Represents an error condition with an associated message and optional supplementary data.
 *
 * This is designed to encapsulate error information in a unified structure. It allows for the
 * storage and retrieval of a descriptive error message, which can be used to communicate the nature
 * of the error to the user or another part of the system. Additionally, it supports optional
 * supplementary data to provide extra context when needed for debugging, logging, etc. This serves
 * as a foundation for custom error handling and can be extended for specific use cases.
 */
class DRN_FOUNDATION_EXPORT Error : public std::runtime_error
{
	::QString message_;

public:
	/**
	 * @brief Error message constructor.
	 *
	 * @param message The error message to be associated with this instance.
	 */
	explicit Error(const ::QString& message);
	/**
	 * @brief Error and supplementary context data constructor.
	 *
	 * The supplementary data can be especially helpful for diagnosing complex errors or passing
	 * crucial information to error-handling routines.
	 *
	 * @param message The error message to be associated with this instance.
	 * @param root The additional context information that describes the cause of the error.
	 */
	Error(const ::QString& message, const std::exception& root);

	explicit Error(const char*) = delete;

	explicit Error(const char*, const std::exception&) = delete;
	/**
	 * @brief An accessor to the error message.
	 *
	 * It is useful when the descriptive error message needs to be used, such as in logging,
	 * debugging, displaying error messages to the user, or capturing error information for further
	 * handling. This does not include any supplementary data or context, only the descriptive error
	 * messages.
	 *
	 * @return The error message describing the nature of this error.
	 */
	[[nodiscard]]
	const ::QString& message() const noexcept;
};

}

#endif
