/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_FOUNDATION_OPTIONAL_HPP_
#define DRN_FOUNDATION_OPTIONAL_HPP_

#include <iosfwd>
#include <optional>

#include <foundation_export.h>

class QString;


namespace std
{

/**
 * @brief Full object stream insertion operator.
 *
 * @note There is potential here for undefined behaviour, but the risk is worth taking to have a
 * uniform way of outputting an optional value.
 *
 * @tparam ValueType The type to place into the stream. This must have support for a stream
 * insertion operator.
 *
 * @param out The stream to place the data into.
 * @param opt The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
template<typename ValueType>
std::ostream& operator<<(std::ostream& out, const std::optional<ValueType>& opt); // NOLINT(*-dcl58-cpp)
/**
 * @brief Full object stream insertion operator.
 *
 * @note There is potential here for undefined behaviour, but the risk is worth taking to have a
 * uniform way of outputting an optional value.
 *
 * @param out The stream to place the data into.
 * @param opt The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
template<>
DRN_FOUNDATION_EXPORT std::ostream& operator<<(
	std::ostream& out,
	const std::optional<::QString>& opt
);

}

namespace drn::foundation
{

/**
 * @brief Converts a value to a human-readable text representation.
 *
 * This is useful for when you need to display an object's value to an end-user in a GUI element.
 *
 * @tparam ValueType The type to format for displaying. It must support a call to the function
 * `presentationText`.
 *
 * @param opt The value to format for displaying.
 *
 * @return The value formatted for displaying when present, else a translated text describing that
 * nothing was set.
 */
template<typename ValueType>
::QString presentationText(const std::optional<ValueType>& opt);
/**
 * @brief A convenience function to retrieve a value from an optional.
 *
 * @tparam ValueType The type to retrieve when present. The type must have a default constructor.
 *
 * @param opt The object containing the value to potentially retrieve.
 *
 * @return Gives the value stored in the object, else a default constructed value when not present.
 */
template<typename ValueType>
ValueType valueOrDefault(const std::optional<ValueType>& opt) noexcept;
/**
 * @brief A convenience function to retrieve a value from an optional when it's present. When a
 * value is not present, an exception is thrown.
 *
 * @tparam ValueType The type to retrieve from the optional.
 * @tparam ExceptionType The type of the exception to throw.
 * @tparam ExceptionArgs The argument types for the thrown exception type.
 *
 * @param value The object containing the value to potentially retrieve.
 * @param args The arguments for the thrown exception.
 *
 * @return Gives the value stored in the object.
 */
template<typename ExceptionType, typename ValueType, typename ... ExceptionArgs>
ValueType valueOrThrow(const std::optional<ValueType>& value, ExceptionArgs ... args);

}

#include "Optional.tpp"

#endif
