/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_FOUNDATION_PERCENTAGE_H_
#define DRN_FOUNDATION_PERCENTAGE_H_

#include <QtCore>

#include <compare>
#include <iosfwd>

#include <foundation_export.h>


class QDebug;
class QString;

namespace drn::foundation
{

/**
 * @brief Represents a percentage value for use in mathematical calculations or user interfaces.
 *
 * It is designed to simplify the representation and manipulation of fractional values expressed as
 * percentages. It provides utility methods for converting between percentages and decimal values,
 * as well as for formatting percentage values for display in graphical user interfaces or other
 * textual outputs.
 */
class DRN_FOUNDATION_EXPORT Percentage
{
	friend DRN_FOUNDATION_EXPORT std::ostream& operator<<(std::ostream&, const Percentage&);
	friend DRN_FOUNDATION_EXPORT ::QString presentationText(const Percentage&);

	quint8 precision_;
	double rate_;

public:
	/// N.B. The number of decimal places is chosen because it seems to be good enough for the
	/// number of digits most percentage rates are advertised in.
	static constexpr const quint8 defaultPrecision{6u};

	/**
	 * @brief Full initialisation constructor.
	 *
	 * @exception Error When the rate is too large to truncate.
	 *
	 * @param rate The percentage rate value. Any digits after the precision is truncated.
	 * @param precision The number of digits after the decimal point.
	 */
	explicit Percentage(
		const double rate = 0.0,
		const quint8 precision = Percentage::defaultPrecision
	);
	/**
	 * @brief Conversion operator to the decimal representation of the percentage.
	 *
	 * @return The current percentage rate.
	 */
	explicit operator double() const noexcept;
	[[nodiscard]]
	std::strong_ordering operator<=>(const Percentage& other) const noexcept;
	[[nodiscard]]
	std::strong_ordering operator<=>(const double rate) const noexcept;
	[[nodiscard]]
	bool operator==(const Percentage&) const noexcept = default;
	[[nodiscard]]
	bool operator==(const double rate) const noexcept;
	/// @brief An accessor to the number of digits to use in any operation involving the rate.
	[[nodiscard]]
	quint8 precision() const noexcept;
	/// @brief An accessor to the percentage rate.
	[[nodiscard]]
	double rate() const noexcept;
};

/**
 * @brief Full object stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param p The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_FOUNDATION_EXPORT std::ostream& operator<<(std::ostream& out, const Percentage& p);
/**
 * @brief Full object Qt stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param p The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_FOUNDATION_EXPORT ::QDebug operator<<(::QDebug out, const Percentage& p);
/**
 * @brief Generates a text representation for a percentage value.
 *
 * This function creates a formatted string displaying the percentage value based on the rate and
 * precision defined in the provided percentage object. It is useful for displaying percentage
 * values in a user-readable format, such as in GUIs, logs, or reports.
 *
 * @param p An object representing the percentage, containing the rate and precision.
 *
 * @return A string representing the percentage value, appended with a percentage symbol.
 */
DRN_FOUNDATION_EXPORT ::QString presentationText(const Percentage& p);

}

#endif
