/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_FOUNDATION_PRESENTATION_TEXT_H
#define DRN_FOUNDATION_PRESENTATION_TEXT_H

#include <QtGlobal>

#include <iosfwd>

#include <foundation_export.h>


class QDate;
class QDebug;
class QString;

namespace pecunia
{

class FloatingPoint;

}
namespace pecunia
{

enum class Currency : quint16;
class Money;

}

namespace drn::foundation
{

struct LocaleAware;
struct WithCurrency;
struct SpaceCurrency;
struct TruncateZeros;

/**
 * @brief Converts a supplied currency object into a human-readable presentation text.
 *
 * @param code The currency object to be converted into a human-readable text form.
 *
 * @return The human-readable representation of the currency value which makes it suitable for
 * display, typically in a locale-aware format, with the currency symbol spaced appropriately.
 */
DRN_FOUNDATION_EXPORT ::QString presentationText(const pecunia::Currency& code);
/**
 * @brief Converts a supplied currency object into a human-readable presentation text with optional
 * spacing. This method formats the given currency object into a string representation that is
 * suitable for display purposes. It accounts for optional spacing based on the provided parameters,
 * making it adaptable for different localization needs.
 *
 * @param code The currency object to be converted into a human-readable text form.
 * @param shouldBeSpaced Indicates whether a space should be inserted between the currency symbol
 * and the amount.
 *
 * @return The human-readable representation of the currency, optionally including a space for
 * better readability.
 */
DRN_FOUNDATION_EXPORT ::QString presentationText(
	const pecunia::Currency& code,
	const SpaceCurrency& shouldBeSpaced
);
/**
 * @brief Converts a supplied money object into its human-readable presentation text.
 *
 * @param money The object to be converted into human-readable text form.
 *
 * @return The human-readable representation of the money value in a format suitable for display
 * with a locale-aware representation, and without its currency.
 */
DRN_FOUNDATION_EXPORT ::QString presentationText(const pecunia::Money& money);
/**
 * @brief Converts a supplied money object into its human-readable presentation text.
 *
 * @param money The object to be converted into human-readable text form.
 * @param shouldAddCurrency Specifies if the currency symbol should be included in the output.
 *
 * @return The human-readable representation of the money value, formatted according to the
 * specified parameters. The decimal point is locale aware, and the currency is spaced.
 */
DRN_FOUNDATION_EXPORT ::QString presentationText(
	const pecunia::Money& money,
	const WithCurrency& shouldAddCurrency
);
/**
 * @brief Converts a supplied money object into its human-readable presentation text.
 *
 * @param money The object to be converted into human-readable text form.
 * @param isLocaleAware Specifies if a locale-specific decimal point should be used.
 * @param shouldAddCurrency Specifies if the currency symbol should be included in the output.
 *
 * @return The human-readable representation of the money value, formatted according to the
 * specified parameters. The currency is spaced.
 */
DRN_FOUNDATION_EXPORT ::QString presentationText(
	const pecunia::Money& money,
	const LocaleAware& isLocaleAware,
	const WithCurrency& shouldAddCurrency
);
/**
 * @brief Converts a supplied money object into its human-readable presentation text.
 *
 * @param money The object to be converted into human-readable text form.
 * @param shouldAddCurrency Specifies if the currency symbol should be included in the output.
 * @param shouldBeSpaced Specifies if a space should be added between the amount, and the currency
 * symbol.
 *
 * @return The human-readable representation of the money value, formatted according to the
 * specified parameters. The decimal point is locale aware.
 */
DRN_FOUNDATION_EXPORT ::QString presentationText(
	const pecunia::Money& money,
	const WithCurrency& shouldAddCurrency,
	const SpaceCurrency& shouldBeSpaced
);
/**
 * @brief Converts the given money object into a human-readable presentation text based on various
 * formatting options.
 *
 * @param money The object representing the monetary value to be converted.
 * @param isLocaleAware Determines if the conversion should consider locale-aware formatting for
 * decimal points.
 * @param shouldAddCurrency Specifies whether the currency symbol should be included in the output
 * string.
 * @param shouldBeSpaced Determines if there should be a space between the currency symbol and the
 * monetary value.
 * @param shouldTruncate Indicates whether trailing zeros should be removed from the formatted
 * string.
 *
 * @return The formatted human-readable representation of the monetary value considering the given
 * options.
 */
DRN_FOUNDATION_EXPORT ::QString presentationText(
	const pecunia::Money& money,
	const LocaleAware& isLocaleAware,
	const WithCurrency& shouldAddCurrency,
	const SpaceCurrency& shouldBeSpaced,
	const TruncateZeros& shouldTruncate
);
/**
 * @brief Converts a given object into a human-readable presentation text.
 *
 * @param value The object to be converted into a human-readable text form.
 *
 * @return The human-readable representation of the date, typically formatted
 * according to local conventions.
 */
DRN_FOUNDATION_EXPORT ::QString presentationText(const ::QDate& value);
/**
 * @brief Converts a signed 8-bit integer value into a string representation.
 *
 * @param value The integer value that needs to be converted into a string.
 *
 * @return A string that represents the integer value, making it suitable for display
 * or other textual operations that require a string format.
 */
DRN_FOUNDATION_EXPORT ::QString presentationText(const ::qint8& value);
/**
 * @brief Converts a unsigned 8-bit numeric value into a human-readable presentation text.
 *
 * @param value The numeric value to be converted into a human-readable text form.
 *
 * @return The human-readable text representation of the numeric value, suitable for display
 * purposes.
 */
DRN_FOUNDATION_EXPORT ::QString presentationText(const ::quint8& value);
/**
 * @brief Converts a floating-point number into a human-readable presentation text for display in
 * GUI elements.
 *
 * This function formats a floating-point value into a locale-appropriate textual representation,
 * ensuring it is displayed with a fixed precision suitable for graphical user interfaces or other
 * display contexts.
 *
 * @param value The floating-point value to be converted into a formatted text representation.
 *
 * @return A formatted string representation of the provided floating-point value, typically
 * displayed with a fixed precision.
 */
DRN_FOUNDATION_EXPORT ::QString presentationText(const pecunia::FloatingPoint& value);

}

/**
 * @brief Stream insertion operator for a Qt string into the output buffer.
 *
 * @param out The output buffer to place the value into.
 * @param value The contents to place into the output buffer.
 *
 * @return The output buffer post modification.
 */
DRN_FOUNDATION_EXPORT std::ostream& operator<<(std::ostream& out, const ::QString& value);
/**
 * @brief Stream insertion operator for a Qt date into the output buffer.
 *
 * @param out The output buffer to place the value into.
 * @param value The contents to place into the output buffer.
 *
 * @return The output buffer post modification.
 */
DRN_FOUNDATION_EXPORT std::ostream& operator<<(std::ostream& out, const ::QDate& value);
/**
 * @brief Stream insertion operator for a currency into the debug buffer.
 *
 * @param out The debug buffer to place the value into.
 * @param value The contents to place into the output buffer.
 *
 * @return The debug buffer post modification.
 */
DRN_FOUNDATION_EXPORT ::QDebug operator<<(::QDebug out, const pecunia::Currency& value);
/**
 * @brief Stream insertion operator for a money amount into the debug buffer.
 *
 * @param out The debug buffer to place the value into.
 * @param value The contents to place into the output buffer.
 *
 * @return The debug buffer post modification.
 */
DRN_FOUNDATION_EXPORT ::QDebug operator<<(::QDebug out, const pecunia::Money& value);

#endif
