/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_FOUNDATION_QT_MEMORY_HPP_
#define DRN_FOUNDATION_QT_MEMORY_HPP_

#include <QMetaObject>
#include <QObject>

#include <memory>
#include <type_traits>

#include <foundation/ObserverPtr.hpp>


//{ Non-exhaustive list of non-QObject Derived Qt types

class QFormLayout;
class QHBoxLayout;
class QListWidgetItem;
class QSpacerItem;
class QTableWidgetItem;
class QTreeWidgetItem;
class QVBoxLayout;

//}

namespace drn::foundation::internal
{

/**
 * @brief Represents the idea of a Qt object that derives from QObject.
 *
 * @tparam ObjectType The type to inspect if it's derived from QObject.
 */
template<typename ObjectType>
concept IsQObjectBase = std::is_base_of_v<::QObject, std::remove_pointer_t<ObjectType>>;
/**
 * @brief Represents the idea of a Qt type whose memory is handled by the parent-child pattern.
 *
 * @tparam ObjectType The type to inspect if it's manageable by Qt.
 */
template<typename ObjectType>
concept IsQtParentableObject = IsQObjectBase<std::remove_pointer_t<ObjectType>>
		|| std::is_same_v<::QListWidgetItem, std::remove_pointer_t<ObjectType>>
		|| std::is_same_v<::QTreeWidgetItem, std::remove_pointer_t<ObjectType>>;
/**
 * @brief Represents the idea of any Qt public type.
 *
 * @tparam ObjectType The type to inspect if it's created by the Qt SDK.
 */
template<typename ObjectType>
concept IsQtType =
	IsQObjectBase<ObjectType>
		|| std::is_same_v<::QFormLayout, std::remove_pointer_t<ObjectType>>
		|| std::is_same_v<::QHBoxLayout, std::remove_pointer_t<ObjectType>>
		|| std::is_same_v<::QListWidgetItem, std::remove_pointer_t<ObjectType>>
		|| std::is_same_v<::QSpacerItem, std::remove_pointer_t<ObjectType>>
		|| std::is_same_v<::QTableWidgetItem, std::remove_pointer_t<ObjectType>>
		|| std::is_same_v<::QTreeWidgetItem, std::remove_pointer_t<ObjectType>>
		|| std::is_same_v<::QVBoxLayout, std::remove_pointer_t<ObjectType>>;

template<typename QtPtrType>
requires requires (QtPtrType p)
{
	typename QtPtrType::element_type;
	p == nullptr;
	{ p.get() } -> std::same_as<typename QtPtrType::element_type*>;
	p = nullptr;
}
QMetaObject::Connection createDestroyableConnection(QtPtrType& self);

}
namespace drn::foundation
{

template<typename>
class UniqueQtPtr;

/**
 * @brief A smart pointer class ensuring the Qt reference counting for automatic memory management.
 * QtPtr is a template-based smart pointer class that ensures Qt's reference counting manages the
 * lifetime of a dynamically allocated object. When the parent owning an object is destroyed or
 * reset, the managed object is automatically deleted. Additionally, when the observed object is
 * destructed, the object observing will turn null.
 *
 * @tparam ObjectType The type of the object stored.
 */
template<typename ObjectType>
class QtPtr : public ObserverPtr<ObjectType>
{
	QMetaObject::Connection connection_;

public:
	/**
	 * @brief Constructs a QtPtr that observes a given Qt object.
	 *
	 * @param observe The Qt object to be observed.
	 */
	QtPtr(ObjectType* const observe = nullptr) requires internal::IsQtParentableObject<ObjectType>; // NOLINT(*-explicit-constructor)
	/**
	 * @brief Constructs a QtPtr that observes a given UniqueQtPtr object.
	 *
	 * @param observe The unique Qt object to be observed.
	 */
	QtPtr(const UniqueQtPtr<ObjectType>& observe); // NOLINT(*-explicit-constructor)

	QtPtr(const QtPtr&) = default;

	QtPtr(QtPtr&&) noexcept = default;

	QtPtr& operator=(const QtPtr&) = default;

	QtPtr& operator=(QtPtr&&) noexcept = default;
	/**
	 * @brief Disconnects the Qt connection associated with the pointer if the observed object is
	 * derived from QObject.
	 */
	~QtPtr();
	/**
	 * @brief Releases the ownership of the Qt object.
	 *
	 * The method removes the Qt object's association with its parent, preventing Qt's internal
	 * memory management from automatically deleting the object when the parent is destroyed. The
	 * caller is responsible for managing the memory.
	 *
	 * @post The object is disassociated from its parent in the case of Qt objects.
	 *
	 * @return The object that was released.
	 *
	 * @warning The caller is responsible for managing the memory of the object that is returned.
	 */
	UniqueQtPtr<ObjectType> release();

	using ObserverPtr<ObjectType>::get;

	using ObserverPtr<ObjectType>::operator*;

	using ObserverPtr<ObjectType>::operator->;

	using ObserverPtr<ObjectType>::operator==;

	using ObserverPtr<ObjectType>::operator!=;

	using ObserverPtr<ObjectType>::operator<;
};
/**
 * @brief A representation of a Qt object that is not handled by the usual Qt memory pattern. When
 * the Qt object derives from a QObject, its deletion is handled on the event loop else it’s deleted
 * immediately upon destruction.
 *
 * @tparam ObjectType The type of the object stored.
 */
template<typename ObjectType>
class UniqueQtPtr final : public std::unique_ptr<ObjectType, void(*)(ObjectType*)>
{
public:
	/**
	 * @brief Constructs an object to manage the given Qt object.
	 *
	 * @pre The pointer to the Qt object to be managed cannot have a parent.
	 *
	 * @param obj A pointer to the Qt object to be managed.
	 */
	explicit UniqueQtPtr(ObjectType* const obj = nullptr) requires internal::IsQtType<ObjectType>;
	/**
	 * @brief Constructs an object with a nullptr, initialising it to manage no object. This is
	 * useful when you need to explicitly initialise a pointer to a null state, ensuring that it
	 * does not point to any Qt object initially.
	 */
	UniqueQtPtr(std::nullptr_t); // NOLINT(*-explicit-constructor)
	using std::unique_ptr<ObjectType, void(*)(ObjectType*)>::operator*;
	using std::unique_ptr<ObjectType, void(*)(ObjectType*)>::operator->;
	using std::unique_ptr<ObjectType, void(*)(ObjectType*)>::get;
	using std::unique_ptr<ObjectType, void(*)(ObjectType*)>::release;
	using std::unique_ptr<ObjectType, void(*)(ObjectType*)>::swap;
	/**
	 * @brief Equality operator for checking null.
	 *
	 * @return True when the pointer is null, false otherwise.
	 */
	bool operator==(std::nullptr_t) const noexcept;
	/**
	 * @brief Inequality operator for checking null.
	 *
	 * @return False when the pointer is null, true otherwise.
	 */
	bool operator!=(std::nullptr_t) const noexcept;
};

/**
 * @brief Creates an object using the Qt memory pattern. Guarantees the object is of a Qt type that
 * can participate in the Qt memory management and is owned by a parent.
 *
 * @tparam ObjectType The type of the object to create.
 * @tparam ObjectArgTypes The argument types for the object constructor to invoke.
 *
 * @param objectArgs The values of to use in the object creation.
 *
 * @return The Qt object that is created on the heap.
 */
template<typename ObjectType, typename ... ObjectArgTypes>
requires internal::IsQtParentableObject<ObjectType>
QtPtr<ObjectType> makeQtPtr(ObjectArgTypes&& ... objectArgs);
/**
 * @brief Creates an object using the Qt memory pattern. Guarantees the object is of a Qt type that
 * can participate in the Qt memory management and is owned by a parent.
 *
 * @tparam ObjectType The type of the object to create.
 * @tparam BaseType The base type of the object being created.
 * @tparam ObjectArgTypes The argument types for the object constructor to invoke.
 *
 * @param objectArgs The values of to use in the object creation.
 *
 * @return The Qt object that is created on the heap.
 */
template<typename ObjectType, typename BaseType, typename ... ObjectArgTypes>
requires internal::IsQtParentableObject<ObjectType>
QtPtr<BaseType> makeBaseQtPtr(ObjectArgTypes&& ... objectArgs);
/**
 * @brief Creates a Qt object that is not handled by the Qt memory pattern.
 *
 * @tparam ObjectType The type of the object to create.
 * @tparam ObjectArgTypes The argument types for the object constructor to invoke.
 *
 * @param objectArgs The values of to use in the object creation.
 *
 * @return The Qt object that is created on the heap.
 */
template<typename ObjectType, typename ... ObjectArgTypes>
requires internal::IsQtType<ObjectType>
UniqueQtPtr<ObjectType> makeUniqueQtPtr(ObjectArgTypes&& ... objectArgs);
/**
 * @brief Creates a Qt object that is not handled by the Qt memory pattern.
 *
 * @tparam ObjectType The type of the object to create.
 * @tparam BaseType The base type of the object being created.
 * @tparam ObjectArgTypes The argument types for the object constructor to invoke.
 *
 * @param objectArgs The values of to use in the object creation.
 *
 * @return The Qt object that is created on the heap.
 */
template<typename ObjectType, typename BaseType, typename ... ObjectArgTypes>
requires internal::IsQtType<ObjectType> && internal::IsQtType<BaseType>
UniqueQtPtr<BaseType> makeBaseUniqueQtPtr(ObjectArgTypes&& ... objectArgs);
/**
 * Equality operator for comparing against a naked type.
 *
 * @tparam ObservedType The type of the naked object to compare.
 * @tparam  ObservedPtrType The type of the Qt object to compare.
 *
 * @param lhs The naked pointer to compare to.
 * @param rhs The pointer to compare.
 *
 * @return True when the two supplied pointers point to the same object, false otherwise.
 */
template<typename ObservedType, typename ObservedPtrType>
bool operator==(const ObservedType* lhs, const QtPtr<ObservedPtrType>& rhs);
/**
 * @brief Detaches the given Qt pointer from its parent, setting its parent to nullptr.
 *
 * @tparam QtPtrType The type of the object to remove the parent.
 *
 * @param ptr The Qt pointer to be de-parented.
 *
 * @return The same Qt pointer passed in, but with its parent set to nullptr.
 */
template<typename QtPtrType>
requires internal::IsQtParentableObject<QtPtrType> && std::is_pointer_v<QtPtrType>
QtPtrType deparent(QtPtrType ptr);

}

#include "QtMemory.tpp"

#endif
