/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_FOUNDATION_QT_MEMORY_HPP_
#error This should never be included directly.
#endif

#include <cassert>

template<typename QtPtrType>
requires requires (QtPtrType p)
{
	typename QtPtrType::element_type;
	p == nullptr;
	{ p.get() } -> std::same_as<typename QtPtrType::element_type*>;
	p = nullptr;
}
QMetaObject::Connection drn::foundation::internal::createDestroyableConnection(QtPtrType& self)
{
	if (self.get() == nullptr)
		return {};

	if constexpr (internal::IsQObjectBase<typename QtPtrType::element_type*>)
	{
		return ::QObject::connect(
			self.get(),
			&::QObject::destroyed,
			[&self] ([[maybe_unused]] auto obj)
			{
				assert(self.get() == obj && "Should be the same object being destroyed.");
				self = nullptr;
			}
		);
	}
	return {};
}

template<typename ObjectType>
drn::foundation::QtPtr<ObjectType>::QtPtr(ObjectType* const observe)
		requires drn::foundation::internal::IsQtParentableObject<ObjectType> :
	ObserverPtr<ObjectType>{observe},
	connection_{internal::createDestroyableConnection(*this)}
{}

template<typename ObjectType>
drn::foundation::QtPtr<ObjectType>::QtPtr(const UniqueQtPtr<ObjectType>& observe) :
	drn::foundation::QtPtr<ObjectType>{observe.get()}
{}

template<typename ObjectType>
drn::foundation::QtPtr<ObjectType>::~QtPtr()
{
	if constexpr (internal::IsQObjectBase<ObjectType>)
		QObject::disconnect(this->connection_);
}

template<typename ObjectType>
drn::foundation::UniqueQtPtr<ObjectType> drn::foundation::QtPtr<ObjectType>::release()
{
	if constexpr (internal::IsQObjectBase<ObjectType>)
		deparent(this->get());
	return UniqueQtPtr{this->get()};
}

template<typename ObjectType>
drn::foundation::UniqueQtPtr<ObjectType>::UniqueQtPtr(ObjectType* const obj)
		requires internal::IsQtType<ObjectType> :
	std::unique_ptr<ObjectType, void(*)(ObjectType*)>{
		obj,
		[] (ObjectType* o)
		{
			if constexpr (internal::IsQObjectBase<ObjectType>)
			{
				if (o != nullptr)
					o->deleteLater();
			}
			else
				delete o;
		}
	}
{
	if constexpr (internal::IsQObjectBase<ObjectType>)
	{
		if (obj != nullptr)
			assert(obj->parent() == nullptr && "All unique Qt objects must not have a parent.");
	}
}

template<typename ObjectType>
drn::foundation::UniqueQtPtr<ObjectType>::UniqueQtPtr(std::nullptr_t) :
	UniqueQtPtr(static_cast<ObjectType*>(nullptr))
{}

template<typename ObjectType>
bool drn::foundation::UniqueQtPtr<ObjectType>::operator==(std::nullptr_t) const noexcept
{
	return this->get() == nullptr;
}

template<typename ObjectType>
bool drn::foundation::UniqueQtPtr<ObjectType>::operator!=(std::nullptr_t) const noexcept
{
	return ! (*this == nullptr);
}

template<typename ObjectType, typename ... ObjectArgTypes>
requires drn::foundation::internal::IsQtParentableObject<ObjectType>
drn::foundation::QtPtr<ObjectType> drn::foundation::makeQtPtr(ObjectArgTypes&& ... objectArgs)
{
	return makeBaseQtPtr<ObjectType, ObjectType>(std::forward<ObjectArgTypes>(objectArgs)...);
}

template<typename ObjectType, typename BaseType, typename ... ObjectArgTypes>
requires drn::foundation::internal::IsQtParentableObject<ObjectType>
drn::foundation::QtPtr<BaseType> drn::foundation::makeBaseQtPtr(ObjectArgTypes&& ... objectArgs)
{
	QtPtr<BaseType> obj{new ObjectType{std::forward<ObjectArgTypes>(objectArgs)...}};

	// BUG: Workaround for the QTreeWidgetItem parent call returning null even though the supplied parent is not null with Qt 5.15.
	if constexpr (std::is_same_v<::QObject, BaseType>)
		assert(obj->parent() != nullptr && "All Qt objects must have a parent.");
	return obj;
}

template<typename ObjectType, typename ... ObjectArgTypes>
requires drn::foundation::internal::IsQtType<ObjectType>
drn::foundation::UniqueQtPtr<ObjectType> drn::foundation::makeUniqueQtPtr(
	ObjectArgTypes&& ... objectArgs
)
{
	return makeBaseUniqueQtPtr<ObjectType, ObjectType>(std::forward<ObjectArgTypes>(objectArgs)...);
}

template<typename ObjectType, typename BaseType, typename ... ObjectArgTypes>
requires drn::foundation::internal::IsQtType<ObjectType>
		&& drn::foundation::internal::IsQtType<BaseType>
drn::foundation::UniqueQtPtr<BaseType> drn::foundation::makeBaseUniqueQtPtr(
	ObjectArgTypes&& ... objectArgs
)
{
	UniqueQtPtr<BaseType> ptr{new ObjectType{std::forward<ObjectArgTypes>(objectArgs)...}};
	return ptr;
}

template<typename ObservedType, typename ObservedPtrType>
bool drn::foundation::operator==(
	const ObservedType* lhs,
	const drn::foundation::QtPtr<ObservedPtrType>& rhs
)
{
	return lhs == static_cast<ObserverPtr<ObservedPtrType>>(rhs);
}

template<typename QtPtrType>
requires drn::foundation::internal::IsQtParentableObject<QtPtrType> && std::is_pointer_v<QtPtrType>
QtPtrType drn::foundation::deparent(QtPtrType ptr)
{
	ptr->setParent(nullptr);
	return ptr;
}
