/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_FOUNDATION_STRONG_TYPE_DEF_HPP
#define DRN_FOUNDATION_STRONG_TYPE_DEF_HPP

#include <type_traits>
#include <utility>


namespace drn::foundation
{

template <typename>
class StrongTypeDef;

/**
 * @brief Creates a strong type that behaves like its underlying weak type form. This is useful for
 * cases where only the type needs to be differentiated, but all behaviour is identical.
 *
 * @tparam WeakType The underlying type the strong type is differentiating.
 */
template <typename WeakType>
requires std::is_fundamental_v<WeakType>
class StrongTypeDef<WeakType>
{

protected:
	WeakType underlying_;

public:
	/**
	 * @brief Constructor that performs default-initialization of the underlying weak type.
	 */
	StrongTypeDef() = default;
	/**
	 * @brief Full initialisation constructor.
	 *
	 * @param value The initial value of the underlying weak type.
	 */
	explicit StrongTypeDef(const WeakType value) :
		// Justification: MSVC cannot find the declaration/definition when separated.
		underlying_{value}
	{}
	/**
	 * @brief Conversion operator to the underlying weak type. N.B. intentionally left implicit to
	 * allow for interoperation.
	 */
	operator WeakType() const
	// Justification: MSVC cannot find the declaration/definition when separated.
	{
		return this->underlying_;
	}
};
/**
 * @brief Creates a strong type that behaves like its underlying weak type form. This is useful for
 * cases where only the type needs to be differentiated, but all behaviour is identical.
 *
 * @tparam WeakType The underlying type the strong type is differentiating.
 */
template <typename WeakType>
requires std::is_class_v<WeakType>
class StrongTypeDef<WeakType> : public WeakType
{
public:
	/**
	 * @brief Full initialisation constructor that initializes with the provided arguments.
	 *
	 * @tparam WeakTypeArguments The types of the arguments for the underlying weak type's
	 * constructor.
	 *
	 * @param args Arguments to forward to the constructor of the underlying weak type.
	 */
	template<typename ... WeakTypeArguments>
	explicit StrongTypeDef(WeakTypeArguments&& ... args) :
		// Justification: MSVC cannot find the declaration/definition when separated.
		WeakType{std::forward<WeakTypeArguments>(args)...}
	{}

};

/// @brief A type to represent any tagged Boolean type.
using TaggedBoolean = StrongTypeDef<bool>;

}

#include "StrongTypeDef.tpp"

#endif
