/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_NAVIGATION_BANK_ACCOUNT_CODE_HPP_
#define DRN_NAVIGATION_BANK_ACCOUNT_CODE_HPP_

#include <pecunia/Currency.hpp>

#include <compare>

#include <accounting/AccountCode.h>
#include <banking/BankAccount.h>
#include <navigation_export.h>


namespace drn::navigation
{

/**
 * Represents a bank account that allows showing account details, such as account number, balance
 * currency, and account holder information.
 */
class DRN_NAVIGATION_EXPORT BankAccountCode final
{
	banking::BankAccount bankAccount_;
	accounting::AccountCode code_;
	pecunia::Currency issuedIn_;

public:
	/**
	 * @brief Default constructor for the BankAccountCode class.
	 *
	 * This constructor initializes an instance of the BankAccountCode class
	 * with default values. No specific bank account, account code, or currency
	 * is assigned when this constructor is used.
	 */
	BankAccountCode();
	/**
	 * @brief Full constructor for a bank account.
	 *
	 * @exception Error If the account number of the bank account does not match the account code.
	 *
	 * @param bankAccount The bank account that is associated with the account code.
	 * @param code The account code that is associated with the bank account.
	 * @param issuedIn The currency in which the account is issued in.
	 */
	BankAccountCode(
		banking::BankAccount bankAccount,
		accounting::AccountCode code,
		pecunia::Currency issuedIn
	);
	/// An accessor to the name of the financial institution.
	[[nodiscard]]
	const banking::BankName& bank() const noexcept;
	/// An accessor to the financial institution account.
	[[nodiscard]]
	const accounting::AccountCode& account() const noexcept;
	/// An accessor to the financial institution's account type.
	[[nodiscard]]
	const banking::SupportedAccountTypes& supportedAccountType() const noexcept;
	/// An accessor to the limited information about the bank account.
	[[nodiscard]]
	const banking::BankAccountType& bankAccountType() const noexcept;
	/// An accessor to the currency used in the financial institution's account.
	[[nodiscard]]
	const pecunia::Currency& issuedIn() const noexcept;
	/**
	 * @brief A three-way comparator based upon the financial institution's account and account
	 * currency.
	 *
	 * @param other The other object to compare against.
	 *
	 * @return The relationship between the current object and the one supplied.
	 */
	[[nodiscard]]
	std::strong_ordering operator<=>(const BankAccountCode& other) const noexcept;
};

}

#endif
