/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_NAVIGATION_CONFIGURATION_OPTIONS_H_
#define DRN_NAVIGATION_CONFIGURATION_OPTIONS_H_

#include <compare>

#include <QStringList>

#include <foundation/StrongTypeDef.hpp>
#include <navigation_export.h>


namespace drn::navigation
{

/**
 * @brief The type to track how the application configures and enables the advanced UI features.
 *
 * This is used to initialize and enable advanced user interface features for the application. These
 * features can provide enhanced interactivity and functionality for end users by displaying
 * additional GUI elements and handling their integration within the application interface.
 */
struct DRN_NAVIGATION_EXPORT EnableAdvancedUi final : foundation::TaggedBoolean
{
	using foundation::TaggedBoolean::TaggedBoolean;
};
/**
 * @brief The type to track how the application configures whether specific accounting jargon should
 * be excluded from the application's GUI.
 *
 * This allows users to enable or disable the display of specialized accounting terminology within
 * the user interface. By disabling such jargon, the application can provide a more accessible and
 * easier-to-understand interface for users who are not familiar with accounting-specific
 * terminology.
 */
struct DRN_NAVIGATION_EXPORT DisableJargonUi final : foundation::TaggedBoolean
{
	using foundation::TaggedBoolean::TaggedBoolean;
};
/**
 * @brief The configurable options for the application's behavior.
 *
 * This is used to collect the user-specified options that define the behavior of the application
 * from the starting of the application.
 */
struct DRN_NAVIGATION_EXPORT ConfigurationOptions
{
	std::strong_ordering operator<=>(const ConfigurationOptions&) const noexcept = default;

	/**
	 * @brief Determines whether advanced UI features are enabled in the application.
	 *
	 * This specifies if the application should render advanced GUI elements as opposed to
	 * simplified ones. Enabling advanced UI provides additional functionality and interactivity
	 * within GUI items, designed for power users or users requiring enhanced features.
	 */
	EnableAdvancedUi isAdvancedUiEnabled_{false};
	/**
	 * @brief Configures whether accounting jargon is disabled in the application's GUI components.
	 *
	 * This determines whether the GUI should present content without any complex or specialized
	 * accounting jargon. When disabled, the user interface becomes more accessible and suitable for
	 * individuals without accounting expertise.
	 */
	DisableJargonUi isJargonUiDisabled_{false};
};

/**
 * @brief Parses arguments for the application.
 *
 * This processes the provided list of arguments and extracts user-specified options, such as GUI
 * preferences, for configuring the application's behavior.
 *
 * @param arguments A list of strings representing the arguments passed to the application.
 *
 * @return A structure containing the options extracted from the arguments.
 */
DRN_NAVIGATION_EXPORT ConfigurationOptions parse(const ::QStringList& arguments) noexcept;

}

#endif
