/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_NAVIGATION_DISPLAY_CURRENCY_H_
#define DRN_NAVIGATION_DISPLAY_CURRENCY_H_

#include <array>
#include <cstdint>
#include <set>

#include <navigation_export.h>


class QString;

namespace pecunia
{

enum class Currency : std::uint16_t;

}

namespace drn::navigation
{

/**
 * @brief Defines options for displaying currencies.
 *
 * This enumeration provides different ways to represent a currency visually. It is useful for
 * applications that need to display currency information in a user-friendly way or format text
 * based on user preferences.
 */
enum class DRN_NAVIGATION_EXPORT CurrencyDisplayOptions
{
	/// Display as the ISO-4217 currency code.
	IsoCode,
	/// Display as the currency symbol.
	Symbol,
	/// Display as both the currency code and symbol.
	SymbolAndCode,
};

/**
 * @brief Converts a currency display option into its corresponding text-based representation.
 *
 * This function maps a given currency display option into a textual description. It is useful
 * within applications that need to store, display, or communicate how a currency should be
 * presented (e.g., using ISO codes, symbols, or both).
 *
 * @exception std::invalid_argument Thrown if the given display option does not have
 * a corresponding text-based representation.
 *
 * @param displayAs An enumeration value representing how the currency should be displayed.
 *
 * @return A string representation of the provided CurrencyDisplayOptions value.
 */
DRN_NAVIGATION_EXPORT ::QString toQString(const CurrencyDisplayOptions& displayAs);
/**
 * @brief Converts a textual representation into a CurrencyDisplayOptions enumeration value.
 *
 * This function maps user-provided or system-provided textual representations of how a currency
 * should be displayed (e.g., "ISO-4217 Code", "Symbol") into the corresponding
 * CurrencyDisplayOptions value. It is useful when receiving input from GUI elements or external
 * sources and standardizing it for internal processing.
 *
 * @exception std::logic_error Thrown if the provided text cannot be mapped to any
 * CurrencyDisplayOptions value or if the input type fails to meet enumeration conversion
 * requirements.
 *
 * @param displayAs The textual representation of the target currency display format.
 *
 * @return The corresponding CurrencyDisplayOptions enumeration value for the provided text.
 */
DRN_NAVIGATION_EXPORT CurrencyDisplayOptions toCurrencyDisplayOptions(const ::QString& displayAs);
/**
 * @brief Converts a currency and its display option into a user-friendly string representation.
 *
 * This function generates a textual representation of a given currency based on the provided
 * display option. It is designed to accommodate different formats of currency visualization,
 * such as ISO codes, symbols, or a combination of both, and is particularly useful for GUI elements
 * or text-based output where clear and standardized currency display is required. If the display
 * option cannot be determined, the function throws an exception.
 *
 * @exception std::logic_error Thrown if the provided display option is not supported or if
 * the function logic fails to map the currency to a valid representation.
 *
 * @param code A reference to the currency object to be converted into a displayable string.
 * @param displayAs The preferred display option describing how the currency should be visualized.
 *
 * @return A string that represents the currency in the desired format, based on the provided
 * display option.
 */
DRN_NAVIGATION_EXPORT::QString toDisplayable(
	const pecunia::Currency& code,
	const CurrencyDisplayOptions displayAs
);
/**
 * @brief Converts a displayed currency representation into a specific currency code.
 *
 * This method is useful for interpreting a user-selected or user-provided text representation
 * of a currency, such as a symbol or ISO code, and mapping it to an actionable currency
 * identifier. It can be used in GUI elements where users input or interact with currencies.
 *
 * @exception logic_error Thrown if the input text could not be interpreted based on the display
 * format.
 * @exception logic_error Thrown if no matching currency is found for the given display options.
 *
 * @param text The displayed text representation of the currency (e.g., symbol or ISO code).
 * @param usableCodes A collection of currencies that are allowed for resolution.
 * @param displayAs Specifies the display option used for representing the currency.
 *
 * @return The resolved currency code corresponding to the provided display text.
 */
DRN_NAVIGATION_EXPORT pecunia::Currency toCurrencyFromDisplayable(
	const ::QString& text,
	const std::set<pecunia::Currency>& usableCodes,
	const CurrencyDisplayOptions displayAs
);
/**
 * @brief Converts a currency display option into its corresponding presentation text.
 *
 * This function is intended to generate a user-friendly textual representation of a currency
 * display option. It is designed for applications that need to display currency options in a GUI
 * element or other text-based contexts.
 *
 * @exception std::invalid_argument Thrown if an unidentified or unsupported currency display option
 * is passed.
 *
 * @param displayAs Specifies the desired currency display format.
 *
 * @return A text representation of the currency display format, formatted for display in GUI
 * elements or other user-facing contexts.
 */
DRN_NAVIGATION_EXPORT ::QString presentationText(const CurrencyDisplayOptions& displayAs);

/**
 * @brief Lists all possible currency display options.
 *
 * This variable provides a collection of all defined options for representing currencies in a
 * standardized way. It is utilized to ensure consistency when presenting currency data across
 * various components of an application.
 *
 * Use this to validate or iterate through the available representations of currency, ensuring no
 * unintentional omissions of any display options. It is particularly useful when creating
 * dropdowns, selectors, or validation mechanisms involving currency formats in GUI elements or
 * other programmatic use cases.
 */
inline constexpr std::array allCurrencyDisplayOptions{
	CurrencyDisplayOptions::IsoCode,
	CurrencyDisplayOptions::Symbol,
	CurrencyDisplayOptions::SymbolAndCode
};

}

#endif
