/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_NAVIGATION_NAVIGATOR_H_
#define DRN_NAVIGATION_NAVIGATOR_H_

#include <pecunia/Money.hpp>

#include <QObject>
#include <QString>

#include <chrono>
#include <map>
#include <memory>
#include <optional>
#include <set>
#include <vector>

#include <accounting/AccountCode.h>
#include <accounting/Account.h>
#include <accounting/LedgerBooks.h>
#include <accounting/TransactionNumber.h>
#include <adviser/BudgetHealth.h>
#include <banking/BankAccount.h>
#include <banking/BankAccountType.h>
#include <banking/BankMap.h>
#include <banking/BankName.h>
#include <banking/ReconciledBalance.h>
#include <budgeting/BillMap.hpp>
#include <budgeting/BudgetedMoney.h>
#include <budgeting/Budget.h>
#include <budgeting/BudgetItemDue.h>
#include <budgeting/BudgetItemIdentifier.h>
#include <budgeting/BudgetItemTypes.h>
#include <budgeting/BudgetSource.h>
#include <budgeting/DebtMap.h>
#include <budgeting/GoalMap.h>
#include <budgeting/NontrackMap.hpp>
#include <budgeting/WageMap.hpp>
#include <conversion/ConversionMap.h>
#include <conversion/CurrencySource.hpp>
#include <foundation/Percentage.h>
#include <foundation/QtMemory.hpp>
#include <navigation_export.h>
#include <navigation/BudgetWizardTemplates.h>
#include <networking/RequestManager.hpp>
#include <surveying/IndexedIncreaseDecreases.hpp>
#include <surveying/DistributedAmount.hpp>
#include <surveying/MappingSurvey.h>
#include <updating/ReleaseInformation.hpp>
#include <updating/Version.hpp>

#include "BankAccountCode.hpp"
#include "BudgetBankAccount.h"
#include "DisplayCurrency.h"
#include "UserPreferences.h"


namespace drn::storage
{

struct BudgetStorage;
struct ExchangeRatesStorage;

}
namespace drn::navigation::internal
{

class BudgetBankLedgers;

}
namespace drn::navigation
{

struct AllowSubAccounts;

class DRN_NAVIGATION_EXPORT Navigator final : public ::QObject
{
	Q_OBJECT

	void clear();
	void markCleared(
		const accounting::TransactionNumber& txnNum,
		const std::optional<drn::accounting::AccountNumber>& actNum
	);
	void markUnreconciled(const accounting::TransactionNumber& tn);

	std::optional<::QString> filePathName_;
	bool isDirty_;
	std::unique_ptr<internal::BudgetBankLedgers> budgetBankLedgers_;
	UserPreferences preferences_;
	foundation::QtPtr<networking::RequestManager> requestManager_;
	foundation::QtPtr<conversion::CurrencySource> exchangeSource_;
	std::unique_ptr<storage::BudgetStorage> budgetStore_;
	std::unique_ptr<storage::ExchangeRatesStorage> exchangeRatesStore_;

public slots:
	void onPrepareShowBills();
	void onPrepareShowDebts();
	void onPrepareShowGoals();
	void onPrepareShowNontracks();
	void onPrepareShowWages();
	/**
	 * @brief Adds a new budget item to the system based on its type, registers it with the
	 * appropriate ledger, marks the state as modified, and emits success or failure signals
	 * in response to the operation.
	 *
	 * This method handles the inclusion of a budget item into the system under various budget
	 * item categories. Each type of budget item requires specific logic to ensure proper handling
	 * and registry.
	 *
	 * @exception std::logic_error Thrown if the type of the budget item is unrecognized or invalid.
	 * @exception Error Thrown if there is a failure in processing the addition of the budget item.
	 *
	 * @param type The category of the budget item to be added.
	 * @param item The budget item object that is being added.
	 * @param ba An optional parameter that represents the associated bank account for the budget
	 * item, e.g. the bank account that stores a debt item's car loan.
	 *
	 * @post The budget item is processed and added to the appropriate ledger based on its type.
	 * @post The internal state is marked as modified to reflect changes.
	 * @post On success, the success signal `saveSucceeded` is emitted with the source of the added
	 * budget item.
	 * @post On failure, the failure signal `saveFailed` is emitted with the error details and the
	 * item's source.
	 *
	 * @return A boolean indicating whether the addition of the budget item was successful.
	 */
	bool onAdded(
		const drn::budgeting::BudgetItemTypes& type,
		const drn::budgeting::BudgetedMoney& item,
		const std::optional<drn::banking::BankAccount>& ba
	);
	void onRenamed(
		const drn::budgeting::BudgetItemTypes& type,
		const drn::budgeting::BudgetSource& original,
		const drn::budgeting::BudgetSource& changed
	);
	void onUpdated(
		const drn::budgeting::BudgetItemTypes& type,
		const drn::budgeting::BudgetedMoney& item,
		const std::optional<drn::banking::BankAccount>& ba
	);
	void onRemoved(
		const drn::budgeting::BudgetItemTypes& type,
		const drn::budgeting::BudgetSource& source
	);
	void onPrepareShowAccounts();
	void onAddedAct(const drn::accounting::Account& act);
	void onRecoded(
		const drn::accounting::AccountCode& current,
		const drn::accounting::AccountCode& changed
	);
	void onClosed(const drn::accounting::AccountCode& code);
	void onOpened(const drn::accounting::AccountCode& code);
	void onRemovedAct(const drn::accounting::AccountCode& code);
	void onPrepareShowLedgers();
	void onSelectedLedger(const drn::accounting::AccountCode& code);
	void onAddedTransaction(const drn::accounting::Transaction& txn);
	void onUpdatedTransaction(
		const drn::accounting::TransactionNumber& tn,
		const drn::accounting::Transaction& txn
	);
	void onMarkCleared(const std::set<drn::accounting::TransactionNumber>& tns);
	void onMarkUnreconciled(const std::set<drn::accounting::TransactionNumber>& tns);
	void onMarkReconciled(const drn::accounting::TransactionNumber& tn);
	void onRemovedTransaction(const drn::accounting::TransactionNumber& tn);
	void onPrepareShowBanks();
	void onBankAdded(
		const drn::banking::Bank& b,
		const std::map<
			drn::accounting::AccountNumber,
			drn::banking::SupportedAccountTypes
		>& accountNumbersBankTypes
	);
	void onBankUpdated(
		const drn::banking::Bank& b,
		const std::map<
			drn::accounting::AccountNumber,
			drn::banking::SupportedAccountTypes
		>& accountNumbersBankTypes
	);
	void onBankRenamed(
		const drn::banking::BankName& original,
		const drn::banking::BankName& change
	);
	void onBankClosed(const drn::banking::BankName& bn);
	void onBankOpened(const drn::banking::BankName& bn);
	void onBankRemoved(const drn::banking::BankName& bn);
	std::optional<drn::accounting::AccountCode> onAddedBankAccount(
		const QString& name,
		const drn::banking::SupportedAccountTypes& type,
		const pecunia::Money& openingBalance
	);
	void onRenamedBankAccount(
		const drn::accounting::AccountCode& original,
		const drn::accounting::AccountCode& changed
	);
	void onRemovedBankAccount(const drn::accounting::AccountCode& ac);
	void onPrepareShowProcessBudgetItems();
	void onPrepareShowSurvey();
	void onLookUpBankAccounts(
		const drn::banking::BankName& bn,
		const drn::navigation::AllowSubAccounts& shouldAllow
	);
	/**
	 * @brief Handles a lookup for sources associated with a specific budget item and emits the
	 * result. This method is used to retrieve all associated sources for a given budget item,
	 * categorized by its type. Once the lookup is completed, the results are emitted
	 * as a signal to notify other components.
	 *
	 * @exception std::logic_error Thrown if the type of the budget item is unrecognized or invalid.
	 *
	 * @param requestedBy The unique identifier for the budget item for which sources are being
	 * queried.
	 */
	void onLookUpBudgetItemSources(const drn::budgeting::BudgetItemDueIdentifier& requestedBy);
	/**
	 * @brief Posts a budget item to the budget bank ledger and emits the result as a signal.
	 * This function is used to update the financial ledger with a budget item due,
	 * associating it with a specific account and post date. Successful operation
	 * triggers a signal containing the posted item's ID and assigned transaction number.
	 * In case of failure, an error signal is emitted to notify the failure details.
	 *
	 * @param bid The budget item amount due being posted to the financial ledger.
	 * @param ac A code representing the target account that is funding the posting.
	 * @param postOn The date on which the budget item is to be posted into the ledger.
	 *
	 * @post The budget item is added to the financial ledger with a unique transaction number.
	 * @post Internal states marking unsaved changes are toggled appropriately.
	 */
	void onPostBudgetItemPosted(
		const drn::budgeting::BudgetItemAmountDue& bid,
		const drn::accounting::AccountCode& ac,
		const QDate& postOn
	);
	void onPostBudgetItemPostedUnexpected(
		const drn::budgeting::BudgetItemDueIdentifier& bidi,
		const pecunia::Money& amount,
		const drn::accounting::AccountCode& ac,
		const QDate& postOn
	);
	void onPostBudgetItemSkipped(const drn::budgeting::BudgetItemDueIdentifier& bidi);
	/**
	 * @brief Prepares the system to start a new, empty budget and emits a confirmation signal
	 * if the current budget or state poses potential data overwrite risks.
	 *
	 * This method is used to transition the application to a new and empty budget state. It is
	 * critical in scenarios where user confirmation is essential to avoid unintentional data loss
	 * when overwriting an existing budget or unsaved changes.
	 *
	 * @post If the current budget is not empty and there are unsaved changes, a confirmation signal
	 * is emitted to the user requiring validation before proceeding further.
	 * @post If it is safe to proceed, the logic to initialize a confirmed new empty budget is
	 * executed.
	 */
	void onPrepareNewBudgetEmpty();
	/**
	 * @brief Clears existing budget values and resets the system to prepare for displaying a new
	 * empty budget, then emits a signal to update the user interface.
	 *
	 * This method is used to reset all budget-related data and transition to a state where a
	 * completely new budget can be created. It ensures that any previous data is cleared and
	 * informs the system to display an empty budget interface through an emitted signal.
	 *
	 * @post All stored budget data is cleared.
	 * @post The `showNewBudget` signal is emitted to notify subscribed components of the
	 * transition.
	 */
	void onConfirmedNewBudgetEmpty();
	void onLoadBudgetFile(const QString& filePathName);
	void onConfirmedLoadBudgetFile(const QString& filePathName, const bool isReloaded);
	void onSaveBudgetFile();
	void onSaveAsBudgetFile(const QString& filePathName);
	void onReloadBudgetFile();
	void onCheckIsBudgetFileDirty();
	void onPrepareShowDashboard();
	void onPrepareShowExchangeRates();
	void onFetchExchangeRates(const std::set<pecunia::Currency>& currencies);
	void onStoreExchangeRates(const drn::conversion::ConversionMap& conversions);
	void onPrepareShowConfigure();
	/**
	 * @brief Updates user configuration preferences and validates the provided input. Emits signals
	 * indicating success or failure of the operation.
	 *
	 * This method is used to update various user preferences. It performs input validation to
	 * ensure the values conform to expected limits and modifies the internal preferences if valid.
	 *
	 * @exception Error Thrown if the number of hours worked per week is outside the allowable
	 * range.
	 * @exception Error Thrown if the number of workdays per week is outside the allowable range.
	 * @exception Error Thrown if the notification range is outside the allowable limits.
	 * @exception Error Thrown if the balance window percentage is not within the permitted range.
	 *
	 * @param workHrInWeek The number of hours the user works in a week for scheduling purposes.
	 * @param workdayInWeek The number of workdays in a week the user intends to work.
	 * @param notifyDueWithin How many days within which users should be notified of approaching tasks.
	 * @param preferred The user's preferred currency for displaying budgets and financial data.
	 * @param displayAs The options for displaying the preferred currency (e.g. symbol, code).
	 * @param usable The set of currencies which are acceptable for use within the system.
	 * @param balanceWindow The user-defined percentage threshold for evaluating balance ranges.
	 *
	 * @post The work hours are updated in user preferences if validation passes.
	 * @post The workdays are updated in user preferences if validation passes.
	 * @post The notification period is updated in user preferences if validation passes.
	 * @post The preferred currency and currency display options are updated in user preferences
	 * if validation passes.
	 * @post The set of usable currencies is updated in user preferences if validation passes.
	 * @post The balance percentage range is updated in user preferences if validation passes.
	 * @post On success, the `configureSaveSucceeded` signal is emitted to indicate configuration
	 * changes were saved successfully.
	 * @post On failure, the `configureSaveFailed` signal is emitted with error details.
	 */
	void onConfigureUpdated(
		const std::chrono::hours& workHrInWeek,
		const std::chrono::days& workdayInWeek,
		const std::chrono::days& notifyDueWithin,
		const pecunia::Currency& preferred,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const std::set<pecunia::Currency>& usable,
		const drn::foundation::Percentage& balanceWindow
	);
	/**
	 * @brief Looks up the last reconciliation data for a given account and emits a signal with
	 * the associated details.
	 *
	 * This method retrieves the last reconciliation data from the system for a specified account.
	 * It is primarily used to enable users or systems to quickly access reconciliation information
	 * relevant to a particular account.
	 *
	 * @param an The identifier of the account for which the last reconciliation data is being
	 * retrieved.
	 */
	void onLookUpLastReconciliation(const drn::accounting::AccountNumber& an);
	void onGenerateMappingSurvey(
		const drn::banking::BankName& bn,
		const drn::accounting::AccountNumber& an,
		const QDate& surveyingOn
	);
	void onCompletedSurvey(
		const QDate& completedOn,
		const drn::banking::BankAccount& ba,
		const pecunia::Money& balance,
		const std::set<drn::accounting::TransactionNumber>& allReconciled,
		const drn::surveying::IndexedIncreaseDecreases<
			drn::budgeting::BudgetItemIdentifier
		>& allAdded,
		const drn::surveying::IndexedIncreaseDecreases<drn::accounting::AccountNumber>& allExtra,
		const std::map<
			drn::budgeting::BudgetItemIdentifier,
			drn::surveying::DistributedAmount
		>& distributed,
		const pecunia::Money& nonDistributed
	);
	void onFetchReleaseNotes();
	void onPrepareShowMoneyTransfer();
	void onTransferRequested(
		const drn::accounting::AccountNumber& from,
		const drn::accounting::AccountNumber& to,
		const pecunia::Money& amount,
		const QString& reason
	);
	void onPrepareShowBudgetItemRefund();
	/**
	 * @brief Handles a lookup for non-reconciled refundable transactions associated with a specific
	 * budget item and emits the result. It is used to retrieve transactions categorized as
	 * refundable for a specific budget item. Once the lookup is completed, the result is emitted
	 * as a signal to notify other components.
	 *
	 * @param bii The unique identifier for the budget item for which refundable transactions are
	 * being queried.
	 */
	void onLookUpRefundableTransactions(const drn::budgeting::BudgetItemIdentifier& bii);
	/**
	 * @brief Handles the user request to post a refund to the budget ledgers.
	 *
	 * @param txnNum The transaction number associated with the transaction that is being refunded.
	 * @param refund The amount to be refunded within the financial system.
	 *
	 * @post When successful, the budget ledger contains a refund transaction, and the signal
	 * 'transactionSaveSucceeded' is emitted. In a failure, no changes to the ledger are made, and
	 * the signal 'transactionSaveFailed' is emitted.
	 */
	void onPostRefund(
		const drn::accounting::TransactionNumber& txnNum,
		const pecunia::Money& refund
	);
	/**
	 * @brief Prepares and emits the signal to show the first-time wizard with the user's
	 * preferred settings. This method is used to initialize and indicate that the application
	 * should display the first-time wizard in the graphical interface.
	 *
	 * @post A signal is emitted with the user's preferred settings for currency and its associated
	 * display format.
	 */
	void onPrepareNewBudgetFirstTimeWizard();
	/**
	 * @brief Processes the confirmation from the user to display the first-time wizard.
	 *
	 * This method is invoked to reset the internal state of the budget system and trigger the
	 * display of a first-time wizard GUI item, ensuring a guided setup experience for first-time
	 * users. It emits a signal to configure the wizard with user preferences.
	 *
	 * @post All existing budget data is cleared from the internal system.
	 * @post The state is updated to reflect an initial setup mode.
	 * @post The signal `showFirstTimeWizard` is emitted with the current preferred values.
	 */
	void onConfirmedNewBudgetFirstTimeWizard();
	/**
	 * @param preferred The user's preferred currency for displaying financial information.
	 * @param usable A collection of currencies that are usable in the application.
	 * @param displayAs Specifies options for how the currencies will be displayed to the user.
	 * @param primary The primary banking institution selected by the user.
	 * @param kind Supported account type associated with the primary banking institution.
	 * @param accountName The user-defined name for the account being set up.
	 * @param initialBalance The initial monetary balance for the specified account.
	 * @param selectedTemplate The primary wizard template chosen by the user for setup guidance.
	 */
	void onCompletedNewBudgetFirstTimeWizard(
		const pecunia::Currency& preferred,
		const std::set<pecunia::Currency>& usable,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const drn::banking::BankName& primary,
		const drn::banking::SupportedAccountTypes& kind,
		const QString& accountName,
		const pecunia::Money& initialBalance,
		const drn::navigation::BudgetWizardTemplates& selectedTemplate
	);

signals:
	void showBills(
		const drn::budgeting::BillMap& bills,
		const pecunia::Currency& preferred,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void showDebts(
		const drn::budgeting::DebtMap& debts,
		const std::set<drn::navigation::BudgetBankAccount>& accounts,
		const std::set<drn::accounting::AccountCode>& codes,
		const std::set<drn::banking::BankName>& banks,
		const pecunia::Currency& preferred,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void showGoals(
		const drn::budgeting::GoalMap& goals,
		const std::set<drn::navigation::BudgetBankAccount>& accounts,
		const std::set<drn::accounting::AccountCode>& codes,
		const std::set<drn::banking::BankName>& banks,
		const pecunia::Currency& preferred,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const pecunia::Money& wagesTotal
	);
	void showNontracks(
		const drn::budgeting::NontrackMap& nontracks,
		const pecunia::Currency& preferred,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void showWages(
		const drn::budgeting::WageMap& wages,
		const pecunia::Currency& preferred,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void saveSucceeded(const drn::budgeting::BudgetSource& source);
	void saveFailed(
		const std::optional<drn::budgeting::BudgetSource>& original,
		const QString& errorMessage,
		const std::optional<drn::budgeting::BudgetSource>& displayed
	);
	void showAccounts(
		const std::set<drn::accounting::Account>& accounts,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void accountSaveSucceeded(const drn::accounting::AccountCode& code);
	void accountSaveFailed(
		const std::optional<drn::accounting::AccountCode>& original,
		const QString& errorMessage,
		const std::optional<drn::accounting::AccountCode>& displayed
	);
	void showLedgers(
		const std::set<drn::accounting::Account>& accounts,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void showLedger(const drn::accounting::LedgerAccount& ledger);
	void transactionSaveSucceeded(
		const std::optional<drn::accounting::TransactionNumber>& displayed,
		const std::optional<drn::accounting::TransactionNumber>& posted
	);
	void transactionSaveFailed(
		const std::optional<drn::accounting::TransactionNumber>& tn,
		const QString& errorMessage
	);
	void showBanks(
		const drn::banking::BankMap& banks,
		const std::set<drn::accounting::Account>& allAccounts,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const std::set<drn::banking::BankAccountType>& accountTypes
	);
	void bankSaveSucceeded(const drn::banking::BankName& bn);
	void bankSaveFailed(
		const std::optional<drn::banking::BankName>& original,
		const QString& errorMessage,
		const std::optional<drn::banking::BankName>& displayed
	);
	void bankAccountSaveSucceeded(const drn::accounting::AccountCode& ac);
	void bankAccountSaveFailed(
		const std::optional<drn::accounting::AccountCode>& original,
		const QString& errorMessage,
		const std::optional<drn::accounting::AccountCode>& displayed
	);
	void showProcessBudgetItems(
		const std::chrono::days& dueWithin,
		const std::vector<drn::budgeting::BudgetItemAmountDue>& items,
		const std::set<drn::banking::BankName>& banks,
		const pecunia::Currency& preferredCurrency,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const std::set<drn::accounting::AccountCode>& codes
	);
	void showSurvey(
		const std::set<drn::banking::BankName>& banks,
		const std::set<pecunia::Currency>& usable,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const drn::foundation::Percentage& balanceWindow,
		const std::set<drn::accounting::AccountCode>& codes
	);
	void lookedUpBankAccounts(
		const drn::banking::BankName& bn,
		const std::set<drn::banking::BankAccountType>& accountTypes
	);
	/**
	 * @brief Emits the result of a lookup for budget item sources associated with a specific budget
	 * item. This method is used to notify other components about the sources identified for a
	 * particular budget item.
	 *
	 * @param requestedBy The identifier of the budget item for which the sources were retrieved.
	 * @param sources The set of budget sources associated with the specified budget item.
	 */
	void lookedUpBudgetItemSources(
		const drn::budgeting::BudgetItemDueIdentifier& requestedBy,
		const std::set<drn::budgeting::BudgetSource>& sources
	);
	void processBudgetItemSaveSucceeded(
		const drn::budgeting::BudgetItemDueIdentifier& bidi,
		const std::optional<drn::accounting::TransactionNumber>& txnNum
	);
	void processBudgetItemSaveFailed(
		const drn::budgeting::BudgetItemDueIdentifier& bidi,
		const QString& errorMessage
	);
	/**
	 * @brief Triggered when a user must confirm that a new budget should be created.
	 */
	void showConfirmNewBudget();
	/**
	 * @brief Triggered when a new budget is created and should be displayed.
	 */
	void showNewBudget();
	void showConfirmLoadBudgetFile(const QString& filePathName);
	void loadedBudgetFile(const QString& filePathName, const bool isReloaded);
	void showSaveAsBudgetFile();
	void savedBudgetFile(const QString& filePathName);
	void showNoBudgetFileOpened();
	void checkedIsBudgetFileDirty(const bool isDirty);
	void showDashboard(
		const drn::budgeting::BudgetTotals& bt,
		const std::vector<drn::budgeting::BudgetItemAmountDue>& due,
		const drn::adviser::BudgetHealth& bh,
		const std::chrono::days& dueWithin,
		const pecunia::Currency& preferred,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void showExchangeRates(
		const drn::conversion::ConversionMap& currentConversions,
		const std::set<pecunia::Currency>& currencies
	);
	void fetchedExchangeRates(
		const std::set<pecunia::Currency>& currencies,
		const drn::conversion::ConversionMap& conversions
	);
	void showConfigure(
		const std::chrono::hours& hrInWeek,
		const std::chrono::days& dayInWeek,
		const std::chrono::days& notifyDueWithin,
		const pecunia::Currency& preferred,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const std::set<pecunia::Currency>& usable,
		const std::set<pecunia::Currency>& used,
		const drn::foundation::Percentage& balanceWindow
	);
	void configureSaveSucceeded();
	void configureSaveFailed(const QString& errorMessage);
	/**
	 * @brief Triggered when the most recent reconciliation record for a previously requested bank
	 * account is available. This is a signal indicating the completion of the lookup.
	 *
	 * This method is used to access the last recorded reconciliation within the system.
	 *
	 * @param code The currency associated with the selected account reconciliation process.
	 * @param reconciliation The last reconciled balance, if available.
	 */
	void lookedUpLastReconciliation(
		const pecunia::Currency& code,
		const std::optional<drn::banking::ReconciledBalance>& reconciliation
	);
	void generatedMappingSurvey(const drn::surveying::MappingSurvey& survey);
	void completedSurveySucceeded();
	void completedSurveyFailed(const QString& errorMessage);
	void fetchedReleaseNotes(
		const std::set<drn::updating::ReleaseInformation>& versions
	);
	void updateReleased(const drn::updating::SemanticVersion& latest);
	void showMoneyTransfer(
		const std::set<drn::navigation::BankAccountCode>& fromBanks,
		const std::set<drn::navigation::BankAccountCode>& toBanks,
		const std::set<pecunia::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void showBudgetItemRefund(
		const std::set<pecunia::Currency>& codes,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	/**
	 * @brief Triggered when the result of a refundable transactions lookup for a specific budget
	 * item is completed.
	 *
	 * @param bii The unique identifier specifying the budget item for which refundable transactions
	 * were retrieved. This ensures that the processing is scoped to the relevant budget item.
	 * @param ac The identifier in the ledger books for the account of the budget item.
	 * @param txns A set of transactions identified as refundable for the specified budget item.
	 */
	void lookedUpRefundableTransactions(
		const drn::budgeting::BudgetItemIdentifier& bii,
		const drn::accounting::AccountCode& ac,
		const std::set<drn::accounting::Transaction>& txns
	);
	/**
	 * @brief Triggered when a first-time new budget wizard should be displayed.
	 *
	 * @param preferred The currency to be used as the user's preferred display currency.
	 * @param usable The set of currencies marked as usable, which will be reflected in the GUI
	 * list.
	 * @param displayAs The manner in which the currencies should be displayed.
	 */
	void showNewBudgetFirstTimeWizard(
		const pecunia::Currency& preferred,
		const std::set<pecunia::Currency>& usable,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void updateAvailableAcknowledged();

public:
	Navigator(
		foundation::UniqueQtPtr<networking::RequestSender> sender,
		std::unique_ptr<storage::BudgetStorage> budgetStore,
		std::unique_ptr<storage::ExchangeRatesStorage> exchangeRatesStore
	);
	~Navigator() override;
};

}

#endif
