/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_NETWORKING_MANAGED_REQUEST_SENDER_HPP_
#define DRN_NETWORKING_MANAGED_REQUEST_SENDER_HPP_

#include <networking_export.h>

namespace pecunia
{

template<typename, typename>
class Expected;

}

namespace drn::foundation
{

template<typename>
class QtPtr;

}

namespace drn::networking
{

class ReplyNotifier;
class RequestBuilder;
struct FailureReply;
struct SuccessReply;

/**
 * @brief Derived objects provide an API for sending network requests and getting their replies
 * as a notification signal.
 */
struct DRN_NETWORKING_EXPORT NotifyingSender
{
	NotifyingSender() = default;
	NotifyingSender(const NotifyingSender&) = delete;
	NotifyingSender(NotifyingSender&&) = default;
	NotifyingSender& operator=(const NotifyingSender&) = delete;
	NotifyingSender& operator=(NotifyingSender&&) = default;
	virtual ~NotifyingSender() = default;
	/**
	 * @brief When implemented, sends a network request and returns a notifier for the reply.
	 *
	 * @exception ReplyError When the request to be sent, cannot be sent.
	 *
	 * @param request The request to be sent.
	 *
	 * @return An object that will be notified about any status of the request.
	 */
	[[nodiscard]]
	virtual foundation::QtPtr<ReplyNotifier> send(const RequestBuilder& request) = 0;
};
/**
 * @brief Sends a network request and returns the response synchronously. It blocks the calling
 * thread until a response is received, or an error occurs.
 *
 * @param sender The means to send the request.
 * @param request The information to use when constructing the network request.
 *
 * @return The reply from the request sent. Either containing a SuccessReply upon successful
 * communication or a FailureReply if the request fails.
 */
DRN_NETWORKING_EXPORT pecunia::Expected<SuccessReply, FailureReply> receiveResponse(
	NotifyingSender& sender,
	const RequestBuilder& request
) noexcept;

}

#endif
