/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_NETWORKING_QT_REQUEST_SENDER_HPP_
#define DRN_NETWORKING_QT_REQUEST_SENDER_HPP_

#include <QNetworkAccessManager>

#include <foundation/QtMemory.hpp>
#include <networking_export.h>

#include "RequestSender.hpp"


namespace drn::networking
{

/**
 * @brief Represents a network request's reply response using Qt's QNetworkReply. This is designed
 * to encapsulate a QNetworkReply object, which is used for handling network responses in Qt
 * applications.
 */
class DRN_NETWORKING_EXPORT QtReplyResponse final : public ReplyResponse
{
	foundation::UniqueQtPtr<::QNetworkReply> reply_;

public:
	/**
	 * @brief Full initialisation constructor for a specified QNetworkReply and parent object. This
	 * constructor initializes the object by taking ownership of a QNetworkReply object and setting
	 * a parent for the object. It also sets up a connection to handle the 'finished' signal from
	 * QNetworkReply, calling the responseReady() slot.
	 *
	 * @param reply The QNetworkReply to be observed and owned by this instance.
	 * @param parent The parent used to manage the memory and lifecycle of this instance, and
	 * ensuring it follows the Qt object hierarchy.
	 */
	QtReplyResponse(
		foundation::UniqueQtPtr<::QNetworkReply> reply,
		const foundation::QtPtr<QObject>& parent
	);
	/**
	 * @brief Retrieves the specified attribute from the network request. This is used to obtain
	 * various attributes associated with a network request, such as HTTP status codes, redirection
	 * targets, etc. It is essential for handling different aspects of the response beyond the raw
	 * data.
	 *
	 * @param code The attribute code to retrieve.
	 *
	 * @return The value of the requested attribute wrapped in a QVariant. If the attribute is not
	 * present, the returned QVariant will be invalid.
	 */
	[[nodiscard]]
	QVariant attribute(const QNetworkRequest::Attribute code) const noexcept override;
	/**
	 * @brief Provides the current error state of the network reply. This is used to retrieve the
	 * error status of a network operation. This information is crucial for handling any issues in
	 * network communication, allowing the application to react accordingly.
	 *
	 * @return It gives a value indicating the specific error that occurred during the request, if
	 * any.
	 */
	[[nodiscard]]
	QNetworkReply::NetworkError error() const noexcept override;
	/**
	 * @brief Gives the URL to which the network request was made.
	 *
	 * @return The URL used for the network request.
	 */
	[[nodiscard]]
	QUrl url() const noexcept override;
	/**
	 * @brief Provides a human-readable description of any network error that has occurred during a
	 * request-response cycle. Useful for logging or displaying error messages to users.
	 *
	 * @return A description of the network error as a string. If no error occurred, the string will
	 * be empty.
	 */
	[[nodiscard]]
	QString errorString() const noexcept override;
	/**
	 * @brief Reads and retrieves all remaining data from the reply. It extracts all data that has
	 * been buffered into the reply object from a network operation. Typical usage involves calling
	 * this method after a network request has completed to obtain the full response payload for
	 * further processing. Use this method to read the entirety of the data content in one go.
	 *
	 * @return All the data read from the reply.
	 */
	[[nodiscard]]
	QByteArray readAll() override;
};
/**
 * @brief This is designed to send network requests with the help of Qt's QNetworkAccessManager.
 */
class DRN_NETWORKING_EXPORT QtRequestSender final : public RequestSender
{
	foundation::QtPtr<::QNetworkAccessManager> manager_;

public:
	/**
	 * @brief Full initialisation constructor that is responsible for sending network requests
	 * within a Qt application. It takes a parent Qt object as a parameter to ensure proper memory
	 * management and adherence to the Qt object hierarchy.
	 *
	 * @param parent A smart pointer to the Qt object that will act as the parent of this instance.
	 * This parameter ensures the proper management of the object's lifecycle.
	 */
	explicit QtRequestSender(const foundation::QtPtr<QObject>& parent);
	/**
	 * Sends a GET request using the specified QNetworkRequest.
	 *
	 * @param request The object containing all the necessary information for the sending GET
	 * request.
	 *
	 * @return An object which wraps the network reply. This object can be used to handle and
	 * process the response data.
	 */
	foundation::UniqueQtPtr<ReplyResponse> get(const QNetworkRequest& request) override;
};

}

#endif
