/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_NETWORKING_REPLY_NOTIFIER_HPP_
#define DRN_NETWORKING_REPLY_NOTIFIER_HPP_

#include <pecunia/Expected.hpp>

#include <QNetworkReply>
#include <QObject>

#include <networking_export.h>
#include <networking/StatusCodes.hpp>

class QByteArray;

namespace drn::networking
{

/**
 * @brief Contains details about a successfully sent network request.
 */
struct DRN_NETWORKING_EXPORT SuccessReply
{
	/// The URL of the network request.
	::QUrl url_;
	/// The data payload associated with the network request response.
	::QByteArray data_;
	/// The HTTP status code reply for the network request.
	std::optional<HttpStatusCode> statusCode_;
};
/**
 * @brief Contains details about a failed network request sent.
 */
struct DRN_NETWORKING_EXPORT FailureReply
{
	/// The URL of the network request.
	::QUrl url_;
	/// A textual, human-readable, reason for the failure.
	::QString reason_;
	/// The error code associated with the failure.
	std::optional<::QNetworkReply::NetworkError> error_{std::nullopt};
	/// The HTTP status code reply for the network request.
	std::optional<HttpStatusCode> statusCode_{std::nullopt};
};
/**
 * @brief Monitors and notifies the status of a network request.
 */
class DRN_NETWORKING_EXPORT ReplyNotifier final : public ::QObject
{
	Q_OBJECT

	bool isFinished_;

signals:
	/**
	 * @brief Emitted when the reply for a sent request is received.
	 *
	 * @param response An object containing the information of the reply.
	 */
	void replyReceived(
		const pecunia::Expected<
			drn::networking::SuccessReply,
			drn::networking::FailureReply
		>& response
	);

public:
	/**
	 * @brief Fully initialises the object for use.
	 */
	ReplyNotifier();
	/**
	 * @brief Checks if the notification of a reply for a sent network request is finished.
	 *
	 * @return True if the reply is finished, otherwise false.
	 */
	[[nodiscard]]
	bool isFinished() const noexcept;
	/**
	 * @brief Notifies all listeners about the result of a network request.
	 *
	 * @param response An object containing the information of the reply.
	 *
	 * @post The object is marked as finished. All subsequent invocations do nothing.
	 */
	void notify(const pecunia::Expected<SuccessReply, FailureReply>& response);
};

}

#endif
