/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_NETWORKING_NETWORK_REQUEST_BUILDER_HPP
#define DRN_NETWORKING_NETWORK_REQUEST_BUILDER_HPP

#include <iosfwd>

#include <QString>
#include <QStringList>
#include <QUrlQuery>

#include <optional>
#include <unordered_map>

#include <networking_export.h>


class QDebug;
class QNetworkRequest;

namespace drn::networking
{

/**
 * @brief Representation of the known possible URI schemes.
 */
enum class DRN_NETWORKING_EXPORT Scheme
{
	/// HTTP scheme
	Http,
	/// HTTPS scheme
	Https,
	/// FTP scheme
	Ftp
};
/**
 * @brief A means to build a network request for sending.
 */
class DRN_NETWORKING_EXPORT RequestBuilder
{
	Scheme scheme_;
	::QString host_;
	std::optional<std::uint16_t> port_;
	std::optional<::QString> userName_;
	std::optional<::QString> password_;
	std::optional<::QString> fragment_;
	::QStringList pathSegments_;
	::QUrlQuery query_;
	std::unordered_map<::QString, ::QString> parameters_;
	std::unordered_map<::QString, ::QString> headers_;

public:
	/**
	 * @brief Full initialisation constructor starting with its scheme and host. When a scheme has a
	 * default port, the port is set to it.
	 *
	 * @exception HostError When the hostname is invalid.
	 *
	 * @param scheme The URI scheme to use (e.g., HTTP, HTTPS, FTP).
	 * @param host The hostname to connect to.
	 * @param pathSegments The list of path segments for the URL. Path segments represent the
	 * individual parts of the URL path.
	 */
	RequestBuilder(Scheme scheme, ::QString host, ::QStringList pathSegments = {});
	/// The current URI scheme used in the network request.
	[[nodiscard]]
	Scheme scheme() const noexcept;
	/// The host name to connect to.
	[[nodiscard]]
	const ::QString& host() const noexcept;
	/// The optional port number for the network request.
	[[nodiscard]]
	const std::optional<std::uint16_t>& port() const noexcept;
	/**
	 * @brief The user-name used for authentication if it is set. If the user-name is not set,
	 * the returned optional will be empty.
	 */
	[[nodiscard]]
	const std::optional<::QString>& userName() const noexcept;
	/**
	 * @brief The optional password for authentication if it has been set. If no password is set,
	 * the optional will be empty.
	 */
	[[nodiscard]]
	const std::optional<::QString>& password() const noexcept;
	/// The fragment portion of the URL, if it has been set.
	[[nodiscard]]
	const std::optional<::QString>& fragment() const noexcept;
	/// The list of path segments for the URL. Path segments represent the individual
	/// parts of the URL path.
	[[nodiscard]]
	const ::QStringList& pathSegments() const noexcept;
	/// The queries that have been added in a key-value pair.
	[[nodiscard]]
	const ::QUrlQuery& queries() const noexcept;
	/// The parameters that have been added in a key-value pair.
	[[nodiscard]]
	const std::unordered_map<::QString, ::QString>& parameters() const noexcept;
	/// The headers that have been added in a key-value pair.
	[[nodiscard]]
	const std::unordered_map<::QString, ::QString>& headers() const noexcept;
	/**
	 * @brief Sets the port number.
	 *
	 * @param port The port number to use.
	 *
	 * @return Reference to the current NetworkRequestBuilder for method chaining.
	 */
	RequestBuilder& setPort(const std::uint16_t port);
	/**
	 * @brief Sets the credentials for authentication.
	 *
	 * @param username The username for authentication.
	 * @param password The optional password for authentication.
	 *
	 * @return Reference to the current NetworkRequestBuilder for method chaining.
	 */
	RequestBuilder& setCredentials(
		const ::QString& username,
		const std::optional<::QString>& password = std::nullopt
	);
	/**
	 * @brief Adds a path segment to the URL.
	 *
	 * @param segment The path segment to add.
	 *
	 * @return Reference to the current NetworkRequestBuilder for method chaining.
	 */
	RequestBuilder& addPathSegment(const ::QString& segment);
	/**
	 * @brief Adds a query parameter to the URL.
	 *
	 * @param key The query parameter key.
	 * @param value The query parameter value.
	 *
	 * @return Reference to the current NetworkRequestBuilder for method chaining.
	 */
	RequestBuilder& addQuery(const ::QString& key, const ::QString& value);
	/**
	 * @brief Adds a parameter to the URL.
	 *
	 * @param key The parameter key.
	 * @param value The parameter value.
	 *
	 * @return Reference to the current NetworkRequestBuilder for method chaining.
	 */
	RequestBuilder& addParameter(const ::QString& key, const ::QString& value);
	/**
	 * @brief Sets the fragment portion of the URL.
	 *
	 * @param fragment The fragment to set.
	 *
	 * @return Reference to the current NetworkRequestBuilder for method chaining.
	 */
	RequestBuilder& setFragment(const ::QString& fragment);
	/**
	 * @brief Adds a header to the request.
	 *
	 * @param key The header field name.
	 * @param value The header field value.
	 *
	 * @return Reference to the current RequestBuilder instance for method chaining.
	 */
	RequestBuilder& addHeader(const ::QString& key, const ::QString& value);
	/**
	 * @brief Builds and returns a QNetworkRequest object based on the current configuration.
	 *
	 * @return The constructed QNetworkRequest.
	 */
	QNetworkRequest build() const noexcept;
};

/**
 * @brief Stream insertion operator for a request builder into the output buffer.
 *
 * @param out The output buffer to place the values into.
 * @param rb The contents to place into the output buffer.
 *
 * @post The supplied output buffer contains the contents of the builder.
 *
 * @return The debug buffer post modification. Any sensitive keywords in the builder will have their
 * values obfuscated.
 */
DRN_NETWORKING_EXPORT std::ostream& operator<<(std::ostream& out, const RequestBuilder& rb);
/**
 * @brief Stream insertion operator for a request builder into the debug buffer.
 *
 * @param out The debug buffer to place the values into.
 * @param rb The contents to place into the output buffer.
 *
 * @post The supplied debug buffer contains the contents of the builder.
 *
 * @return The debug buffer post modification. Any sensitive keywords in the builder will have their
 * values obfuscated.
 */
DRN_NETWORKING_EXPORT ::QDebug operator<<(::QDebug out, const RequestBuilder& rb);

}

#endif
