/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_NETWORKING_REQUEST_MANAGER_HPP_
#define DRN_NETWORKING_REQUEST_MANAGER_HPP_

#include <QObject>
#include <QTimer>

#include <map>

#include <foundation/QtMemory.hpp>
#include <foundation/StrongTypeDef.hpp>
#include <networking_export.h>

#include "NotifyingSender.hpp"
#include "ReplyNotifier.hpp"
#include "RequestSender.hpp"

namespace drn::networking
{

class RequestBuilder;

/**
 * @brief Manages all network requests and handles their responses.
 */
class DRN_NETWORKING_EXPORT RequestManager final : public ::QObject, public NotifyingSender
{
	Q_OBJECT

	struct RequestId : foundation::StrongTypeDef<quint64>
	{
		using foundation::StrongTypeDef<quint64>::StrongTypeDef;
		RequestId& operator++() noexcept;
	};
	struct SentRequestReply
	{
		foundation::UniqueQtPtr<ReplyResponse> reply_;
		ReplyNotifier notifier_;
	};

	static RequestId currentId_;

	foundation::UniqueQtPtr<RequestSender> sender_;
	std::map<RequestId, SentRequestReply> replies_;
	foundation::QtPtr<::QTimer> cleanUp_;

private slots:
	void processReply(const RequestId id) noexcept;
	void cleanUpFinishedNotifications();

public:
	/**
	 * @brief Full initialisation constructor initializes the object with a given network request
	 * sender. It also sets up a cleanup timer to manage finished notifications, ensuring efficient
	 * resource management.
	 *
	 * @param sender An object responsible for sending network requests and responses.
	 * @param parent The parent used to manage the memory and lifecycle of this instance, and
	 * ensuring it follows the Qt object hierarchy.
	 */
	explicit RequestManager(
		foundation::UniqueQtPtr<RequestSender> sender,
		const foundation::QtPtr<::QObject>& parent
	);
	RequestManager(const RequestManager&) = delete;
	RequestManager(RequestManager&&) = delete;
	RequestManager& operator=(const RequestManager&) = delete;
	RequestManager& operator=(RequestManager&&) = delete;
	~RequestManager() override;
	[[nodiscard]]
	RequestSender& requestSender() noexcept;
	/**
	 * @brief Sends a network request and returns a notifier for the reply.
	 *
	 * @exception ReplyError When the request to be sent, cannot be sent.
	 *
	 * @param request The network request to be sent.
	 *
	 * @return An object that will be notified about any status of the request.
	 */
	[[nodiscard]]
	foundation::QtPtr<ReplyNotifier> send(const RequestBuilder& request) override;
};

}

#endif
