/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_NETWORKING_REQUEST_SENDER_HPP_
#define DRN_NETWORKING_REQUEST_SENDER_HPP_

#include <QNetworkReply>
#include <QNetworkRequest>
#include <QObject>

#include <foundation/QtMemory.hpp>
#include <networking_export.h>

namespace drn::networking
{

/**
 * @brief Interface for retrieving network replies from a sent request. When deriving, it'll provide
 * methods for accessing various response information from a received network reply.
 */
class DRN_NETWORKING_EXPORT ReplyResponse : public ::QObject
{
	Q_OBJECT

signals:
	/**
	 * @brief Triggered when a response from a network request is ready for processing. This signal
	 * is meant to be connected to a slot or lambda function that handles the response once a
	 * network request completes. It indicates that the reply object contains data that is ready to
	 * be processed, or is in an error state.
	 */
	void ready();

public:
	/**
	 * @brief Fully initialises the object with a specified parent.
	 *
	 * @param parent The parent used to manage the memory and lifecycle of this instance, and
	 * ensuring it follows the Qt object hierarchy.
	 */
	explicit ReplyResponse(const foundation::QtPtr<QObject>& parent);
	/**
	 * @brief When implemented, retrieves the specified attribute from the network request. This
	 * is used to obtain various attributes associated with a network request, such as HTTP status
	 * codes, redirection targets, etc. It is essential for handling different aspects of the
	 * response beyond the raw data.
	 *
	 * @param code The attribute code to retrieve.
	 *
	 * @return The value of the requested attribute wrapped in a QVariant. If the attribute is not
	 * present, the returned QVariant will be invalid.
	 */
	[[nodiscard]]
	virtual ::QVariant attribute(const QNetworkRequest::Attribute code) const noexcept = 0;
	/**
	 * @brief When implemented, provides the current error state of the network reply. This is used
	 * to retrieve the error status of a network operation. This information is crucial for handling
	 * any issues in network communication, allowing the application to react accordingly.
	 *
	 * @return It gives a value indicating the specific error that occurred during the request, if
	 * any.
	 */
	[[nodiscard]]
	virtual ::QNetworkReply::NetworkError error() const noexcept = 0;
	/**
	 * @brief When implemented, gives the URL to which the network request was made.
	 *
	 * @return The URL used for the network request.
	 */
	[[nodiscard]]
	virtual ::QUrl url() const noexcept = 0;
	/**
	 * @brief When implemented, provides a human-readable description of any network error that has
	 * occurred during a request-response cycle. Useful for logging or displaying error messages to
	 * users.
	 *
	 * @return A description of the network error as a string. If no error occurred, the string will
	 * be empty.
	 */
	[[nodiscard]]
	virtual ::QString errorString() const noexcept = 0;
	/**
	 * @brief When implemented, reads and retrieves all remaining data from the reply. This will
	 * extract all data that has been buffered into the reply object from a network operation.
	 * Typical usage involves calling this method after a network request has completed to obtain
	 * the full response payload for further processing. Use this method to read the entirety of the
	 * data content in one go.
	 *
	 * @return All the data read from the reply.
	 */
	[[nodiscard]]
	virtual ::QByteArray readAll() = 0;
};
/**
 * @brief Interface for sending network requests. When deriving, it'll provide methods for
 * sending various network requests.
 */
class DRN_NETWORKING_EXPORT RequestSender : public ::QObject
{
	Q_OBJECT

public:
	/**
	 * @brief Fully initialises the object with a specified parent.
	 *
	 * @param parent The parent used to manage the memory and lifecycle of this instance, and
	 * ensuring it follows the Qt object hierarchy.
	 */
	explicit RequestSender(const foundation::QtPtr<QObject>& parent);
	/**
	 * @brief When implemented, sends a GET request to the specified network address.
	 *
	 * @param request The network request to be sent. It should contain all necessary
	 * information such as the URL and any required headers.
	 *
	 * @return An object that encapsulates the network reply information.
	 */
	virtual foundation::UniqueQtPtr<ReplyResponse> get(const QNetworkRequest& request) = 0;
};

}

#endif
