/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_NETWORKING_MANAGED_STATUS_CODES_HPP_
#define DRN_NETWORKING_MANAGED_STATUS_CODES_HPP_

#include <array>
#include <cstdint>
#include <iosfwd>
#include <string>

#include <networking_export.h>

class QDebug;

namespace drn::networking
{
/**
 * @brief Represents all the various HTTP status codes as defined in RFC 7231 and other relevant
 * RFCs.
 */
enum class DRN_NETWORKING_EXPORT HttpStatusCode
{
	//{ 1xx Informational

	/// The HTTP status code for Continue
	Continue = 100,
	/// The HTTP status code for Switching Protocols
	SwitchingProtocols = 101,
	/// The HTTP status code for Processing
	Processing = 102,
	/// The HTTP status code for Early Hints
	EarlyHints = 103,

	//}

	//{ 2xx Success

	/// The HTTP status code for OK
	Ok = 200,
	/// The HTTP status code for Created
	Created = 201,
	/// The HTTP status code for Accepted
	Accepted = 202,
	/// The HTTP status code for Non-Authoritative Information
	NonAuthoritativeInformation = 203,
	/// The HTTP status code for No Content
	NoContent = 204,
	/// The HTTP status code for Reset Content
	ResetContent = 205,
	/// The HTTP status code for Partial Content
	PartialContent = 206,
	/// The HTTP status code for Multi-Status
	MultiStatus = 207,
	/// The HTTP status code for Already Reported
	AlreadyReported = 208,
	/// The HTTP status code for IM Used
	ImUsed = 226,

	//}

	//{ 3xx Redirection

	/// The HTTP status code for Multiple Choices
	MultipleChoices = 300,
	/// The HTTP status code for Moved Permanently
	MovedPermanently = 301,
	/// The HTTP status code for Found
	Found = 302,
	/// The HTTP status code for See Other
	SeeOther = 303,
	/// The HTTP status code for Not Modified
	NotModified = 304,
	/// The HTTP status code for Use Proxy
	UseProxy = 305,
	/// The HTTP status code for Temporary Redirect
	TemporaryRedirect = 307,
	/// The HTTP status code for Permanent Redirect
	PermanentRedirect = 308,

	//}

	//{ 4xx Client Errors

	/// The HTTP status code for Bad Request
	BadRequest = 400,
	/// The HTTP status code for Unauthorized
	Unauthorized = 401,
	/// The HTTP status code for Payment Required
	PaymentRequired = 402,
	/// The HTTP status code for Forbidden
	Forbidden = 403,
	/// The HTTP status code for Not Found
	NotFound = 404,
	/// The HTTP status code for Method Not Allowed
	MethodNotAllowed = 405,
	/// The HTTP status code for Not Acceptable
	NotAcceptable = 406,
	/// The HTTP status code for Proxy Authentication Required
	ProxyAuthenticationRequired = 407,
	/// The HTTP status code for Request Timeout
	RequestTimeout = 408,
	/// The HTTP status code for Conflict
	Conflict = 409,
	/// The HTTP status code for Gone
	Gone = 410,
	/// The HTTP status code for Length Required
	LengthRequired = 411,
	/// The HTTP status code for Precondition Failed
	PreconditionFailed = 412,
	/// The HTTP status code for Payload Too Large
	PayloadTooLarge = 413,
	/// The HTTP status code for URI Too Long
	UriTooLong = 414,
	/// The HTTP status code for Unsupported Media Type
	UnsupportedMediaType = 415,
	/// The HTTP status code for Range Not Satisfiable
	RangeNotSatisfiable = 416,
	/// The HTTP status code for Expectation Failed
	ExpectationFailed = 417,
	/// The HTTP status code for I'm a Teapot
	ImATeapot = 418,
	/// The HTTP status code for Misdirected Request
	MisdirectedRequest = 421,
	/// The HTTP status code for Unprocessable Entity
	UnprocessableEntity = 422,
	/// The HTTP status code for Locked
	Locked = 423,
	/// The HTTP status code for Failed Dependency
	FailedDependency = 424,
	/// The HTTP status code for Too Early
	TooEarly = 425,
	/// The HTTP status code for Upgrade Required
	UpgradeRequired = 426,
	/// The HTTP status code for Precondition Required
	PreconditionRequired = 428,
	/// The HTTP status code for Too Many Requests
	TooManyRequests = 429,
	/// The HTTP status code for Request Header Fields Too Large
	RequestHeaderFieldsTooLarge = 431,
	/// The HTTP status code for Unavailable For Legal Reasons
	UnavailableForLegalReasons = 451,

	//}

	//{ 5xx Server Errors

	/// The HTTP status code for Internal Server Error
	InternalServerError = 500,
	/// The HTTP status code for Not Implemented
	NotImplemented = 501,
	/// The HTTP status code for Bad Gateway
	BadGateway = 502,
	/// The HTTP status code for Service Unavailable
	ServiceUnavailable = 503,
	/// The HTTP status code for Gateway Timeout
	GatewayTimeout = 504,
	/// The HTTP status code for HTTP Version Not Supported
	HttpVersionNotSupported = 505,
	/// The HTTP status code for Variant Also Negotiates
	VariantAlsoNegotiates = 506,
	/// The HTTP status code for Insufficient Storage
	InsufficientStorage = 507,
	/// The HTTP status code for Loop Detected
	LoopDetected = 508,
	/// The HTTP status code for Not Extended
	NotExtended = 510,
	/// The HTTP status code for Network Authentication Required
	NetworkAuthenticationRequired = 511

	//}
};

/**
 * @brief Converts an integer value to its corresponding HTTP status code.
 *
 * @exception logic_error When the supplied value is not a known HTTP status code.
 *
 * @param value The integer representation of the HTTP status code.
 *
 * @return The corresponding HttpStatusCode for the given integer value.
 */
DRN_NETWORKING_EXPORT HttpStatusCode toHttpStatusCode(const std::int32_t value);
/**
 * @brief Converts an HttpStatusCode to its human-readable representation.
 *
 * @param value The HttpStatusCode to be converted to a string.
 *
 * @return A representation of the given HttpStatusCode.
 */
DRN_NETWORKING_EXPORT std::string toStdString(const HttpStatusCode value);
/**
 * @brief Converts an HttpStatusCode to its integer representation.
 *
 * @param value The HttpStatusCode to be converted to an integer.
 *
 * @return The integer representation of the given HttpStatusCode.
 */
DRN_NETWORKING_EXPORT std::int32_t toInteger(const HttpStatusCode value);
/**
 * @brief Stream insertion operator.
 *
 * @param out The stream to insert the HTTP status code information.
 * @param code The HTTP status code information to insert into the stream.
 *
 * @post The stream contains the HTTP status code information in the form:
 *		- code [Description]
 *
 * @return The modified stream.
 */
DRN_NETWORKING_EXPORT std::ostream& operator<<(std::ostream& out, const HttpStatusCode code);
/**
 * @brief Qt stream insertion operator.
 *
 * @param out The stream to insert the HTTP status code information.
 * @param code The HTTP status code information to insert into the stream.
 *
 * @post The stream contains the HTTP status code information in the form:
 *		- code [Description]
 *
 * @return The modified stream.
 */
DRN_NETWORKING_EXPORT ::QDebug operator<<(::QDebug out, const HttpStatusCode code);

/// Every known HTTP status code.
inline constexpr const std::array allStatusCodes{
	HttpStatusCode::Continue,
	HttpStatusCode::SwitchingProtocols,
	HttpStatusCode::Processing,
	HttpStatusCode::EarlyHints,
	HttpStatusCode::Ok,
	HttpStatusCode::Created,
	HttpStatusCode::Accepted,
	HttpStatusCode::NonAuthoritativeInformation,
	HttpStatusCode::NoContent,
	HttpStatusCode::ResetContent,
	HttpStatusCode::PartialContent,
	HttpStatusCode::MultiStatus,
	HttpStatusCode::AlreadyReported,
	HttpStatusCode::ImUsed,
	HttpStatusCode::MultipleChoices,
	HttpStatusCode::MovedPermanently,
	HttpStatusCode::Found,
	HttpStatusCode::SeeOther,
	HttpStatusCode::NotModified,
	HttpStatusCode::UseProxy,
	HttpStatusCode::TemporaryRedirect,
	HttpStatusCode::PermanentRedirect,
	HttpStatusCode::BadRequest,
	HttpStatusCode::Unauthorized,
	HttpStatusCode::PaymentRequired,
	HttpStatusCode::Forbidden,
	HttpStatusCode::NotFound,
	HttpStatusCode::MethodNotAllowed,
	HttpStatusCode::NotAcceptable,
	HttpStatusCode::ProxyAuthenticationRequired,
	HttpStatusCode::RequestTimeout,
	HttpStatusCode::Conflict,
	HttpStatusCode::Gone,
	HttpStatusCode::LengthRequired,
	HttpStatusCode::PreconditionFailed,
	HttpStatusCode::PayloadTooLarge,
	HttpStatusCode::UriTooLong,
	HttpStatusCode::UnsupportedMediaType,
	HttpStatusCode::RangeNotSatisfiable,
	HttpStatusCode::ExpectationFailed,
	HttpStatusCode::ImATeapot,
	HttpStatusCode::MisdirectedRequest,
	HttpStatusCode::UnprocessableEntity,
	HttpStatusCode::Locked,
	HttpStatusCode::FailedDependency,
	HttpStatusCode::TooEarly,
	HttpStatusCode::UpgradeRequired,
	HttpStatusCode::PreconditionRequired,
	HttpStatusCode::TooManyRequests,
	HttpStatusCode::RequestHeaderFieldsTooLarge,
	HttpStatusCode::UnavailableForLegalReasons,
	HttpStatusCode::InternalServerError,
	HttpStatusCode::NotImplemented,
	HttpStatusCode::BadGateway,
	HttpStatusCode::ServiceUnavailable,
	HttpStatusCode::GatewayTimeout,
	HttpStatusCode::HttpVersionNotSupported,
	HttpStatusCode::VariantAlsoNegotiates,
	HttpStatusCode::InsufficientStorage,
	HttpStatusCode::LoopDetected,
	HttpStatusCode::NotExtended,
	HttpStatusCode::NetworkAuthenticationRequired
};
}

#endif
