/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_SURVEYING_COMPLETED_SURVEY_H_
#define DRN_SURVEYING_COMPLETED_SURVEY_H_

#include <pecunia/Money.hpp>

#include <QString>

#include <compare>
#include <iosfwd>
#include <map>
#include <optional>

#include <budgeting/BudgetItemIdentifier.h>
#include <banking/BankAccount.h>
#include <banking/Reconciliation.h>

#include <surveying_export.h>


class QDate;
class QDebug;

namespace drn::accounting
{

class AccountCode;
class AccountNumber;

}
namespace drn::foundation
{

class Percentage;

}
namespace drn::surveying
{

class DistributedAmount;

/**
 * @brief Represents a completed survey of a reconciled bank account, including distributions
 * and associated data. It is primarily designed to allow tracking and managing a bank account’s
 * reconciled state with respect to budgeting and distribution of account balance amount.
 */
class DRN_SURVEYING_EXPORT CompletedSurvey final
{
	banking::ReconciledBankAccount reconciled_;
	std::map<budgeting::BudgetItemIdentifier, pecunia::Money> distribution_;
	pecunia::Money nonDistributed_;
	std::optional<::QString> checksum_;

public:
	CompletedSurvey() = default;
	/**
	 * @brief Partial initialisation constructor for creating a new survey.
	 *
	 * @param ba The bank account that is reconciled.
	 * @param currency The currency held within the bank account.
	 */
	CompletedSurvey(banking::BankAccount ba, const pecunia::Currency& currency);
	/**
	 * @brief Full initialisation constructor. It is designed to ensure consistency and validity of
	 * the survey’s data.
	 *
	 * @exception SurveyError When the currency of the non-distributed amount does not match
	 * the reconciled bank account’s currency.
	 * @exception SurveyError When a checksum is provided, but does not match the calculated
	 * value based on the values supplied.
	 * @exception SurveyError When a zero-amount is distributed in a budget item.
	 *
	 * @param reconciled The reconciled bank account that reflects the current state of the account.
	 * @param distribution A mapping of budget items to monetary allocations that distributes the
	 * account’s reconciled balance.
	 * @param nonDistributed The amount of funds that remain after allocation to budget items.
	 * @param checksum A checksum to validate the integrity of the survey content. No validation
	 * occurs when null.
	 */
	CompletedSurvey(
		banking::ReconciledBankAccount reconciled,
		std::map<budgeting::BudgetItemIdentifier, pecunia::Money> distribution,
		pecunia::Money nonDistributed,
		std::optional<::QString> checksum
	);

	CompletedSurvey(const CompletedSurvey& other) noexcept = default;

	CompletedSurvey(CompletedSurvey&& other) noexcept = default;

	~CompletedSurvey() noexcept = default;
	/**
	 * @brief Assigns the contents of another instance to the current instance.
	 *
	 * @param other An instance whose data will be duplicated into the current instance.
	 *
	 * @post The current survey’s data will be updated to match the provided instance.
	 *
	 * @return The current modified instance.
	 */
	CompletedSurvey& operator=(const CompletedSurvey& other) noexcept;
	/**
	 * @brief Moves the contents of another instance into this instance, effectively transferring
	 * ownership of all data.
	 *
	 * This ensures efficient transfer of resources without the need for duplicating data, and it is
	 * used to handle ownership transfer scenarios where copying data would negatively impact
	 * performance or where the original instance’s data is no longer needed.
	 *
	 * @param other The instance whose data will be transferred to this instance.
	 *
	 * @post The “other” instance will be in an undefined state where all moved-from members in it
	 * are no longer safe to use until they are reassigned.
	 * @post This instance will own the data previously contained in “other”.
	 *
	 * @return This instance, now containing the moved data.
	 */
	CompletedSurvey& operator=(CompletedSurvey&& other) noexcept;
	/**
	 * @brief Compares this completed survey with another to determine their relative order.
	 * The comparison considers reconciliation state, non-distributed amounts, distribution data,
	 * and a checksum in that order.
	 * @pre Both surveys must be valid and fully initialised before invoking this function.
	 *
	 * @param other The other completed survey to compare against.
	 *
	 * @return The result of the comparison expressed in a strong ordering manner. It is
	 * `strong_ordering::less` if this survey is less, `strong_ordering::greater` if greater,
	 * and `strong_ordering::equal` if both are considered equivalent.
	 */
	[[nodiscard]]
	std::strong_ordering operator<=>(const CompletedSurvey& other) const noexcept;

	[[nodiscard]]
	bool operator==(const CompletedSurvey&) const noexcept = default;

	[[nodiscard]]
	bool operator!=(const CompletedSurvey&) const noexcept = default;
	/**
	 * @brief Retrieves the reconciled bank account associated with the completed survey. This
	 * information is intended to facilitate the examination and tracking of the reconciled state
	 * within a budgeting or financial context.
	 *
	 * @return The reconciled bank account associated with the completed survey.
	 */
	[[nodiscard]]
	const banking::ReconciledBankAccount& reconciled() const noexcept;
	/**
	 * @brief Provides access to the distribution of allocated amounts from the reconciled balance
	 * in the completed survey, enabling consumers to review the allocation results.
	 *
	 * @return The mapping of budget items to their corresponding allocated amounts in the completed
	 * survey.
	 */
	[[nodiscard]]
	const std::map<
		budgeting::BudgetItemIdentifier,
		pecunia::Money
	>& distribution() const noexcept;
	/**
	 * @brief Retrieves the part of the total reconciled amount that has not been distributed.
	 *
	 * @return The amount that remains undistributed.
	 */
	[[nodiscard]]
	const pecunia::Money& nonDistributed() const noexcept;
	/**
	 * @brief Retrieves the stored checksum value associated with the survey’s finalised state. This
	 * provides a mechanism to verify the integrity and consistency of the survey data after it has
	 * been reconciled.
	 *
	 * @return The checksum value, representing the integrity validation data. If no checksum has
	 * been set, it returns an empty optional, and the validity cannot be guaranteed.
	 */
	[[nodiscard]]
	const std::optional<::QString>& checksum() const noexcept;
	/**
	 * @brief Finalises and records the completion of a survey of a reconciled bank account,
	 * including its reconciled balance, distribution, and related details.
	 *
	 * @exception SurveyError When the non-distributed balance currency is not the same as the
	 * reconciled account’s currency.
	 * @exception SurveyError When the combined total distribution balance and non-distributed
	 * balance is not within the specified percentage window of the reconciled balance.
	 * @exception SurveyError When a zero-amount is distributed in a budget item.
	 *
	 * @param completedOn The date when the survey was finalised.
	 * @param reconciledBalance The reconciled balance of the bank account associated with the
	 * survey.
	 * @param distribution A map of budget item identifiers to distributed amounts, representing how
	 * the reconciled balance is distributed amongst budget items.
	 * @param nonDistributed The portion of the reconciled balance that has not been distributed to
	 * any budget item.
	 * @param balanceWindow The allowed percentage deviation within which the distribution balance
	 * must align with the reconciled balance.
	 *
	 * @post Updates the reconciled bank account to mark it as reconciled at the specified date and
	 * balance.
	 * @post Internal state reflecting distribution details and non-distributed balances is updated.
	 * @post A checksum representing the reconciliation state is recalculated and stored.
	 */
	void completed(
		const ::QDate& completedOn,
		const pecunia::Money& reconciledBalance,
		const std::map<budgeting::BudgetItemIdentifier, DistributedAmount>& distribution,
		const pecunia::Money& nonDistributed,
		const foundation::Percentage& balanceWindow
	);
	/**
	 * @brief Updates the surveyed bank account’s name.
	 *
	 * @exception SurveyError When the original bank name does not match the currently surveyed
	 * bank account.
	 *
	 * @param original The current name of the surveyed bank account.
	 * @param newName The new name to assign to the surveyed bank account.
	 *
	 * @post Updates the internal state of the surveyed bank account to use the new bank name
	 * whilst retaining account type, balance, and reconciliation details.
	 * @post Re-calculates the checksum of the surveyed account to maintain an up-to-date internal
	 * checksum following the name update.
	 */
	void renamed(const banking::BankName& original, const banking::BankName& newName);
	/**
	 * @brief Updates the reconciled bank account with a new account number.
	 *
	 * @exception SurveyError If the provided original account number does not match the account
	 * number of the reconciled bank account.
	 *
	 * @param original The account number of the bank account being renamed.
	 * @param changed The new account number to replace the original number for the reconciled
	 * bank account.
	 *
	 * @post Updates the internal record of the reconciled bank account to reflect the new
	 * account number.
	 * @post Recalculates the checksum to ensure internal consistency with the updated
	 * account number.
	 */
	void renamed(
		const accounting::AccountNumber& original,
		const accounting::AccountNumber& changed
	);
	/**
	 * @brief Renames a specific budget distribution item identified by its identifier.
	 *
	 * @exception BudgetItemError Thrown when the specified identifier does not correspond to any
	 * existing budget item in the distribution.
	 *
	 * @param id The identifier of the budget item to be renamed.
	 * @param newName The new name to be assigned to the identified budget item.
	 *
	 * @post A new budget item is added with the specified new name, preserving the corresponding
	 * distribution value.
	 * @post The original budget item is removed from the distribution.
	 * @post The checksum representing the integrity of the distribution is recalculated to reflect
	 * the changes.
	 */
	void renamed(const budgeting::BudgetItemIdentifier& id, const budgeting::BudgetSource& newName);
};
/**
 * @brief Represents a collection of completed surveys identified by a bank account. It provides
 * extended functionality built upon a map-like structure for efficient access and modification of
 * surveys.
 */
class DRN_SURVEYING_EXPORT CompletedSurveys final : std::map<banking::BankAccount, CompletedSurvey>
{
public:
	/**
	 * @brief Constructs a collection of completed surveys associated with bank accounts.
	 *
	 * @exception SurveyError Thrown when an indexed bank account does not match the surveyed bank
	 * account.
	 *
	 * @param surveys A collection mapping bank accounts to their completed surveys.
	 */
	explicit CompletedSurveys(std::map<banking::BankAccount, CompletedSurvey> surveys);

	using std::map<banking::BankAccount, CompletedSurvey>::at;

	using std::map<banking::BankAccount, CompletedSurvey>::begin;

	using std::map<banking::BankAccount, CompletedSurvey>::cbegin;

	using std::map<banking::BankAccount, CompletedSurvey>::cend;

	using std::map<banking::BankAccount, CompletedSurvey>::const_iterator;

	using std::map<banking::BankAccount, CompletedSurvey>::count;

	using std::map<banking::BankAccount, CompletedSurvey>::empty;

	using std::map<banking::BankAccount, CompletedSurvey>::end;

	using std::map<banking::BankAccount, CompletedSurvey>::erase;

	using std::map<banking::BankAccount, CompletedSurvey>::map;

	using std::map<banking::BankAccount, CompletedSurvey>::size;
	/**
	 * @brief Looks up a completed survey for a given bank account. This query provides information
	 * regarding the survey associated with the reconciled state of the bank account, aiding in
	 * identifying its distribution and budgeting details.
	 *
	 * @param ba The bank account for which the completed survey is being looked up.
	 *
	 * @return An optional object containing the completed survey for the given bank account, or an
	 * empty result if no associated survey is found.
	 */
	[[nodiscard]]
	std::optional<CompletedSurvey> lookUpSurvey(const banking::BankAccount& ba) const noexcept;
	/**
	 * @brief Searches for a completed survey linked to a specific bank and account. This is used to
	 * retrieve a previously reconciled survey for a bank account, enabling review or further
	 * processing of the completed survey data.
	 *
	 * @param bn The name of the bank associated with the survey being searched.
	 * @param an The number of the account associated with the survey being searched.
	 *
	 * @return An optional containing the completed survey if a match is found, or an empty result
	 * if no matching survey exists.
	 */
	[[nodiscard]]
	std::optional<CompletedSurvey> lookUpSurvey(
		const banking::BankName& bn,
		const accounting::AccountNumber& an
	) const noexcept;
	/**
	 * @brief Inserts or updates a completed survey for a specific bank account, ensuring
	 * distribution identifiers are unique within the context of all existing surveys.
	 *
	 * @exception SurveyError When the provided survey contains a distribution identifier that
	 * overlaps with an existing survey’s identifiers for a different bank account.
	 *
	 * @param survey The completed survey to insert or update. This survey includes reconciled data,
	 * distributions, and other associated information.
	 *
	 * @post If the bank account associated with the provided survey does not exist within the
	 * system, the survey is added as a new entry. If the bank account does exist, the existing
	 * survey is replaced with the provided survey.
	 *
	 * @return Returns true if a new survey is inserted, and false if an existing survey is
	 * replaced.
	 */
	bool upsert(CompletedSurvey survey);
};

/**
 * @brief Compares two collections of completed surveys to determine if they are equivalent.
 * This function evaluates whether the sets of surveys are identical based on the survey data
 * and the associated bank accounts.
 *
 * @param lhs The first collection of completed surveys for comparison.
 * @param rhs The second collection of completed surveys for comparison.
 *
 * @return Returns `true` if both collections are equivalent, meaning they contain the same
 * surveys corresponding to the same bank accounts. Otherwise, returns `false`.
 */
DRN_SURVEYING_EXPORT bool operator==(
	const CompletedSurveys& lhs,
	const CompletedSurveys& rhs
);
/**
 * @brief Compares two collections of completed surveys to determine if they are not equivalent.
 * This function evaluates whether the sets of surveys are not identical based on the survey data
 * and the associated bank accounts.
 *
 * @param lhs The first collection of completed surveys for comparison.
 * @param rhs The second collection of completed surveys for comparison.
 *
 * @return Returns `false` if both collections are equivalent, meaning they contain the same
 * surveys corresponding to the same bank accounts. Otherwise, returns `true`.
 */
DRN_SURVEYING_EXPORT bool operator!=(
	const CompletedSurveys& lhs,
	const CompletedSurveys& rhs
);
/**
 * @brief Full object standard stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_SURVEYING_EXPORT std::ostream& operator<<(std::ostream& out, const CompletedSurvey& data);
/**
 * @brief Full object standard stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_SURVEYING_EXPORT std::ostream& operator<<(std::ostream& out, const CompletedSurveys& data);
/**
 * @brief Full object Qt stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_SURVEYING_EXPORT ::QDebug operator<<(::QDebug out, const CompletedSurvey& data);
/**
 * @brief Full object Qt stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_SURVEYING_EXPORT ::QDebug operator<<(::QDebug out, const CompletedSurveys& data);

}

#endif
