/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_SURVEYING_DISTRIBUTED_AMOUNT_HPP_
#define DRN_SURVEYING_DISTRIBUTED_AMOUNT_HPP_

#include <pecunia/Money.hpp>

#include <QDate>

#include <compare>
#include <iosfwd>
#include <surveying_export.h>


namespace drn::surveying
{

/**
 * @brief This represents a budget item's amount distributed in a survey.
 */
class DRN_SURVEYING_EXPORT DistributedAmount
{
	pecunia::Money distributed_{};
	::QDate nextOccursOn_{};
	pecunia::Money maximum_{};

public:
	DistributedAmount() = default;
	/**
	 * @brief Full initialisation constructor.
	 *
	 * @exception Error Thrown when the distributed amount exceeds the maximum permissible value.
	 *
	 * @pre The next occurrence date must be a valid date.
	 *
	 * @param distributed The distributed amount to initialise with.
	 * @param nextOccursOn The date when the distributed amount is next scheduled.
	 * @param maximum The maximum permissible value for the distributed amount.
	 */
	DistributedAmount(
		pecunia::Money distributed,
		::QDate nextOccursOn,
		pecunia::Money maximum
	);
	/**
	 * @brief Creates a new instance by copying the values from another instance.
	 *
	 * @param other The instance to copy from.
	 *
	 * @post The newly created instance will have identical values for distributed amount, the next
	 * occurrence date, and the maximum permissible value as the provided instance.
	 */
	DistributedAmount(const DistributedAmount& other);
	/**
	 * @brief Constructs an instance by transferring ownership from another instance.
	 *
	 * @param other The instance to efficiently transfer resources and state from.
	 *
	 * @post The new instance will take ownership of all resources that belonged to the `other`
	 * instance prior to the call. The `other` instance will remain valid but its state will be
	 * unspecified.
	 */
	DistributedAmount(DistributedAmount&& other) noexcept;
	/**
	 * @brief Full object assignment.
	 *
	 * @param other The other instance to assign from.
	 *
	 * @post The current instance will have identical values for distributed amount, the next
	 * occurrence date, and the maximum permissible value as the provided instance.
	 *
	 * @return The current instance post assignment.
	 */
	DistributedAmount& operator=(const DistributedAmount& other);
	/**
	 * @brief Full object assignment ownership transfer.
	 *
	 * @param other The instance to efficiently transfer resources and state from.
	 *
	 * @post The new instance will take ownership of all resources that belonged to the `other`
	 * instance prior to the call. The `other` instance will remain valid but its state will be
	 * unspecified.
	 *
	 * @return The current instance post assignment.
	 */
	DistributedAmount& operator=(DistributedAmount&& other) noexcept;
	/// @brief An accessor the current distributed amount for a budget item in a survey.
	[[nodiscard]]
	const pecunia::Money& distributed() const noexcept;
	/// @brief An accessor the date when the distributed amount is next scheduled.
	[[nodiscard]]
	const ::QDate& nextOccursOn() const noexcept;
	/// @brief An accessor the maximum permissible value for the distributed amount.
	[[nodiscard]]
	const pecunia::Money& maximum() const noexcept;

	[[nodiscard]]
	std::strong_ordering operator<=>(const DistributedAmount& other) const noexcept = default;
	/**
	 * @brief Distributed amount assignment.
	 *
	 * @exception Error Thrown when the distributed amount exceeds the maximum permissible value.
	 *
	 * @param amount The amount to assign to the distributed amount.
	 *
	 * @post The current instance will have identical values for the distributed amount. The next
	 * occurrence date, and the maximum permissible value remain unchanged.
	 *
	 * @return The current instance post assignment.
	 */
	DistributedAmount& operator=(const pecunia::Money& amount);
	/**
	 * @brief Performs addition on the distributed amount.
	 *
	 * @exception Error Thrown when the new distributed amount exceeds the maximum permissible
	 * value.
	 *
	 * @param amount The amount to increase the distributed amount by.
	 *
	 * @return The result of the mathematical operation.
	 */
	[[nodiscard]]
	DistributedAmount operator+(const pecunia::Money& amount) const;
	/**
	 * @brief Performs addition on the distributed amount.
	 *
	 * @exception Error Thrown when the new distributed amount exceeds the maximum permissible
	 * value.
	 *
	 * @param amount The amount to increase the distributed amount by.
	 *
	 * @post The current instance will have the distributed amount increased by the requested
	 * amount. The next occurrence date, and the maximum permissible value remain unchanged.
	 *
	 * @return The current instance after the mathematical operation.
	 */
	DistributedAmount& operator+=(const pecunia::Money& amount);
	/**
	 * @brief Performs subtraction on the distributed amount.
	 *
	 * @exception Error Thrown when the new distributed amount exceeds the maximum permissible
	 * value.
	 *
	 * @param amount The amount to decrease the distributed amount by.
	 *
	 * @return The result of the mathematical operation.
	 */
	[[nodiscard]]
	DistributedAmount operator-(const pecunia::Money& amount) const;
	/**
	 * @brief Performs subtraction on the distributed amount.
	 *
	 * @exception Error Thrown when the new distributed amount exceeds the maximum permissible
	 * value.
	 *
	 * @param amount The amount to increase the distributed amount by.
	 *
	 * @post The current instance will have the distributed amount decreased by the requested
	 * amount. The next occurrence date, and the maximum permissible value remain unchanged.
	 *
	 * @return The current instance after the mathematical operation.
	 */
	DistributedAmount& operator-=(const pecunia::Money& amount);
};

/**
 * @brief Full object stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_SURVEYING_EXPORT std::ostream& operator<<(std::ostream& out, const DistributedAmount& data);

}

#endif
