/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_UPDATING_RELEASE_INFORMATION_HPP
#define DRN_UPDATING_RELEASE_INFORMATION_HPP

#include <QDateTime>
#include <QString>

#include <compare>

#include <updating_export.h>

#include "Version.hpp"


namespace drn::updating
{

/**
 * @brief Represents details about a specific version of a software component.
 *
 * The structure is designed to provide essential metadata about a software version. It includes
 * fields for the version number, the release date, and a description of the changes or features
 * included in the release. This structure is particularly useful in scenarios where version
 * management or release notes need to be displayed, logged, or processed by application components.
 *
 * Use this structure to encapsulate version-related information in your application. For example,
 * it can be employed in dialogs showing release notes, such as the ReleaseNotesDialog class, or
 * in APIs that fetch and process version history, like version-fetching mechanisms.
 */
struct DRN_UPDATING_EXPORT ReleaseInformation
{
	/// Represents a semantic version associated with an entity, typically used for
	/// version management.
	SemanticVersion version_{};
	/// Represents the release date and time of a version. This field is used to store the date and
	/// time when a specific version of the software or update was released. It plays an essential
	/// role in providing contextual information about the timing of updates and changes.
	::QDateTime releasedAt_{};
	/// A member variable to store a textual description of the version information.
	::QString description_{};
};

/**
 * @brief Compares two objects using a three-way comparison, in the order of the semantic version,
 * release date and time, and then version description.
 *
 * @param lhs The left-hand operand used in the operation.
 * @param rhs The right-hand operand used in the operation.
 *
 * @return The result of the comparison of this object to the supplied object.
 */
DRN_UPDATING_EXPORT std::strong_ordering operator<=>(
	const ReleaseInformation& lhs,
	const ReleaseInformation& rhs
);

}

#endif
