/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_UPDATING_RELEASE_INFORMATION_FETCHER_HPP
#define DRN_UPDATING_RELEASE_INFORMATION_FETCHER_HPP

#include <pecunia/Expected.hpp>

#include <QObject>

#include <set>

#include <foundation/QtMemory.hpp>
#include <networking/ReplyNotifier.hpp>
#include <updating_export.h>

#include "ReleaseInformation.hpp"

namespace drn::networking
{

struct NotifyingSender;

}
namespace drn::updating
{
/**
 * @brief A utility class for retrieving the application version release information.
 *
 * Usage:
 * - Create an instance of VersionFetcher to initialize the version fetching process.
 * - Use provided methods to retrieve version information from the desired source.
 * - Process or store the fetched version data as needed within your application.
 *
 * It abstracts the complex logic behind version retrieval by providing a simple interface. It
 * ensures that version information is consistently parsed and represented, regardless of the source
 * of the data.
 */
class DRN_UPDATING_EXPORT ReleaseInformationFetcher final : public ::QObject
{
	Q_OBJECT

	networking::NotifyingSender& sender_;

private slots:
	void processReply(
		const pecunia::Expected<
			drn::networking::SuccessReply,
			drn::networking::FailureReply
		>& reply
	);

signals:
	/**
	 * @brief Triggered when the version information has been successfully fetched.
	 *
	 * This signal provides a list of version information objects that have been retrieved,
	 * allowing you to act upon or process them. It is typically used to notify other parts
	 * of the application about the available versions, whether for display, storage, or comparison.
	 *
	 * Usage:
	 * - Connect this signal to slots where you want to handle the retrieved version data.
	 * - The connected slot can process, display, or act upon the provided version information as
	 * required.
	 *
	 * @param versions A collection of version information representing the fetched version data.
	 * These objects may contain details such as version numbers, metadata, and other related
	 * information.
	 */
	void fetched(const std::set<drn::updating::ReleaseInformation>& versions);

public:
	/**
	 * @brief Full initialisation constructor for retrieving application version release information.
	 *
	 * @param sender The reference to a NotifyingSender instance used to send network requests.
	 * This object should handle the communication requirements needed to fetch version information.
	 * @param parent The Qt object that acts as the parent of the instance. This parameter ensures
	 * proper memory management and cleanup of the QObject hierarchy. If no parent is needed, an
	 * empty QtPtr can be used.
	 */
	ReleaseInformationFetcher(
		networking::NotifyingSender& sender,
		const drn::foundation::QtPtr<::QObject>& parent
	);
	/**
	 * @brief Initiates the process of retrieving version release information.
	 *
	 * This method constructs a network request to communicate with the server hosting the
	 * version release data. It formats the request URL, sets appropriate headers for
	 * authentication, and sends the request through a NotifyingSender instance.
	 *
	 * Usage:
	 * - Call this method to start the process of fetching version information.
	 * - Connect the `fetched` signal to your application logic to handle the retrieved data.
	 *
	 * The method establishes a connection to process and handle the server's response. This allows
	 * seamless integration of fetched version data into your application.
	 */
	void fetch();
};

}

#endif
