/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_UPDATING_UPDATE_CHECKER_HPP
#define DRN_UPDATING_UPDATE_CHECKER_HPP

#include <QObject>

#include <set>

#include <foundation/QtMemory.hpp>
#include <updating_export.h>

#include "ReleaseInformation.hpp"
#include "ReleaseInformationFetcher.hpp"


namespace drn::networking
{

struct NotifyingSender;

}
namespace drn::updating
{

/**
 * @brief This class is responsible for checking and notifying about software updates.
 */
class DRN_UPDATING_EXPORT UpdateChecker final : public ::QObject
{
	Q_OBJECT

	const foundation::QtPtr<ReleaseInformationFetcher> fetcher_;

private slots:
	void process(const std::set<drn::updating::ReleaseInformation>& versions);

signals:
	/**
	 * @brief Triggered when a new version has been released relative to the built version.
	 *
	 * Usage:
	 * - Connect to this signal in your application to receive updates when a new version
	 *   is detected.
	 *
	 * @param latest The newest version of the software available.
	 */
	void updateReleased(const drn::updating::SemanticVersion& latest);
	/**
	 * @brief Triggered when the completion of the application update check process.
	 *
	 * This signal is emitted after the update checking operation concludes, regardless of
	 * whether a new update was found, no updates were available, or an error occurred during
	 * the process. It serves as a notification that the check operation has finished and ensures
	 * any dependent resources or operations can proceed accordingly.
	 *
	 * Usage:
	 * This signal is commonly connected to a slot or lambda to handle any clean-up, logging, or
	 * subsequent actions once the update check is complete. For example, it is often used to signal
	 * that the `UpdateChecker` instance can be safely deleted to release resources.
	 */
	void finishedCheck();

public:
	/**
	 * @brief Full initialisation constructor that manages the update fetching operations. It
	 * initialises itself with a given NotifyingSender for sending network requests, and a parent
	 * QObject to maintain the lifespan of this object within an existing object hierarchy.
	 *
	 * @param sender Reference to the object responsible for sending network requests.
	 * @param parent A Qt smart pointer to a QObject that acts as the parent of this instance. It
	 * ensures the object is properly managed within Qt's ownership model.
	 */
	UpdateChecker(
		networking::NotifyingSender& sender,
		const foundation::QtPtr<::QObject>& parent
	);
	/**
	 * @brief Begins the process of checking if there is an application update.
	 */
	void check();
};

}

#endif
