/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_UPDATING_VERSION_HPP
#define DRN_UPDATING_VERSION_HPP

#include <QString>

#include <cstdint>
#include <optional>
#include <iosfwd>
#include <compare>

#include <updating_export.h>

namespace drn::updating
{

/**
 * @brief Represents a semantic version with major, minor, patch. Additionally, supports the
 * inclusion of optional pre-release, and build metadata components. It adheres to the Semantic
 * Versioning 2.0.0 standard.
 *
 * @see https://semver.org/ for more information.
 */
class DRN_UPDATING_EXPORT SemanticVersion
{
	friend DRN_UPDATING_EXPORT std::ostream& operator<<(
		std::ostream& out,
		const SemanticVersion& sv
	);

	std::uint8_t major_;
	std::uint8_t minor_;
	std::uint8_t patch_;
	std::optional<::QString> preRelease_;
	std::optional<::QString> buildMetadata_;

public:
	/**
	 * @brief Default constructor that initializes version to 0.0.0 with any pre-release nor build
	 * metadata.
	 */
	SemanticVersion();
	/**
	 * @brief Full initialisation constructor.
	 *
	 * @exception Error When the pre-release string is invalid.
	 * @exception Error When the build metadata string is invalid.
	 *
	 * @param major The major component of the version.
	 * @param minor The minor component of the version.
	 * @param patch The patch component of the version.
	 * @param preRelease Any pre-release identifier for the version.
	 * @param buildMetadata Any build metadata information.
	 */
	SemanticVersion(
		std::uint8_t major,
		std::uint8_t minor,
		std::uint8_t patch,
		std::optional<::QString> preRelease = std::nullopt,
		std::optional<::QString> buildMetadata = std::nullopt
	);
	/**
	 * @brief Full initialisation string constructor following the semantic version format.
	 *
	 * @exception Error When the version string is invalid.
	 *
	 * @param version The semantic version string, e.g. "1.2.3-alpha.1+20241019".
	 */
	explicit SemanticVersion(const ::QString& version);
	/// An accessor to the major version.
	[[nodiscard]]
	std::uint8_t major() const noexcept;
	/// An accessor to the minor version.
	[[nodiscard]]
	std::uint8_t minor() const noexcept;
	/// An accessor to the patch version.
	[[nodiscard]]
	std::uint8_t patch() const noexcept;
	/// An accessor to the pre-release version.
	[[nodiscard]]
	const std::optional<::QString>& preRelease() const noexcept;
	/// An accessor to the build metadata.
	[[nodiscard]]
	const std::optional<::QString>& buildMetadata() const noexcept;
	/**
	 * @brief Compares two objects using a three-way comparison.
	 *
	 * The comparison follows semantic versioning rules:
	 *		- If the major versions are different, the one with the higher major version is greater.
	 *		- If the major versions are identical, the minor versions are compared next.
	 *		- If the minor versions are identical, the patch versions are compared next.
	 *		- If the patch versions are identical, pre-release versions are compared next.
	 *		- Pre-release versions have lower precedence than the associated normal version.
	 *		- Build metadata is not considered at any point.
	 *
	 * @param other The other object to compare with.
	 *
	 * @return The result of the comparison of this object to the supplied object.
	 */
	[[nodiscard]]
	std::strong_ordering operator<=>(const SemanticVersion& other) const noexcept;

	[[nodiscard]]
	bool operator==(const SemanticVersion& other) const noexcept;

	[[nodiscard]]
	bool operator!=(const SemanticVersion& other) const noexcept = default;
};

/**
 * @brief Populates an output stream with a semantic version.
 *
 * @param out The output stream to populate.
 * @param sv The version object to fill the stream with.
 *
 * @post The output stream contains the version in the form:
 *		- major.minor.patch-preRelease+buildMetadata
 *
 * @return The output stream with the version.
 */
DRN_UPDATING_EXPORT std::ostream& operator<<(std::ostream& out, const SemanticVersion& sv);
/**
 * @brief Converts a SemanticVersion object to its string representation. It provides an
 * human-readable string representation. Use this function when you need a textual representation of
 * a SemanticVersion object, particularly for logging or displaying version information in the UI or
 * output.
 *
 * @param sv The SemanticVersion object to be converted to a string.
 *
 * @return A string representation of the provided SemanticVersion object. The result will
 * follow the semantic versioning format, i.e., "major.minor.patch-preRelease+buildMetadata".
 */
DRN_UPDATING_EXPORT ::QString toString(const SemanticVersion& sv);
/**
 * @brief Full object Qt stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param data The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_UPDATING_EXPORT ::QDebug operator<<(::QDebug out, const SemanticVersion& data);

}

#endif
